#===============================================================
#
#   flincs.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package flincs;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
FLINCS_get_incs
FLINCS_get_incs_files
FLINCS_get_incs_tree
FLINCS_get_bld_incs
FLINCS_get_stubs_incs
FLINCS_get_flags
FLINCS_get_flags_files
FLINCS_get_flags_tree
FLINCS_get_bld_flags
);

use strict;
use integer;

use glo::env;
use gbsenv;
use target;
use gbsfileglo;
use scope;
use incs;
use flags;




sub FLINCS_get_incs($$);
sub FLINCS_get_incs_files($$);
sub FLINCS_get_incs_tree($$);
sub FLINCS_get_bld_incs($);
sub FLINCS_get_stubs_incs();
sub FLINCS_get_flags($$);
sub FLINCS_get_flags_tree($$);
sub FLINCS_get_flags_files($$);
sub FLINCS_get_bld_flags($);

sub ext_files($$);
sub get_scope_path($);




my $GBS_ROOT_PATH = GBSENV_gbs_root_path();
my $GBS_SYSBUILD_PATH = "$GBS_ROOT_PATH/sysbuild";
my $GBS_SCRIPTS_PATH = GBSENV_scripts_path();
my $GBS_TARGET = ENV_getenv( 'GBS_TARGET');
my $GBS_TARGET_PLUGIN = ENV_getenv( 'GBS_TARGET_PLUGIN');




sub FLINCS_get_incs($$)
{
my ($src_type,
$component,	    # '' = for non-gbs subsys
) = @_;
my @includes;

foreach my $ref (FLINCS_get_incs_tree( $src_type, $component))
{
my (undef, $type, $lines_ref) = @{$ref};
push @includes, @{$lines_ref}
if (index( 'LSE', $type) >= 0);
}


return @includes;
}







sub FLINCS_get_incs_tree($$)
{
my ($src_type,
$component,	    # '' = for non-gbs subsys
) = @_;
my @incsfile_refs;

my $src_type_name = substr( $src_type, 1);
my $file = "incs_$src_type_name";

my $is_glkb_type;
my $include_inc;
my $include_bld;
if ($src_type eq '.glkb')  # .include in glkb files
{
$is_glkb_type = 0;
$include_inc = 1;
$include_bld = 0;
} elsif (TARGET_is_src_type( $GBS_TARGET, $src_type))
{
(my $gen_type, $include_inc, $include_bld) =
TARGET_get_src_items( $GBS_TARGET, $src_type, qw( TYPE INCLUDE_INC INCLUDE_BLD));
$is_glkb_type = (substr( $gen_type, 0, 2) eq 'gl');
} else
{
ENV_error_exit( "Unknown src_type '$src_type'");
}

if ($component && !$is_glkb_type)
{



my @includes = ( "../loc", "../inc" ) if ($include_inc);
push @includes, "../bld/$GBS_TARGET" if ($include_bld);
push @incsfile_refs, [ 'Local', 'L', [ @includes ] ];




my @subdirs;
push @subdirs, '/inc' if ($include_inc);
push @subdirs, "/bld/$GBS_TARGET" if ($include_bld);
my @lines = SCOPE_get_includes( $component, @subdirs);
push @incsfile_refs, [ 'scope.gbs', 'S', [ @lines ] ];
}




foreach my $file_spec (reverse( ext_files( $component, $file)))
{
my @lines = INCS_get_includes( $file_spec);
push @incsfile_refs, [ $file_spec, 'E', [ @lines ] ];
}




my $file_spec = TARGET_get_bld_incs_file( $GBS_TARGET, $src_type);
if ($file_spec)
{
my @lines = INCS_get_includes( $file_spec);
push @incsfile_refs, [ $file_spec, 'B', [ @lines ] ];
}




if (!$is_glkb_type)
{
my @includes;
my $file_spec;
$file_spec = "$GBS_ROOT_PATH/sysbuild/makemake_stubs/ALL";
push @includes, $file_spec if (-s $file_spec);

$file_spec = "$GBS_ROOT_PATH/sysbuild/makemake_stubs/$GBS_TARGET";
push @includes, $file_spec if (-s $file_spec);

push @incsfile_refs, [ 'makemake_stubs', 'M', [ @includes ] ]
if (@includes);
}

return @incsfile_refs;
}




sub FLINCS_get_incs_files($$)
{
my ($src_type,
$component,	    # '' = for non-gbs subsys
) = @_;


my $src_type_name = substr( $src_type, 1);
my $file = "incs_$src_type_name";

return reverse( ext_files( $component, $file));
}




sub FLINCS_get_bld_incs($)
{
my ($src_type) = @_;
my @includes;

my @files = TARGET_get_bld_incs_file( '', $src_type);
@includes = map { INCS_get_includes( $_) } @files;


return @includes;
}




sub FLINCS_get_stubs_incs()
{
my @includes;

@includes = ( "$GBS_ROOT_PATH/sysbuild/makemake_stubs/ALL",
"$GBS_ROOT_PATH/sysbuild/makemake_stubs/$GBS_TARGET",
"$GBS_SCRIPTS_PATH/plugins/build/ALL/makemake_stubs",
);
push @includes, "$GBS_SCRIPTS_PATH/plugins/build/$GBS_TARGET_PLUGIN/makemake_stubs"
if ($GBS_TARGET_PLUGIN ne '');


return @includes;
}




sub FLINCS_get_flags($$)
{
my ($src_type,
$component,	    # '' = for non-gbs subsys
) = @_;
my @flags;

my @files = FLINCS_get_flags_files( $src_type, $component);
push @flags, map { FLAGS_get_flags( $_) } @files;


return @flags;
}







sub FLINCS_get_flags_tree($$)
{
my ($src_type,
$component,
) = @_;
my @flagsfile_refs;

my $src_type_name = substr( $src_type, 1);
my $file = "flags_$src_type_name";

foreach my $file_spec (ext_files( $component, $file))
{
my @lines = FLAGS_get_flags( $file_spec);
push @flagsfile_refs, [ $file_spec, 'E', [ @lines ] ];
}
my $file_spec = TARGET_get_bld_flags_file( '', $src_type);
if ($file_spec)
{
my @lines = FLAGS_get_flags( $file_spec);
push @flagsfile_refs, [ $file_spec, 'B', [ @lines ] ];
}

return @flagsfile_refs;
}




sub FLINCS_get_flags_files($$)
{
my ($src_type,
$component,	    # '' = for non-gbs subsys
) = @_;


my $src_type_name = substr( $src_type, 1);
my $file = "flags_$src_type_name";

return ext_files( $component, $file);
}




sub FLINCS_get_bld_flags($)
{
my ($src_type) = @_;
my @flags;

my @files = TARGET_get_bld_flags_file( '', $src_type);
push @flags, map { FLAGS_get_flags( $_) } @files;

return @flags;
}




sub ext_files($$)
{
my ($component,	# '' = for non-gbs subsys
$file,	    	# without .gbs
) = @_;
my @files;

foreach my $path (get_scope_path( $component))
{
my $file_spec = GBSFILEGLO_select_usr_gbs( "$path/$file");
push @files, $file_spec
if ( defined $file_spec);
}


return @files;
}




sub get_scope_path($)
{
my ($component) = @_;  # '' = for non-gbs subsys
my @dirs;

my $subsys = ENV_getenv( 'GBS_SUBSYS');
my $subsys_path = "$GBS_ROOT_PATH/dev/$subsys";

@dirs = ( "$GBS_ROOT_PATH/sysbuild",
"$GBS_ROOT_PATH/sysbuild/$GBS_TARGET",
"$subsys_path/build",
"$subsys_path/build/$GBS_TARGET",
);

if ($component ne '')
{
my $gbs_component_path = "$subsys_path/comp/$component";

push @dirs,
"$gbs_component_path/opt",
"$gbs_component_path/opt/$GBS_TARGET";
}

return @dirs;
}

return 1;

