#===============================================================
#
#   gbsbgglo.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package gbsbgglo;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
GBSBGGLO_init
GBSBGGLO_main
GBSBGGLO_set
GBSBGGLO_exec
GBSBGGLO_handle_rc
);

use strict;
use integer;

use glo::env;
use glo::time;
use glo::slurp;
use gbsenv;
use fail;




sub GBSBGGLO_init($$);
sub GBSBGGLO_main($$);
sub GBSBGGLO_set($$);
sub GBSBGGLO_exec($$$@);
sub GBSBGGLO_handle_rc($$$$);

sub do_banner(@);
sub do_echo(@);
sub do_glkb($$$$@);
sub do_sca($$$@);
sub do_scasum($$);
sub banner($$@);
sub print_fail_msg($);




my $GBS_IGNORE_ERRORS = $ENV{GBS_IGNORE_ERRORS};

my $GBS_ROOT_PATH = GBSENV_gbs_root_path();
my $GBS_SUBSYS = $ENV{GBS_SUBSYS};
my $GBS_COMPONENT = ENV_getenv( 'GBS_COMPONENT');




my $BATCH_NAME = '';	# To be used in logging. E.g.: gbsgen, gbsgen_post, gbsaudit, gbsaudit_post




my $NR_FILES = 0;
my $FAIL_COUNT = 0;

my $CWD;	    # filled from $GBSBG_SCRIPT_SPEC file




my %FUNCTIONS = (

BANNER  => [ \&do_banner,	    1, 0 ],	# $first_line, @lines
ECHO    => [ \&do_echo,	    1, 0 ],	# @lines
SET     => [ \&GBSBGGLO_set,    2, 0 ],	# $name, $value
EXEC    => [ \&GBSBGGLO_exec,   7, 1 ],	# $file, $must_print_commands, $out_files, @command_data_lines
GLKB    => [ \&do_glkb,	    8, 1 ],	# $file, $must_print_commands, $glkb_args, $out_files, @command_data_lines
SCA	    => [ \&do_sca,	    3, 1 ],	# $file, $unix_style_search, $out_files, @command_data_lines
SCASUM  => [ \&do_scasum,       2, 1 ],	# $file, $sum_filespec
);




sub GBSBGGLO_init($$)
{
my ($batch_name,
$component,	    # may be '-'
) = @_;




$BATCH_NAME = uc $batch_name;
{
my ($os, $host) = ENV_uname();
ENV_say( 1, "$BATCH_NAME: Running on $host ($os)");
}




if ($component eq '' || $component eq '-')
{
$GBS_COMPONENT = '';

$ENV{GBS_COMPONENT} = '';
$ENV{GBS_COMPONENT_PATH} = '';

$CWD = ENV_chdir( "$GBS_ROOT_PATH/dev/$GBS_SUBSYS");
} else
{
my $new_component_path = "$GBS_ROOT_PATH/dev/$GBS_SUBSYS/comp/$component";
$GBS_COMPONENT = $component;
banner( '*', "Component: $GBS_COMPONENT");
$ENV{GBS_COMPONENT} = $GBS_COMPONENT;
$ENV{GBS_COMPONENT_PATH} = $new_component_path;

$CWD = ENV_chdir( "$new_component_path/src");
}
}




sub GBSBGGLO_main($$)
{
my ($gbsbg_script_spec,
$func_ref,		# [ $key, $function ]
) = @_;
my $rc = 0;




{
my ($key, $function) = @{$func_ref};
ENV_fatal( "Invalid function-key '$key'")
if (!exists $FUNCTIONS{$key});
$FUNCTIONS{$key}->[0] = $function;
}




my @lines = SLURP_file( $gbsbg_script_spec);
while (@lines)
{
my $command = shift @lines;

my ($function, $min_nr_lines, $must_count) = @{$FUNCTIONS{$command}};
my $nr_lines = shift @lines;
ENV_fatal( "$command: nr_lines ($nr_lines) < min_nr_lines ($min_nr_lines)")
if ($nr_lines < $min_nr_lines);




$NR_FILES++ if ($must_count);
my @sub_lines;
while ($nr_lines > 0)
{
my $sub_line = shift @lines;
push @sub_lines, $sub_line;
$nr_lines--;
}




my $this_rc = &$function( @sub_lines);
$rc = $this_rc if ($this_rc > $rc);
if ($this_rc != 0)
{
print_fail_msg( scalar @lines);
last if (!$GBS_IGNORE_ERRORS);
}
}

return $rc;
}




sub END
{
my $rc = $?;
my $rc_txt;
if ($rc == 0)
{
$rc_txt = 'Completed';
} else
{
$rc_txt = "Unsuccessful ($rc)";
$rc_txt .= " (Count: $FAIL_COUNT)"
if ($FAIL_COUNT > 0);
}
ENV_say( 1, "$BATCH_NAME: $rc_txt");
}




sub do_banner(@)
{
my ($first_line,
@lines
) = @_;

banner( '*', $first_line, @lines);

return 0;
}




sub do_echo(@)
{
my (@lines) = @_;

ENV_say( 1, @lines);

return 0;
}




sub GBSBGGLO_set($$)
{
my ($name,
$value,
) = @_;
$value = '' if (!defined $value);


$value = ' ' if ($value eq '');
$ENV{$name} = $value;

return 0;
}




sub GBSBGGLO_exec($$$@)
{
my ($file,
$must_print_commands,	# scripts should print their own command(s)
$out_files,		# List
@command_data_lines,	# ($command_type, $command_line, $ok_values, $exec_condition), ...

) = @_;
my $rc = 0;



my $nr_command_data_lines = @command_data_lines;
my $nr_commands = $nr_command_data_lines / 4;
my $count = 0;
for (my $i = 0; $i < $nr_command_data_lines; $i += 4)
{


$count++;
ENV_say( 0, '.')
if ($i != 0);
my ($command_type, $command_line, $ok_values, $exec_condition) = @command_data_lines[$i..$i + 3];





my $must_exec = 1;
if ($exec_condition ne '-')
{
my $type = substr( $exec_condition, 0, 1);	# F or V
my $cond = substr( $exec_condition, 1, 1);	# 0 or 1
if ($type eq 'F')
{
my $file_spec = substr( $exec_condition, 2);
$file_spec = ENV_expand_envs( $file_spec);
my $exists = (-e $file_spec) ? 1 : 0;

$must_exec = ($cond == $exists);
$exec_condition = "File exists ($file_spec != $cond)";
} elsif ($type eq 'V')
{
my ($name, $value) = split( / /, substr( $exec_condition, 2), 2);
my $actual_value = ENV_getenv( $name);
$actual_value = ENV_expand_envs( $actual_value);
my $equal = ($actual_value eq $value) ? 1 : 0;

$must_exec = ($cond == $equal);
$exec_condition = "EnvVar ($name==$value) != $cond";
} else
{
ENV_fatal( "Invalid exec-condition type ($type)");
}
}
if ($must_exec)
{



if ($command_type eq 'B')	# direct builtin (e.g. echo)
{
my ($command, $args) = split( / /, $command_line, 2);
if ($command eq 'ECHO')
{
ENV_say( 0, $args);
} elsif ($command eq 'SET')
{
my ($name, $value) = split( / /, $args, 2);

$value = ENV_expand_envs( $value);
ENV_setenv( $name => $value);
} else
{
ENV_fatal( 'Invalid Special Command',
"'$command_line'");
}
} elsif ($command_type eq 'D' || $command_type eq 'S')  # direct or script
{
my $full_command = ENV_expand_envs( $command_line);
if ($must_print_commands eq '1' || ($must_print_commands eq '-' && $command_type eq 'D'))
{
if ($nr_commands == 1)
{
ENV_say( 0, $full_command);
} else
{
ENV_say( 0, "$count:$full_command");
}
}
$rc = ENV_system( $full_command, undef, undef);





my $rc_ok;
if ($ok_values eq '0')
{
$rc_ok = ($rc == 0)
} else
{
my @ok_values = split( ' ', $ok_values);
$rc_ok = grep( $_ == $rc, @ok_values);
}
if ($rc_ok)
{
$rc = 0;
} else
{
GBSBGGLO_handle_rc( $rc, $file, 'exec', $out_files);
last;
}
} else
{
ENV_fatal( "Invalid command_type '$command_type'");
}
} else
{
my $text = "Skipped because of pre-condition: $exec_condition";
if ($nr_commands == 1)
{
ENV_say( 0, $text);
} else
{
ENV_say( 0, "$count:$text");
}
}
}
ENV_say( 0, '-');

return $rc;
}




sub GBSBGGLO_handle_rc($$$$)
{
my ($rc,
$file,
$sub_command,
$out_files,
) = @_;


if ($rc > 0)
{
ENV_say( 1, "$BATCH_NAME: ### Failed ($rc) ***");
FAIL_log( $sub_command, $rc, $GBS_SUBSYS, $GBS_COMPONENT, $file);
$FAIL_COUNT++;




if ($out_files ne '')
{
foreach my $out_file (split( ' ', $out_files))
{

if (-f $out_file)
{
ENV_say( 1, "  Deleting '$out_file'...");
unlink $out_file;
}
}
}
}
ENV_say( 0, '-');

return $rc
}




sub do_glkb($$$$@)
{
my ($file,
$must_print_commands,
$glkb_args,	    # "$gen_type $glkb_type $glkb_file $out_file_spec"
$out_files,	    # List
@command_line,
) = @_;


ENV_fatal( 'GLKB_func not defined');


}





sub do_sca($$$@)
{
my ($file,
$unix_style_search,	# bool
$out_files,		# List
@command_lines
) = @_;


ENV_fatal( 'SCA_func not defined');


}





sub do_scasum($$)
{
my ($file,
$sum_filespec,
) = @_;


ENV_fatal( 'SCASUM_func not defined');


}




sub banner($$@)
{
my ($pre_char,
$first_line,
@other_lines
) = @_;

my $pre = $pre_char x 3;
my $time_stamp =  TIME_time2num();
my $time_stamp_indent = ' ' x length( $time_stamp);
ENV_say( 1, $pre);
ENV_say( 1, "$pre $time_stamp - $first_line");
map { ENV_say( 1, "$pre $time_stamp_indent - $_") } @other_lines;
ENV_say( 1, $pre);
}




sub print_fail_msg($)
{
my ($nr_remaining_lines) = @_;

my ($title, $sub_line);

$title = $BATCH_NAME;

if ($nr_remaining_lines > 0)
{
if ($GBS_IGNORE_ERRORS)
{
$sub_line = "Continuing with remaining commands";
} else
{
$sub_line = "Remaining commands will be skipped";
}
} else
{
$sub_line = "No Remaining commands";
}
banner( '#', "$title: ### FAILED ###",
$sub_line);
}

return 1;


