#===============================================================
#
#   gbscleanup.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package gbscleanup;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
GBSCLEANUP_main
);

use strict;
use integer;

use glo::env;
use glo::ask;
use gbsenv;
use cleanup;




sub GBSCLEANUP_main($$$);

sub cleanup_most();
sub cleanup_all();
sub do_cleanup_most();
sub cleanup_bld();
sub cleanup_aud();
sub cleanup_make();
sub select_targets();
sub select_audits();




my $GBS_ROOT_PATH = GBSENV_gbs_root_path();
my $GBS_SUBSYS = ENV_getenv( 'GBS_SUBSYS');
my $GBS_COMPONENT = ENV_getenv( 'GBS_COMPONENT');
my $GBS_TARGET = ENV_getenv('GBS_TARGET');
my $GBS_AUDIT = ENV_getenv('GBS_AUDIT');
my @GBS_ALL_TARGETS = ENV_getenv( 'GBS_ALL_TARGETS');
my @GBS_AUDITS = ENV_getenv( 'GBS_AUDITS');

my $ALL = '* (ALL)';




sub GBSCLEANUP_main($$$)
{
my ($menu_entry,
$data_ref,
$entries_ref) = @_;

my @menu_items;
if ($GBS_ROOT_PATH eq '')
{
ENV_say( 1, "No current Root");
push @menu_items,
[ "Cleanup User currency file (.gbsrc) (immediate)",
\&CLEANUP_currencies ];
} else
{
ENV_say( 1, "Root: $GBS_ROOT_PATH");
push @menu_items, (
[ "Cleanup MOST (*)                                 (immediate)",
\&cleanup_most ],
[ "*Cleanup Build <target> directories",
\&cleanup_bld ],
[ "*Cleanup Audit <audit>/<target> directories",
\&cleanup_aud ],
[ "*Cleanup Make-files",
\&cleanup_make ],
[ "*Cleanup EXPORT directories                      (immediate)",
sub { CLEANUP_export( '*') } ],
[ "*Cleanup RES directories                         (immediate)",
sub { CLEANUP_res( '*') } ],
[ "*Cleanup TMP directories                         (immediate)",
\&CLEANUP_tmp_directory ],
[ "*Cleanup obsolete BLD files                      (immediate)",
\&CLEANUP_obsolete_bld_files ],
[ " Cleanup SILO user directories (e.g.: doxygen)   (immediate)",
\&CLEANUP_silo_user_directory ],
[ " Cleanup SILO cache directories                  (immediate)",
\&CLEANUP_silo_cache_directory ],
[ " Cleanup .usr files                              (immediate)",
\&CLEANUP_usr_files ],
[ " Cleanup currency files (.gbsrc)                 (immediate)",
\&CLEANUP_currencies ],
[ "Cleanup ALL                                      (immediate - sure?)",
\&cleanup_all ],
);
}
ASK_menu( 'Select Cleanup function to perform',
\@menu_items,
$entries_ref);
}




sub cleanup_most()
{
ENV_say( 1, "** Cleanup MOST derived files and directories **");

do_cleanup_most();

ENV_say( 1, "** Cleanup of MOST completed **");
}




sub cleanup_all()
{
ENV_say( 1, "** Cleanup ALL derived files and directories **");

do_cleanup_most();

CLEANUP_silo_user_directory();
CLEANUP_silo_cache_directory();
CLEANUP_usr_files();
CLEANUP_currencies();

ENV_say( 1, "** Cleanup of ALL completed **");
}




sub do_cleanup_most()
{
foreach my $target (@GBS_ALL_TARGETS)
{
CLEANUP_system_bld( $target);
CLEANUP_system_make_files( $target);
foreach my $audit (@GBS_AUDITS)
{
CLEANUP_system_aud( $audit, $target);
}
}

CLEANUP_export( '*');
CLEANUP_res( '*');
CLEANUP_tmp_directory();

CLEANUP_small_stuff();
}




sub cleanup_bld()
{
ENV_say( 1, "Cleanup Build Directories");

my @targets = select_targets();
if (@targets)
{
my @entries = ('');	    # <none>
push @entries,         [ "Cleanup current All       Build directories ($GBS_ROOT_PATH)",
sub { CLEANUP_system_bld( $_[0]) } ];
if ($GBS_SUBSYS ne '')
{
push @entries,     [ "Cleanup current SubSystem Build directories ($GBS_SUBSYS)",
sub { CLEANUP_subsys_bld( $GBS_SUBSYS, $_[0]) } ];
if ($GBS_COMPONENT ne '')
{
push @entries, [ "Cleanup current Component Build directories ($GBS_COMPONENT)",
sub { CLEANUP_comp_bld( $_[0]) } ];
}
}
my $index = ASK_index_from_menu( "Select Deletion Level for Target(s)", 0, @entries);
if ($index > 0)
{
my $cleanup_func = $entries[$index]->[1];
foreach my $target (@targets)
{
$cleanup_func->( $target);
}
}
}
}




sub cleanup_aud()
{
ENV_say( 1, "Cleanup Audit Directories");

my @audits = select_audits();
if (@audits)
{
my @targets = select_targets();
if (@targets)
{
my @entries = ('');	    # <none>
push @entries,         [ "Cleanup current All       Audit directories ($GBS_ROOT_PATH)",
sub { CLEANUP_system_aud( $_[0], $_[1]) } ];
if ($GBS_SUBSYS ne '')
{
push @entries,     [ "Cleanup current SubSystem Audit directories ($GBS_SUBSYS)",
sub { CLEANUP_subsys_aud( $GBS_SUBSYS, $_[0], $_[1]) } ];
if ($GBS_COMPONENT ne '')
{
push @entries, [ "Cleanup current Component Audit directories ($GBS_COMPONENT)",
sub { CLEANUP_comp_aud( $_[0], $_[1]) } ];
}
}
my $index = ASK_index_from_menu( "Select Deletion Level for Audit(s)/Target(s) ", 0, @entries);
if ($index > 0)
{
my $cleanup_func = $entries[$index]->[1];
foreach my $audit (@audits)
{
foreach my $target (@targets)
{
$cleanup_func->( $audit, $target);
}
}
}
}
}
}




sub cleanup_make()
{
ENV_say( 1, "Cleanup Make-files");

my @targets = select_targets();
if (@targets)
{
my @entries = ('');	    # <none>
push @entries,         [ "Cleanup current System    Make files ($GBS_ROOT_PATH)",
sub { CLEANUP_system_make_files( $_[0]) } ];
if ($GBS_SUBSYS ne '')
{
push @entries,     [ "Cleanup current SubSystem Make files ($GBS_SUBSYS)",
sub { CLEANUP_subsys_make_files( $_[0]) } ];
}
my $index = ASK_index_from_menu( "Select Deletion Level for Target(s) ", 0, @entries);
if ($index > 0)
{
my $cleanup_func = $entries[$index]->[1];
foreach my $target (@targets)
{
$cleanup_func->( $target);
}
}
}
}




sub select_targets()
{
my @targets;

if (@GBS_ALL_TARGETS == 0)
{
ENV_say( 1, "No Targets");
} elsif( @GBS_ALL_TARGETS == 1)
{
@targets = @GBS_ALL_TARGETS;
ENV_say( 1, "Single Target '@targets'");
} else
{
my @list = ('', $ALL, @GBS_ALL_TARGETS);	# '' == <none>
my $target = ASK_value_from_menu( "Select Target ('*' == ALL)", \$GBS_TARGET, @list);
@targets = ($target eq '') ? () : ($target eq $ALL) ? @GBS_ALL_TARGETS : ($target);
}

return @targets;
}




sub select_audits()
{
my @audits;

if (@GBS_AUDITS == 0)
{
ENV_say( 1, "No Audits");
} elsif( @GBS_AUDITS == 1)
{
@audits = @GBS_AUDITS;
ENV_say( 1, "Single Audit '@audits'");
} else
{
my @list = ('', $ALL, @GBS_AUDITS);	# '' == <none>
my $audit = ASK_value_from_menu( "Select Audit ('*' == ALL)", \$GBS_AUDIT, @list);
@audits = ($audit eq '') ? () : ($audit eq $ALL) ? @GBS_AUDITS : ($audit);
}

return @audits;
}

return 1;

