#===============================================================
#
#   gbsexport.pl
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

print "*GBSEXPORT*\n" if ($ENV{GBSDEBUG_FILE});
use strict;
use integer;

use lib $ENV{GBS_SCRIPTS_PATH};
use glo::env;
use glo::slurp;
use glo::file;
use glo::genopt;
use fspec;
use gbsenv;
use gbsglo;
use gbsoptenv;
use validate;
use target;
use cleanup;
use export;




sub do_all_exports(@);
sub select_export_locs(@);
sub select_export(@);
sub do_export($$$);
sub prune_directory($$);
sub error(@);
sub contains_dir($$);
sub debug_keeps();




my $GBS_ROOT_PATH;
my $GBS_SUBSYS;
my $GBS_TARGET = ENV_getenv( 'GBS_TARGET');
my $GBS_TARGETS = ENV_getenv( 'GBS_TARGETS');

my %OUT_DIRS;



my %KEEPS;



my @EXP_REFS;

my @RES_REFS;







$| = 1;
my $RC = 0;
GBSENV_init();




my @genopts = (
[ '<*>',     'export_file_locations', 'sao',  "", [ "<Component>... and/or GEN: or SUB:",
". == Current Component. <empty> == <all>"] ],
[ 'i',       'ignore_errors',         'bso',   0, "Continue generation after error(s)" ],
[ 'c',       'cleanup_all',	          'bso',   0, "Delete all, independent from target" ],
[ 'verbose', 'verbose',	          'bso',   0, "Extensive messaging" ],
);
my @genenvs = qw( APP_* );
GENOPT_set_optdefs( 'gbsexport', \@genopts,
'Fill the export and/or res directories from the Components and/or Gen or SubSystem by Target',
undef);
GENOPT_set_envdefs( \&GBSOPTENV_set, \&GBSOPTENV_print, 'GBS_', \@genenvs);
GENOPT_parse();
my @EXPORT_LOCS = GENOPT_get( 'export_file_locations');
my $IGNORE_ERRORS = GENOPT_get( 'ignore_errors');
my $MUST_CLEANUP_ALL = GENOPT_get( 'cleanup_all');
my $VERBOSE = GENOPT_get( 'verbose');




ENV_setenv( GBS_IGNORE_ERRORS => $IGNORE_ERRORS);




($GBS_ROOT_PATH, $GBS_SUBSYS) = VALIDATE_subsys();
my $GBS_SUBSYS_PATH = "$GBS_ROOT_PATH/dev/$GBS_SUBSYS";
my $IS_FULL_GBS = GBSGLO_subsystem_is_full_gbs( $GBS_SUBSYS);
if ($IS_FULL_GBS)
{
VALIDATE_target();
} else
{
ENV_error_exit( "No current Target")
if ($GBS_TARGET eq '');
}
ENV_say( 1, "SUBSYS=$GBS_SUBSYS TARGET=$GBS_TARGET EXPORT=(@EXPORT_LOCS)");

my $GBS_RES_PATH = "$GBS_ROOT_PATH/res";
my $GBS_COMP_PATH = "$GBS_SUBSYS_PATH/comp";
my $GBS_BUILD_PATH = "$GBS_SUBSYS_PATH/build";
my $GBS_EXPORT_PATH = "$GBS_SUBSYS_PATH/export";
my $GBS_RES_SUBSYS_PATH = "$GBS_RES_PATH/$GBS_SUBSYS";

my $EXPORT_DIR_EXISTS = -d $GBS_EXPORT_PATH;
my $RESSUB_DIR_EXISTS = -d $GBS_RES_SUBSYS_PATH;

my @GBS_ALL_TARGETS = ENV_getenv('GBS_ALL_TARGETS');
my @OTHER_TARGETS = grep( $GBS_TARGET ne $_, @GBS_ALL_TARGETS);
my $NR_SPECIFIED_FILES = 0;
my $NR_COMPONENTS = 0;
my $GEN_FILE_INCLUDED = 0;




if ($RESSUB_DIR_EXISTS || $EXPORT_DIR_EXISTS)
{

my $export_fh = FILE_lock_ex_wait( "$GBS_SUBSYS_PATH/.export.lck") if ($EXPORT_DIR_EXISTS);
my $ressub_fh = FILE_lock_ex_wait( "$GBS_RES_PATH/.$GBS_SUBSYS.lck") if ($RESSUB_DIR_EXISTS);

TARGET_set_envs( $GBS_TARGET);

$RC = do_all_exports( @EXPORT_LOCS);

FILE_unlock( $ressub_fh) if ($RESSUB_DIR_EXISTS);
FILE_unlock( $export_fh) if ($EXPORT_DIR_EXISTS);
} else
{
ENV_say( 1, "No 'res' or 'export' dir for this SybSystem '$GBS_SUBSYS'");
}

ENV_exit( $RC);




sub END
{
ENV_print_end_msg( 1);
}




sub do_all_exports(@)
{
my @export_locs = @_;    # GEN:, SUB:, Components (including '.')
my $rc = 0;




my @selected_export_locs = select_export_locs( @export_locs);
ENV_say( 1, "Selected: @selected_export_locs")
if (@selected_export_locs);




select_export( @selected_export_locs);






if ($EXPORT_DIR_EXISTS)
{
ENV_say( 1, "");
ENV_say( 1, "Handling 'EXPORT'");
if ($MUST_CLEANUP_ALL)
{
CLEANUP_export( $GBS_SUBSYS);
} else
{
prune_directory( 'EXPORT', $GBS_EXPORT_PATH);
}
$rc = do_export( 'EXPORT', $GBS_EXPORT_PATH, \@EXP_REFS);
}
if ($RESSUB_DIR_EXISTS)
{
ENV_say( 1, "");
ENV_say( 1, "Handling 'RESSUB'");
if ($MUST_CLEANUP_ALL)
{
CLEANUP_res( $GBS_SUBSYS);
} else
{
prune_directory( 'RESSUB', $GBS_RES_SUBSYS_PATH);
}
$rc = do_export( 'RESSUB', $GBS_RES_SUBSYS_PATH, \@RES_REFS);
}
return $rc;
}




sub select_export_locs( @)
{
my @export_locs = @_;
my @selected_export_locs;

if ($IS_FULL_GBS)
{
if (@export_locs)
{



ENV_error_exit( "No comp directory in SubSystem '$GBS_SUBSYS'")
if (! -d $GBS_COMP_PATH);

foreach my $export_loc (@export_locs)
{
if ($export_loc eq 'GEN:')
{
ENV_error_exit( "No export.gbs in GEN:'!")
if (! -f "$GBS_BUILD_PATH/export.gbs");
$GEN_FILE_INCLUDED = 1;
} elsif ($export_loc eq 'SUB:')
{
ENV_error_exit( "Cannot use 'SUB:' in Full-GBS SubSystem. Did you mean 'GEN:'?");
} else
{



if ($export_loc eq '.')
{
$export_loc = ENV_getenv( 'GBS_COMPONENT');
ENV_error_exit( "No current Component!")
if ($export_loc eq "");
}
ENV_error_exit( "Component '$export_loc' does not exist!")
if (! GBSGLO_component_exists( $GBS_ROOT_PATH, $GBS_SUBSYS, $export_loc));
ENV_error_exit( "No export.gbs in Component '$export_loc'!")
if (! -f "$GBS_COMP_PATH/$export_loc/export.gbs");
ENV_error_exit( "Component '$export_loc' does not build for Target '$GBS_TARGET'!")
if (! GBSGLO_component_does_build( $GBS_SUBSYS, $export_loc, $GBS_TARGET));
$NR_COMPONENTS++;
}
push @selected_export_locs, $export_loc;
}
} else
{
@selected_export_locs = grep( -f  FSPEC_export( 'export', $GBS_ROOT_PATH, $GBS_SUBSYS, $_), GBSGLO_components( $GBS_SUBSYS));
$NR_COMPONENTS = @selected_export_locs;
if (-f FSPEC_export( 'export', $GBS_ROOT_PATH, $GBS_SUBSYS, ''))
{
push @selected_export_locs, 'GEN:';
$GEN_FILE_INCLUDED = 1;
}
}
} else
{
if (@EXPORT_LOCS)
{
ENV_error_exit( "Can only specify 'SUB:' for Non-GBS SubSystems (@EXPORT_LOCS)")
if (@EXPORT_LOCS != 1 || $EXPORT_LOCS[0] ne 'SUB:');
} else
{
@selected_export_locs = ( 'SUB:');
}
ENV_error_exit( "No export.gbs in this Non-GBS SubSys!")
if (! -f "$GBS_SUBSYS_PATH/export.gbs");
}

return @selected_export_locs;
}






sub select_export(@)
{
my @export_locs = @_;	# Component and/or 'GEN:' or 'SUB:'

foreach my $export_loc (@export_locs)
{
my $export_type;
if ($export_loc eq 'GEN:')
{
$export_type = 'GEN';
} elsif ($export_loc eq 'SUB:')
{
$export_type = 'SUB';
} else
{
$export_type = 'CMP';
}

my @specs_refs = EXPORT_parse( $export_type, $export_loc, $GBS_TARGET);









my $skip_this_component = ($export_type eq 'CMP' &&
! GBSGLO_component_does_build( $GBS_SUBSYS, $export_loc, $GBS_TARGET));
ENV_say( 1, "Skipped Component '$export_loc' (Does not generate for this Target)")
if ($skip_this_component);
foreach my $specs_ref (@specs_refs)
{
my ($selected, $destination, $to_dir, $chmods_dir, @from_refs) = @{$specs_ref};





my @destinations = ($destination eq 'ANY') ? qw( EXPORT RESSUB) : ($destination);




my $must_export_this_dir;
if ($selected)
{
if (contains_dir( $to_dir, [ '*', $GBS_TARGET ]))
{
$must_export_this_dir = 1;
} elsif (contains_dir( $to_dir, \@OTHER_TARGETS ))
{
$must_export_this_dir = 0;
ENV_say( 1, "Skipped output directory '$to_dir' (Other Target)");
} else
{
$must_export_this_dir = 1;
}
} else
{
$must_export_this_dir = 0;
ENV_say( 1, "Skipped output directory '$to_dir' (Target Line Selector)");
}
my @dirs;
foreach my $dir (split '/', $to_dir)
{
push @dirs, $dir;
my $dir_spec = join( '/', @dirs);

foreach my $dest (@destinations)
{
if ($must_export_this_dir)
{
if (!exists $OUT_DIRS{$dest}->{$dir_spec})
{
$OUT_DIRS{$dest}->{$dir_spec} = $chmods_dir
}
}
$KEEPS{$dest}->{$dir_spec} = 1;
}
}
$to_dir =~ s/\*/$GBS_TARGET/g;




foreach my $ref (@from_refs)
{
my ($source, $to_file, $chmods_file) = @{$ref};



$NR_SPECIFIED_FILES++;

my $to_spec = ($to_dir) ? "$to_dir/$to_file" : $to_file;
map { $KEEPS{$_}->{$to_spec} = 1 } @destinations;
next if ($skip_this_component);

if ($must_export_this_dir)
{
if ($source eq 'INSTALL:')
{
my (undef, undef, undef, $command, @args) =  @{$ref};

my $dest_ref = [ $to_spec, $chmods_file, 'I', $export_type, $command, @args ];
push @EXP_REFS, $dest_ref
if ($destination eq 'EXPORT' || $destination eq 'ANY');
push @RES_REFS, $dest_ref
if ($destination eq 'RESSUB' || $destination eq 'ANY');
} else
{
my (undef, undef, undef, $from_dir, $from_file, $from_file_spec) = @{$ref};


if (!contains_dir( $from_dir, \@OTHER_TARGETS))
{
if (-f $from_file_spec)
{


my $dest_ref = [ $to_spec, $chmods_file, 'C', $from_file_spec ];
push @EXP_REFS, $dest_ref
if ($destination eq 'EXPORT' || $destination eq 'ANY');
push @RES_REFS, $dest_ref
if ($destination eq 'RESSUB' || $destination eq 'ANY');
} else
{

error( "File not found ($source$from_file)",
"($from_file_spec)");
}
} else
{
ENV_say( 1, "Skipped '$to_dir' <== '$from_dir/$from_file' (Other Target)");
}
}
} else
{

}
}
}
}

}





sub do_export($$$)
{
my ($destination,	    	# 'EXPORT' or 'RESSUB'
$export_path,
$export_refs_ref,
) = @_;
my $rc = 0;

ENV_say( 1, "* Exporting: '$destination'...");




my $out_dir_destinations_ref = $OUT_DIRS{$destination};
foreach my $to_dir (sort( keys( %{$out_dir_destinations_ref})))
{
my $new_dir = "$export_path/$to_dir";
$new_dir =~ s/\*/$GBS_TARGET/g;

if (! -d $new_dir)
{
ENV_say( 1, "Create Directory: $new_dir");
my $chmods = $out_dir_destinations_ref->{$to_dir};
ENV_mkdir( $new_dir, 0751);		    # u=rwx,go=rx,o=x
map { ENV_chmod( $_, $new_dir) } split( ',', $chmods)
if ($chmods ne '');

} else
{

}
}




my $exported = 0;
my $copied = 0;
my $failed = 0;
my @copy_refs;
foreach my $ref (@{$export_refs_ref})
{
my ($to_subdir_file, $to_chmods, $action) = @{$ref};
$exported++;
my $to_file = "$export_path/$to_subdir_file";
my $replicated;
if ($action eq 'I')
{



my (undef, undef, undef, $export_type, $command, @args) = @{$ref};
unlink $to_file;
ENV_say( 1, "  Executing INSTALL: $command, @args");
ENV_pushd( ($export_type eq 'SUB') ? $GBS_SUBSYS_PATH : $GBS_BUILD_PATH);
my $rc = ENV_system( ENV_command_ref( $command, \@args) , undef, undef);
ENV_say( 0, '');
if ($rc == 0)
{
if (-e $to_file)
{
$replicated = 1;
} else
{
$replicated = -1;
ENV_say( 1, "  INSTALL Outputfile missing / not created ($to_subdir_file)");
}
} else
{
$replicated = -1;
ENV_say( 1, "  INSTALL Failed (rc=$rc)");
}
ENV_popd();
} else
{



my (undef, undef, undef, $from_file) =  @{$ref};
ENV_say( 1, "  $to_subdir_file <== $from_file")
if ($VERBOSE);
if (ENV_file_is_same( $from_file, $to_file))
{
$replicated = 0;
} else
{
$replicated = ENV_copy_file( $from_file, $to_file, 0);   # $exit_on_error = 0
}
}
if ($replicated == -1)
{
$failed++;
} elsif ($replicated == 1)
{
$copied++;
}




if ($replicated == 1)
{
ENV_chmod( 'ugo=r', $to_file);
map { ENV_chmod( $_, $to_file) } split( ',', $to_chmods)
if ($to_chmods ne '');
}
}

if ($failed > 0)
{
ENV_warn( "$failed copies failed");
$rc = 2;
}




if ($IS_FULL_GBS)
{
my $gen_file = ($GEN_FILE_INCLUDED) ? " and 'GEN:'" : '';
ENV_say( 1, "$NR_COMPONENTS Component(s)$gen_file. $NR_SPECIFIED_FILES file(s) specified, $exported exported, $copied copied");
} else
{
ENV_say( 1, "'SUB:'. $NR_SPECIFIED_FILES file(s) specified, $exported exported, $copied copied");
}

return $rc;
}




sub prune_directory($$)
{
my ($destination,	    	# 'EXPORT' or 'RESSUB'
$export_path,
) = @_;

ENV_say( 1, "Searching for obsolete items in $destination...");


debug_keeps()
if (ENV_is_debug());
my $keeps_ref = $KEEPS{$destination};
my @files_to_delete;
my @dirs_to_delete;
foreach my $spec_ref (SLURP_dir_tree( $export_path, 1))
{
my ($is_dir, $spec) = @{$spec_ref};

my $dir = ($is_dir) ? $spec : ENV_split_spec_d( $spec);



my $must_delete = 0;
if (grep( ENV_is_in_path( $dir, $_), @dirs_to_delete))
{
$must_delete = 1;
} elsif (!contains_dir( $spec, \@GBS_ALL_TARGETS))
{
if (!exists $keeps_ref->{$spec})
{
$must_delete = 1;
}
}
if ($must_delete)
{
if ($is_dir)
{
ENV_say( 1, "Must delete Dir '$spec'");
push @dirs_to_delete, $spec;
} else
{
ENV_say( 1, "Must delete File '$spec'");
push @files_to_delete, $spec;
}
} else
{
ENV_say( 1, "$destination: $spec")
if ($VERBOSE);
}
}

map { $_ = "$export_path/$_" } @dirs_to_delete;
map { $_ = "$export_path/$_" } @files_to_delete;
if (@files_to_delete)
{
my $nr_files_to_delete = @files_to_delete;
my $nr_files_deleted = unlink @files_to_delete;
ENV_say( 1, "$nr_files_deleted/$nr_files_to_delete file(s) deleted");
}
if (@dirs_to_delete)
{
my $nr_dirs_to_delete = @dirs_to_delete;
my $nr_dirs_deleted = 0;
foreach my $dir (reverse( @dirs_to_delete))
{
if (rmdir $dir)
{
$nr_dirs_deleted++;
} else
{
ENV_say( 1, "Unable to delete dir: '$dir'",
"- $!");
}
}
ENV_say( 1, "$nr_dirs_deleted/$nr_dirs_to_delete dir(s) deleted");
}
}




sub error(@)
{
my @msgs = @_;

if ($IGNORE_ERRORS)
{
ENV_error( @msgs);
} else
{
ENV_error_exit( @msgs);
}
}




sub contains_dir($$)
{
my ($spec,
$sub_dirs_ref) = @_;
my $contains = 0;

my $this_spec = "/$spec/";
foreach my $sub_dir (@{$sub_dirs_ref})
{
if (index( $this_spec, "/$sub_dir/") >= 0)
{
$contains = 1;
last;
}
}

return $contains;
}




sub debug_keeps()
{
foreach my $dest (qw( RESSUB EXPORT))
{
foreach my $keep (sort( keys( %{$KEEPS{$dest}})))
{
ENV_say( 0, "DBG: KEEPS: $dest => $keep => $KEEPS{$dest}->{$keep}");
}
}

foreach my $dest (qw( RESSUB EXPORT))
{
foreach my $dir (sort( keys( %{$OUT_DIRS{$dest}})))
{
ENV_say( 0, "DBG: OUT_DIRS: $dest => $dir => $OUT_DIRS{$dest}->{$dir}");
}
}
}


