#===============================================================
#
#   gbsjob.pl
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

print "*GBSJOB*\n" if ($ENV{GBSDEBUG_FILE});

use strict;
use integer;

use lib $ENV{GBS_SCRIPTS_PATH};
use glo::env;
use glo::args;
use glo::time;
use glo::shell;
use glo::slurp;
use glo::spit;
use glo::banner;
use glo::build;
use gbsres;
use bgnotifier;
use notify;




sub do_job();
sub finish($);
sub trap_sig($);
sub write_summary($$);
sub cpu_usage();
sub read_job_file();
sub get_command_args();








my $JOB_FILE;




my (
$SUBMODE,			    # Fixed: 'SUBWIN' or 'SUBMIT'
$JOBNAME,			    # Fixed
$TARGET,			    # Vary
$ACTION,			    # Vary: 'gen' or 'make' or audit_name
$DATE_TIME,			    # Fixed
$LOGFILE,			    # Vary: '-' or logfile
$MUST_DISPLAY_NOTIFY_MSG,	    # Fixed
$ROOT,			    # Fixed
$EXEC,			    # Vary. E.g.: gbssysgen begin bart superglo non_gbs copy_export --files=*:*.*
$DISPLAY,			    # Vary
$COMMAND_ARGS,		    # Vary. E.g.: --targets=test,mingw*
$COMMENT,			    # Vary
);
my $JOB;




my $TIME_START;
my $TIME_END;
my $TIME_DIFF;

my @TIMER_START;    # ($clock, $user, $system, $cuser, $csystem) in seconds
my @TIMER_END;	    # ($clock, $user, $system, $cuser, $csystem) in seconds
my @TIMER_DIFF;     # ($clock, $user, $system, $all)

my $UNAME;
{
my ($os, $host, @rest) = ENV_uname();
$UNAME = "$host ($os [@rest])";
}

my $TITLE = GBSRES_title();
my ($LICS, $LICN, $LICE) = BUILD_get_lic();






$| = 1;
my $RC = 0;

get_command_args();

read_job_file();
$JOB = "$JOBNAME ($TARGET / $ACTION)";

ENV_system( SHELL_xterm_follow_up(), 0, undef)
if ($SUBMODE eq 'SUBWIN');




$RC = do_job();




ENV_exit( $RC);




sub END
{
ENV_print_end_msg( 1);

}




sub do_job()
{
my $rc;

@TIMER_START = ( time, times );
$TIME_START = TIME_time2num( $TIMER_START[0]);
my $gbs_show_version = BUILD_get_show_version();

BANNER_print( 'JOB', [
[ '',         $TITLE ],
undef,
[ 'VERSION:', "$gbs_show_version" ],
[ 'LICENSE:', "$LICN ($LICS/$LICE)" ],
[ 'START:',   $TIME_START ],
[ 'JOB:',     $JOB ],
[ 'PID:',     $$ ],
[ 'HOST:',    $UNAME ],
[ 'ROOT:',    $ROOT ],
[ 'EXEC:',    $EXEC ],
[ 'LOGFILE:', $LOGFILE ],
[ 'DISPLAY:', $DISPLAY ],
[ 'ARGS:',    $COMMAND_ARGS ],
] );

if ($LOGFILE eq '-')
{
$ENV{GBS_EXEC_MODE} = 'FOREGROUND';
} else
{
$ENV{GBS_EXEC_MODE} = 'BACKGROUND';
BGNOTIFIER_start( $JOBNAME, $DATE_TIME, $TARGET, $ACTION, $LOGFILE);
}
$ENV{GBS_LOGFILE} = $LOGFILE;
$ENV{DISPLAY} = $DISPLAY;
$ENV{GBS_PID} = $$;
$ENV{GBS_EXEC_ID} = '';
$ENV{GBS_JOB_ARGS} = $COMMAND_ARGS;

use sigtrap 'handler', \&trap_sig, 'untrapped';

my $exec = $EXEC;
if (ENV_is_win32())
{
$exec =~ s/ /\$_/g;
} else
{
$exec =~ s/ /%20/g;
}

ENV_chdir( $ROOT);
my $command_ref = ENV_command_ref( 'gbsstart.pl', [ '!!', '-', "--target=$TARGET", "--exec=$exec" ]);
my $command = ENV_command_line( $command_ref);

$rc = ENV_system( $command, undef, undef);


finish( $rc);
unlink( $JOB_FILE);

return $rc;
}




sub finish($)
{
my ($rc) = @_;

ENV_say( 0,  "-");

my $state;
if ($rc == -1)
{
ENV_say( 0, "*******************",
"*** JOB KILLED  ***",
"*******************");
$state = 'KILLED';
} elsif ($rc == 0)
{
$state = 'NORMAL';
} else
{
$state = 'FAILED';
}

@TIMER_END = ( time, times );
$TIME_END = TIME_time2num( $TIMER_END[0]);
my $cpu_usage = cpu_usage();		# sets $TIME_DIFF and @TIMER_DIFF
my $display = ENV_getenv( 'DISPLAY');

if ($LOGFILE ne '-')
{
write_summary( $state, $rc);

BGNOTIFIER_end( $state, $rc);
}

if ($MUST_DISPLAY_NOTIFY_MSG)
{
NOTIFY_submit( $JOB, $rc, $LOGFILE, $state);
}

ENV_say( 0,  "=");
BANNER_print( 'JOB', [
[ '',         $TITLE ],
undef,
[ 'LICENSE:', "$LICN ($LICS/$LICE)" ],
[ 'START:',   $TIME_START ],
[ 'END:',     $TIME_END ],
[ 'ELAPSED:', $TIME_DIFF ],
[ 'JOB:',     $JOB ],
[ 'ROOT:',    $ROOT ],
[ 'EXEC:',    $EXEC ],
[ 'LOGFILE:', $LOGFILE ],
[ 'USAGE:',   $cpu_usage ],
[ 'STATUS:',  "$state ($rc)" ],
[ 'DISPLAY:', $display ],
[ 'ARGS:',    $COMMAND_ARGS ],
] );

if ($SUBMODE eq 'SUBWIN')
{
ENV_say( 1, "Hit <enter> key to exit this window: ");
<STDIN>;
}
}




sub trap_sig($)
{
my ($sig) = @_;

ENV_say( 1, "Signal: $sig");

finish( -1);
ENV_exit( 8);
}




sub write_summary($$)
{
my ($state,
$rc,
) = @_;




my @out = (
$JOBNAME,
$TARGET,
$UNAME,
$ROOT,
$LOGFILE,
$TIME_START,
$TIME_END,
$TIME_DIFF,
"@TIMER_DIFF",
$state,
$rc,
$ACTION,
$DATE_TIME,
$COMMAND_ARGS,
$COMMENT,
'<<EOF>>'
);
SPIT_file_nl( "$LOGFILE.sum", \@out);




my $gbs_root_path = ENV_perl_paths( $ROOT);
my $gbssum_dir = "$gbs_root_path/silo/.gbs/gbssum";
my $sumline = join( "\t", @out);
SPIT_file_nl( ">$gbssum_dir/gbssyssum.log", $sumline);	    #append

}




sub cpu_usage()
{
my $text;
no integer;

my @diff;
for ( my $i = 0; $i < 5; $i++)
{
$diff[$i] = $TIMER_END[$i] - $TIMER_START[$i];
}





my ($c, $pu, $ps, $cu, $cs) = @diff;
my $cpu_s = $ps + $cs;
my $cpu_u = $pu + $cu;
my $cpu_a = $cpu_s + $cpu_u;

$TIME_DIFF = TIME_deltatime2num( $c);
$text = sprintf "Clock: %s, System: %.2f, User: %.2f, Total: %.2f",
"$c ($TIME_DIFF)", $cpu_s, $cpu_u, $cpu_a;
@TIMER_DIFF = ($c, $cpu_s, $cpu_u, $cpu_a);

use integer;
return $text;
}




sub read_job_file()
{
(
$SUBMODE,
$JOBNAME,
$TARGET,
$ACTION,
$DATE_TIME,
$LOGFILE,
$MUST_DISPLAY_NOTIFY_MSG,
$ROOT,
$EXEC,
$DISPLAY,
$COMMAND_ARGS,
$COMMENT,
) = SLURP_file( $JOB_FILE);

ENV_fatal( "Invalid job-file '$JOB_FILE'")
if (!defined $EXEC);
ENV_fatal( "Invalid mode '$SUBMODE'")
if ($SUBMODE ne 'SUBMIT' && $SUBMODE ne 'SUBWIN');
}




sub get_command_args()
{


($JOB_FILE,
) = ARGS_get( 1, 'perl -w gbsjob.pl <job-file>', undef);
}


