#===============================================================
#
#   gbssilo.pl
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

print "*GBSSILO*\n" if ($ENV{GBSDEBUG_FILE});
use strict;
use integer;

use lib $ENV{GBS_SCRIPTS_PATH};
use glo::env;
use glo::genopt;
use glo::html;
use glo::spit;
use glo::slurp;
use glo::time;
use gbsenv;
use gbsglo;
use ssprop;
use gbssilo;
use gbshtml;




sub create_summaries();
sub gen_sum_file($$$);
sub create_target_audits_overview();




my $GBS_ROOT_PATH = GBSENV_gbs_root_path();
my $GBS_SILO_PATH = "$GBS_ROOT_PATH/silo";


my $GBS_SYSTEM_NAME = ENV_getenv( 'GBS_SYSTEM_NAME');
my $GBS_LOG_PATH = ENV_getenv( 'GBS_LOG_PATH');






my $RC = 0;
GBSENV_init();




my @genopts = (
[ 'ci',     'create_index',	     'bso',   1, "Create the HTML Index" ],
[ 'cs',     'create_summaries',  'bso',   1, "Create the HTML Summaries" ],
[ 'vi',     'view_index',	     'bso',   1, "View index in HTML browser" ],
);
GENOPT_set_optdefs( 'gbssilo', \@genopts,
'Generate silo index and show in HTML browser',
undef);
GENOPT_parse();
my $CREATE_SUMMARIES = GENOPT_get( 'create_summaries');
my $CREATE_INDEX = GENOPT_get( 'create_index');
my $VIEW_INDEX = GENOPT_get( 'view_index');




if ($CREATE_SUMMARIES)
{
create_summaries();
create_target_audits_overview()
}




GBSSILO_create_index()
if ($CREATE_INDEX);




GBSSILO_view_index()
if ($VIEW_INDEX);




ENV_exit( $RC);




sub END
{
ENV_print_end_msg( 1);
}




sub create_summaries()
{
my $sum_file_spec = "$GBS_SILO_PATH/.gbs/gbssum/gbssyssum.log";
if (-e $sum_file_spec)
{

my @gen_lines;
my @audit_lines;
my $line_nr = 0;
foreach my $sum_line (SLURP_file( $sum_file_spec))
{
$line_nr++;
next if ($sum_line eq '');

if (substr( $sum_line, 0, 11) eq 'gbssysaudit')
{
push @audit_lines, "$line_nr\t$sum_line";
} else	# gen and make
{
push @gen_lines, "$line_nr\t$sum_line";
}
}
if (@gen_lines)
{
my $gen_sum_file_spec = "$GBS_SILO_PATH/.gbs/gbssum/build.html";
gen_sum_file( $gen_sum_file_spec, 0, \@gen_lines)
}
if (@audit_lines)
{
my $audit_sum_file_spec = "$GBS_SILO_PATH/.gbs/gbssum/audit.html";
gen_sum_file( $audit_sum_file_spec, 1, \@audit_lines)
}
}
}




sub gen_sum_file($$$)
{
my ($file_spec,
$is_audit,			# 'Audit' or 'Generate/Make'
$lines_ref,
) = @_;

my $name = ($is_audit) ? 'Audit' : 'Generate/Make';
my @lines;




my $title = "$GBS_SYSTEM_NAME - Silo Overview: $name";

push @lines, GBSHTML_doc_start( $file_spec, $title, 1);
push @lines, GBSHTML_doc_top( [$title,  HTML_link( $GBS_LOG_PATH, HTML_bold( 'GBS Log Directory'))], ());




my @row_refs;




if ($is_audit)
{
push @lines, HTML_txt( 'Please be aware that only the HTML-summary files were saved. Not the files referred to.'),
HTML_br(),
HTML_bold_txt( 'The files referred to will always be the last audited file versions.'),
HTML_br();
}
push @lines, HTML_br(),
HTML_txt( 'Note that the following table is sortable by the items in the top row.');




foreach my $line (reverse( @{$lines_ref}))
{
my (
$line_nr,
$jobname,
$target,
$uname,
$root_path,
$logfile,			    # os format
$time_start,
$time_end,
$time_diff,
$timer_diff,
$state,			    # NORMAL, FAILED, KILLED
$rc,
$action,
$date_time,
$command_args,
$comments,
undef
) = split( /\t/, $line);


if (defined $comments)
{
$logfile = ENV_perl_paths( $logfile);
$comments = ''
if ($comments eq '<<EOF>>');

my $logfile_html;
if (-e $logfile)
{
$logfile_html = HTML_link( $logfile, 'VIEW');
} else
{
my ($path, $file) = ENV_split_spec_df( $logfile);
my $moved_logfile = "$GBS_LOG_PATH/$file";
$logfile_html = (-e $moved_logfile) ? HTML_link( $moved_logfile, 'VIEW') : 'Missing';
}
my $state_html = HTML_div( {class => $state}, $state);
$date_time = TIME_unpack( $date_time);
$date_time =~ s/ /&nbsp;/;
if ($is_audit)
{
my ($audit) = $logfile =~ /gbssysaudit_${target}_(.*)_\d/;
$audit = '*UNK*'
if (!defined $audit);
my $htmlfile = "$GBS_SILO_PATH/audits/" . ENV_split_spec_n( $logfile) . "_sum.html";
my $htmlfile_html = (-e $htmlfile) ? HTML_link( $htmlfile, 'VIEW') : 'Missing';
push @row_refs, [ $jobname, $audit, $target, $date_time, $state_html, $rc, $logfile_html, $htmlfile_html, $command_args, $comments ];
} else
{
push @row_refs, [ $jobname, $target, $date_time, $state_html, $rc, $logfile_html, $command_args, $comments ];
}
} else
{
$line =~ s/^\d+\t//;
ENV_warn( "Invalid log line ($line_nr)", $line);
}
}

push @lines, HTML_br();
if ($is_audit)
{
push @lines, HTML_table( [ qw( - - - - -C -R -C -C - - ) ],
[ qw( Job Audit Target DateTime State RC LogFile HTML<BR>Sum Comand-Args Comments) ],
\@row_refs, border => 1, class => 'sortable border1');
} else
{
push @lines, HTML_table( [ qw( - - - -C -R -C - - ) ],
[ qw( Job Target DateTime State RC LogFile Comand-Args Comments) ],
\@row_refs, border => 1, class => 'sortable border1');
}




push @lines, GBSHTML_doc_bottom( $title, undef);
push @lines, GBSHTML_doc_end();

SPIT_file_nl( $file_spec, \@lines);
ENV_say( 1, "Created Silo $name");
}




sub create_target_audits_overview()
{
my $path = "$GBS_SILO_PATH/target_audits_overview";
ENV_mkdir( $path);
my $file_spec = "$path/index.html";
my @lines;




my $title = "$GBS_SYSTEM_NAME - Targets and Audits Overview";

push @lines, GBSHTML_doc_start( $file_spec, $title, 1);
push @lines, GBSHTML_doc_top( [ $title, HTML_link( '../index.html', HTML_bold( 'Silo Home')) ], ());








push @lines, HTML_br(),
HTML_txt( 'Note that the following table is sortable by the items in the top row.');




my @row_refs;
my @gbs_all_targets = ENV_getenv( 'GBS_ALL_TARGETS');
my %gbs_all_targets = map { $_ => 1 } @gbs_all_targets;
my $nr_targets = @gbs_all_targets;
my @gbs_all_audits =  ENV_getenv( 'GBS_ALL_AUDITS');
my %gbs_all_audits = map { $_ => 1 } @gbs_all_audits;
my $nr_audits = @gbs_all_audits;

foreach my $subsys (GBSGLO_subsystems())
{
my $sstype = SSPROP_get_type( undef, $subsys);

if ($sstype eq 'GBS')
{



foreach my $component (GBSGLO_components( $subsys))
{

my $component_path = "$GBS_ROOT_PATH/dev/$subsys/comp/$component";
foreach my $target (@gbs_all_targets)
{
my $ss_target = (GBSGLO_subsystem_does_build( $subsys, $target)) ? $target : "$target*";
my @found_audits;
foreach my $audit (@gbs_all_audits)
{
if (-e "$component_path/aud/$audit/$target")
{
push @found_audits, $audit;
}
}
if (@found_audits)
{
foreach my $audit (@found_audits)
{
my $ss_audit = (GBSGLO_subsystem_does_audit( $subsys, $audit, $target)) ? $audit : "$audit*";
push @row_refs, [ $subsys, $sstype, $ss_target, $ss_audit, $component, $ss_target, $audit ];
}
} else
{
push @row_refs, [ $subsys, $sstype, $ss_target, '', $component, $ss_target, '' ]
if ($ss_target ne "$target*");
}
}
}
} else
{



my $nr_ss_targets = 0;
foreach my $ss_target( @gbs_all_targets)
{
if (GBSGLO_subsystem_does_build( $subsys, $ss_target))
{
my $nr_ss_audits = 0;
foreach my $ss_audit (@gbs_all_audits)
{
if (GBSGLO_subsystem_does_audit( $subsys, $ss_audit, $ss_target))
{
push @row_refs, [ $subsys, $sstype, $ss_target, $ss_audit, '', '', ''];
$nr_ss_audits++;
$nr_ss_targets++;
}
}
if ($nr_ss_audits == 0)
{
push @row_refs, [ $subsys, $sstype, $ss_target, '', '', '', '' ];
$nr_ss_targets++;
}
}
}
push @row_refs, [ $subsys, $sstype, '', '', '', '', '' ]
if ($nr_ss_targets == 0);

}
}

push @lines, HTML_table( [ qw( - - - - - - - ) ],
[ qw( SubSys SSTYPE Target Audit Component Target Audit) ],
\@row_refs, border => 1, class => 'sortable border1');




push @lines, GBSHTML_doc_bottom( $title, undef);
push @lines, GBSHTML_doc_end();

SPIT_file_nl( $file_spec, \@lines);
ENV_say( 1, "Created Silo Targets and Audits Overview");
}


