#===============================================================
#
#   gbssys.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package gbssys;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
GBSSYS_init
GBSSYS_submit
GBSSYS_execute
GBSSYS_print_errors
);

use strict;
use integer;

use glo::env;
use glo::genopt;
use glo::slurp;
use glo::banner;
use gbsenv;
use gbsglo;
use gbsjob;
use exec;
use validate;
use cleanup;
use steps;
use sws;
use audit;




sub GBSSYS_init($$$$$$$$);
sub GBSSYS_submit($$);
sub GBSSYS_execute($$);
sub GBSSYS_print_errors();

sub do_sub_command($);
sub do_sub_build($$);
sub do_sub_audit($$);
sub do_audit_sum($);




my $GBS_ROOT_PATH = GBSENV_gbs_root_path();




my $JOBTYPE;		    # 'gen', 'make' or 'audit'

my $IGNORE_ERRORS;	    # bool
my $TARGETS_REF;
my $AUDITS_REF;
my $FILES_REF;
my $FOREGROUND;		    # bool
my $SUBMIT_REF,		    # [ $DELAY # bool , $MAIL  # bool, $NOTIFY # bool, $COMMENT # string]
my $WAIT;		    # bool




my @ARG_TARGETS;
my @ARG_AUDITS;








my $GBS_TARGET;
my $FOR_GEN = 0;
my $FOR_MAKE = 0;
my $FOR_AUDIT = 0;
my $ACTION_TEXT;
my $GBS_AUDIT;

my $AUDIT_COUNT;





sub GBSSYS_init($$$$$$$$)
{
($JOBTYPE,			# 'gen', 'make' or 'audit'
$IGNORE_ERRORS,
$TARGETS_REF,
$AUDITS_REF,		# may be undef
$FILES_REF,
$FOREGROUND,
$SUBMIT_REF,		# [ $DELAY, $MAIL, $NOTIFY, $COMMENT ]
$WAIT,
) = @_;

}





sub GBSSYS_submit($$)
{
my ($step_names_ref,
$opts_ref,	# selection of: steps targets audits files delay
) = @_;
my $rc = 0;




@ARG_TARGETS = grep( $_ ne '-', VALIDATE_targets( @{$TARGETS_REF}));
my %selected_targets = map { $_ => 1 } @ARG_TARGETS;




my @selected_audits;
if ($JOBTYPE eq 'audit')
{
@ARG_AUDITS  = grep( $_ ne '-', VALIDATE_audits( @{$AUDITS_REF}));
@selected_audits = @ARG_AUDITS;
}




my @options = @{$opts_ref};
if (@{$FILES_REF})
{
my $gbs_component = ENV_getenv( 'GBS_COMPONENT');
my @files;
foreach my $cfs (@{$FILES_REF})
{
my ($comp) = $cfs =~ /(.*):/;
if ($comp)
{
if ($comp eq '.')
{
ENV_error_exit( "No current Component (--files=$cfs)")
if ($gbs_component eq '');
$cfs = $gbs_component . substr( $cfs, 1);
ENV_say( 1, "--files set to '$cfs'");
}
} else
{
$cfs = "*:$cfs";
ENV_say( 1, "--files set to '$cfs'");
}
push @files, $cfs;
}
my $files = '--files=' . join( ',', @files);
push @options, $files;
}




my $jobname ="gbssys$JOBTYPE";
if ($FOREGROUND)
{
GBSJOB_init( $jobname, undef);
} else
{
GBSJOB_init( $jobname, $SUBMIT_REF);
}




my $count = 0;
my @selectable_targets = STEPS_analyze_steps( $JOBTYPE, $step_names_ref);
if ($JOBTYPE eq 'audit')
{



foreach my $audit (@selected_audits)
{

my @audit_options = GENOPT_makeopt( audits => $audit);
my $nr_selected = 0;
foreach my $target (@selectable_targets)
{

if (exists $selected_targets{$target})
{
my @full_steps = STEPS_select_steps( $target, $audit);
if (@full_steps)
{
my $exec = "@full_steps @options @audit_options";

GBSJOB_sub( $target, $audit, $exec);
$count++;
$nr_selected++;
} else
{

}
}
}
if ($nr_selected == 0)
{
ENV_say( 1, "Audit '$audit' not selected");
}
}
} else
{
foreach my $target (@selectable_targets)
{

if (exists $selected_targets{$target})
{



my @full_steps = STEPS_select_steps( $target, '');
if (@full_steps)
{
my $exec = "@full_steps @options";

GBSJOB_sub( $target, $JOBTYPE, $exec);
$count++;
} else
{

}
} else
{
ENV_say( 1, "Target '$target' not selected");
}
}
}

if ($count > 0)
{
ENV_say( 1, "Submitting $count jobs");
GBSJOB_run();
$rc = GBSJOB_wait()
if ($WAIT);
} else
{
ENV_warn( "Nothing to submit");
}

return $rc;
}






sub GBSSYS_execute($$)
{
my ($steps_ref,
$options_ref,
) = @_;
my $rc = 0;

my @steps = @{$steps_ref};
my @command_options = @{$options_ref};  # --i, --mm, --export, --summary, --audit=<audit>, etc




$GBS_TARGET = ENV_getenv( 'GBS_TARGET');
$AUDIT_COUNT = 1;
if ($JOBTYPE eq 'gen')
{
$FOR_GEN = 1;
$ACTION_TEXT = 'Generating';
} elsif ($JOBTYPE eq 'make')
{
$FOR_MAKE = 1;
$ACTION_TEXT = 'Making';
} else  #($JOBTYPE eq 'audit')
{
$FOR_AUDIT = 1;
$GBS_AUDIT = ENV_getenv( 'GBS_AUDIT');
$ACTION_TEXT = "Auditing $GBS_AUDIT";
}
$ENV{GBS_IGNORE_ERRORS} = $IGNORE_ERRORS;




BANNER_print( 'SYSTEM', "ALL_STEPS: @steps");

while (@steps)
{



my $step_name = shift @steps;
my ($subsys, $command, $must_force_execute) = STEPS_get_command( $step_name);

my $must_execute = ($IGNORE_ERRORS || $must_force_execute);
if ($rc == 0 || $IGNORE_ERRORS || $must_force_execute)
{
my $banner_name = ($must_force_execute) ? 'EXPLICIT STEP' : 'STEP';
my $this_rc;
if ($subsys)
{



BANNER_print( 'SUBSYS', [
"$banner_name: $step_name - SubSystem: $subsys",
"Silent switch to SubSystem: $subsys" ]);
SWS_set( $subsys, 0);
my $subsys_path = "$GBS_ROOT_PATH/dev/$subsys";
ENV_chdir( $subsys_path);




if ($FOR_AUDIT)
{
$this_rc = do_sub_audit( $subsys, \@command_options);
} else
{
$this_rc = do_sub_build( $subsys, \@command_options);
}
} else
{
ENV_chdir( $GBS_ROOT_PATH);
BANNER_print( 'SUBSYS', "$banner_name: $step_name - Command: $command");
$this_rc = do_sub_command( $command);
}
$rc = $this_rc if ($this_rc > $rc);

if ($this_rc == 0)
{
BANNER_print( 'NOTE', "DONE: SubSystem $step_name - TARGET=$GBS_TARGET");
} else
{
BANNER_print( 'FAIL', "FAILED: SubSystem $step_name - TARGET=$GBS_TARGET - RC=$this_rc");
}
if ($this_rc != 0 && !$IGNORE_ERRORS)
{
if (@steps)
{
BANNER_print( 'NOTE', 'Remaining STEPs will be skipped');
} else
{
BANNER_print( 'NOTE', 'No Remaining STEPs');
}
last;
}
}

}




if ($FOR_AUDIT)
{
my $this_rc = do_audit_sum( \@command_options);
$rc = $this_rc if ($this_rc > $rc);
}

return $rc;
}





sub do_sub_build($$)
{
my ($subsys,
$gbs_build_opts_ref,
) = @_;
my $rc = 0;




if ($FOR_GEN)
{
BANNER_print( 'NOTE', "Implicit: Cleanup the bld directories for SubSys: $subsys");
CLEANUP_subsys_bld( $subsys, $GBS_TARGET);
}




if (GBSGLO_subsystem_is_full_gbs( $subsys))
{



my @gbs_build_opts = @{$gbs_build_opts_ref};
ENV_say( 1, "$ACTION_TEXT Full GBS SubSystem: $subsys (@gbs_build_opts)...");
$rc = EXEC_run( "gbs$JOBTYPE.pl", [ @{$FILES_REF}, @gbs_build_opts ]);
} else
{



ENV_say( 1, "$ACTION_TEXT Non-Full GBS SubSystem: $subsys...");




$rc = EXEC_run( 'gbssub.pl', [ $JOBTYPE, $IGNORE_ERRORS ]);
}

return $rc;
}




sub do_sub_audit($$)
{
my ($subsys,
$gbsaudit_opts_ref,
) = @_;
my $rc = 0;




if (GBSGLO_subsystem_is_full_gbs( $subsys))
{



my @gbsaudit_opts = @{$gbsaudit_opts_ref};
ENV_say( 1, "$ACTION_TEXT Full GBS SubSystem: $subsys (@gbsaudit_opts)...");
my @args = ( @{$FILES_REF}, @gbsaudit_opts, "--call=$AUDIT_COUNT");

$ENV{GBS_AUDIT_FILE} = 'NONE';
$rc = EXEC_run( "gbsaudit.pl", \@args);
$AUDIT_COUNT++;
} else
{



ENV_say( 1, "$ACTION_TEXT Non-Full GBS SubSystem: $subsys...");
$rc = EXEC_run( 'gbssub.pl', [ $JOBTYPE, $IGNORE_ERRORS ]);
}

return $rc;
}




sub do_audit_sum($)
{
my ($gbsaudit_opts_ref) = @_;
my $rc = 0;





if ($AUDIT_COUNT > 1)
{
ENV_chdir( $GBS_ROOT_PATH);
my @gbsaudit_opts = @{$gbsaudit_opts_ref};
BANNER_print( 'SUBSYS', "IMPLICIT STEP: Create Audit Summary (@gbsaudit_opts)" );

AUDIT_set_envs( $GBS_AUDIT);
my @args = ('summary', @gbsaudit_opts, "--call=-$AUDIT_COUNT");
$rc = EXEC_batch_run( ENV_command_ref( 'gbsaudit.pl', \@args));
if ($rc == 0)
{
BANNER_print( 'NOTE', "AUDIT SUMMARY: TARGET=$GBS_TARGET");
} else
{
BANNER_print( 'FAIL', "AUDIT SUMMARY FAILED: TARGET=$GBS_TARGET");
}
}
$AUDIT_COUNT = 1;

return $rc;
}




sub do_sub_command($)
{
my ($command_line) = @_;
my $rc = 0;

BANNER_print( 'NOTE', "Executing Command: $command_line...");
my ($command_file, @args) = split( ' ', $command_line);
$rc = EXEC_batch_run( ENV_command_ref( $command_file, \@args));

return $rc;
}





sub GBSSYS_print_errors()
{
my $gbs_logfile = ENV_getenv( 'GBS_LOGFILE');
if ($gbs_logfile && $gbs_logfile ne '-')
{
my $fail_file_spec = "$gbs_logfile.err";
if (-e $fail_file_spec)
{
BANNER_print( 'FAIL', [
'Error Log:',
'',
SLURP_file( $fail_file_spec) ]);
}
}
}

return 1;


