#===============================================================
#
#   file.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package glo::file;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
FILE_mtime
FILE_recurse_tree
FILE_parse_tree
FILE_replicate_tree
FILE_del_tree
FILE_copy_template
FILE_copy_translate
FILE_replicate
FILE_move
FILE_mkdir
FILE_mkpath
FILE_del_dirs
FILE_del_files
FILE_lock_ex_wait
FILE_lock_sh_wait
FILE_unlock
FILE_which
FILE_which_path
);

use strict;
use integer;

use Fcntl ":flock";  #import LOCK_* constants
use glo::env;
use glo::slurp;
use glo::spit;




sub FILE_mtime($);
sub FILE_parse_tree($$$$);
sub FILE_recurse_tree($$$);
sub FILE_replicate_tree($$$$);
sub FILE_del_tree($$$);
sub FILE_copy_template($$;@);
sub FILE_copy_translate($$;@);
sub FILE_replicate($$);
sub FILE_move($$$);
sub FILE_mkdir($;$);
sub FILE_mkpath($;$);
sub FILE_del_dirs($);
sub FILE_del_files($);
sub FILE_lock_ex_wait($);
sub FILE_lock_sh_wait($);
sub FILE_unlock($);
sub FILE_which($$);
sub FILE_which_path($@);

sub parse_tree($$);
sub replicate_tree($$);
sub copy_file($$);
sub recurse_tree($);
sub del_tree($);
sub init_recurse($$$);
sub must_ignore($);
sub exit_recurse();
sub get_command_types();
sub match_type($);














































my $IS_WIN32 = (ENV_is_win32());

my $FUNC_REF;		# $state = $func_ref->( $level, $is_dir, $dir_spec, $name);
my @IGNORES;		#
my $VERBOSE_LEVEL;	# 0 = nothing, 1 = ignore, 2 = directories, 3 = files, 4 = dir + files, 5 = max
my $RECURSING;		# BOOL
my $NR_DIRS_CREATED;
my $NR_FILES_COPIED;
my $NR_DIRS_DELETED;
my $NR_FILES_DELETED;
my $NR_TOTAL_SCANNED;

my $DEFAULT_MODE = 0751;    # u=rwx g=rx o=x

my %LOCKS;


my @COMMAND_TYPES;  	# do not initialize!	e.g. .exe, .com , .bat, .sh, etc
my @PATH;  	    	# do not initialize!	from EnvVar PATH
my %COMMANDS;






sub FILE_mtime($)
{
my ($file_or_dir_spec) = @_;
my $mtime;



$mtime = (stat( $file_or_dir_spec))[9];	# $mtime
ENV_fatal( "'stat' of file '$file_or_dir_spec' failed")
if (!defined $mtime);

return $mtime;
}




sub FILE_parse_tree($$$$)
{
my ($dir_spec,
$func_ref,	    # $state = $func_ref->( $level, $is_dir, $dir_spec, $name);

$ignore_ref,
$verbose_level,	    # 0 = nothing, 1 = ignore, 2 = directories, 3 = files, 4 = dir + files, 5 = max
) = @_;
my $last_state;

init_recurse( $func_ref, $ignore_ref, $verbose_level);

if (!must_ignore( $dir_spec))
{
parse_tree( 0, $dir_spec);
}

exit_recurse();
}





sub parse_tree($$)
{
my ($level,
$dir_spec,
) = @_;
my $state = 0;		# 0 == continue, 1 = stop this_dir (next), 2 = total stop

$level++;
foreach my $entry (SLURP_dir_all( $dir_spec, 0))
{
if (!must_ignore( $entry))
{
my $entry_spec = "$dir_spec/$entry";
if (-d $entry_spec)
{
ENV_say( 1, " From  Dir $entry_spec...")
if ($VERBOSE_LEVEL >= 2);
$state = $FUNC_REF->( $level, 1, $dir_spec, $entry);
$state = parse_tree( $level, $entry_spec)
if ($state == 0);
} elsif (-f _)
{
ENV_say( 1, " From File $entry_spec")
if ($VERBOSE_LEVEL >= 3);
$state = $FUNC_REF->( $level, 0, $dir_spec, $entry);
} elsif (-l _)
{
ENV_error( "Symbolic links not supported",
"'$entry_spec'");
} else
{
ENV_error( "Unknown Fil/Dir type",
"'$entry_spec'");
}
last
if ($state > 0);
}
}
$state = 0
if ($state == 1);

return $state;
}




sub FILE_recurse_tree($$$)
{
my ($dir_spec,
$ignore_ref,
$verbose_level,	    # 0 = nothing, 1 = ignore, 2 = directories, 3 = files, 4 = dir + files, 5 = max
) = @_;
my @entry_refs;	    # [ F or D, $full_spec ]

init_recurse( undef, $ignore_ref, $verbose_level);

if (!must_ignore( $dir_spec))
{
@entry_refs = recurse_tree( $dir_spec);
}

exit_recurse();

return @entry_refs;
}





sub recurse_tree($)
{
my ($dir_spec,
) = @_;
my (@file_refs, @dir_refs);

foreach my $entry (SLURP_dir_all( $dir_spec, 0))
{
if (!must_ignore( $entry))
{
my $entry_spec = "$dir_spec/$entry";
if (-d $entry_spec)
{
ENV_say( 1, " From Dir $entry_spec...")
if ($VERBOSE_LEVEL >= 2);
push @dir_refs, [ 'D', $entry_spec ];
push @dir_refs, recurse_tree( $entry_spec);
} elsif (-f _)
{
ENV_say( 1, " From File $entry_spec")
if ($VERBOSE_LEVEL >= 3);
push @file_refs, [ 'F', $entry_spec ];
} elsif (-l _)
{
ENV_error( "Symbolic links not supported",
"'$entry_spec'");
} else
{
ENV_error( "Unknown File/Dir type",
"'$entry_spec'");
}
}
}

return (@file_refs, @dir_refs);
}




sub FILE_replicate_tree($$$$)
{
my ($from_dir_spec,
$to_dir_spec,
$ignore_ref,
$verbose_level,	    # 0 = nothing, 1 = ignore, 2 = directories, 3 = files, 4 = dir + files, 5 = max
) = @_;


init_recurse( undef, $ignore_ref, $verbose_level);




FILE_mkdir( $to_dir_spec)
if (! -d $to_dir_spec);

replicate_tree( $from_dir_spec, $to_dir_spec); # recurse

exit_recurse();

return ($NR_DIRS_CREATED, $NR_FILES_COPIED, $NR_DIRS_DELETED, $NR_FILES_DELETED);
}





sub replicate_tree($$)
{
my ($from_dir_spec,
$to_dir_spec,
) = @_;




my %to_items = map { $_ => 1 } SLURP_dir_all( $to_dir_spec, 0);




foreach my $item (SLURP_dir_all( $from_dir_spec, 0))
{
if (!must_ignore( $item))
{
my $from_spec = "$from_dir_spec/$item";
my $to_spec = "$to_dir_spec/$item";
if (-d $from_spec)
{



ENV_say( 1, " From  Dir $from_spec...")
if ($VERBOSE_LEVEL >= 2);
if ($to_items{$item})
{
if (-f $to_spec)
{
FILE_del_files( $to_spec);
FILE_mkdir( $to_spec);
} else
{
;	# dir exists
}
delete $to_items{$item};
} else
{
FILE_mkdir( $to_spec);
}
replicate_tree( $from_spec, $to_spec); # recurse
} elsif (-f _)
{



ENV_say( 1, " From File $from_spec")
if ($VERBOSE_LEVEL >= 3);
if ($to_items{$item})
{
if (-f $to_spec)
{
FILE_replicate( $from_spec, $to_spec);
} else	# item is dir
{
del_tree( $to_spec);
FILE_del_dirs( $to_spec);
copy_file( $from_spec, $to_spec);
}
delete $to_items{$item};
} else
{
copy_file( $from_spec, $to_spec);
}
} elsif (-l _)
{
ENV_error( "Symbolic links not supported",
"'$from_spec'");
} else
{
ENV_error( "Unknown Fil/Dir type",
"'$from_spec'");
}
}
}




foreach my $item (keys %to_items)
{
my $to_spec = "$to_dir_spec/$item";
if (-d $to_spec)
{
del_tree( $to_spec);
FILE_del_dirs( $to_spec);
} else
{
FILE_del_files( $to_spec);
}
}
}







sub FILE_del_tree($$$)
{
my ($from_dir_spec,
$top_exclude_re_ref,
$verbose_level,		# 0 = nothing, 1 = ignore, 2 = directories, 3 = files, 4 = dir + files, 5 = max
) = @_;


init_recurse( undef, undef, $verbose_level);

foreach my $item (SLURP_dir_all( $from_dir_spec, 1))
{
if (!grep( $item =~ $_, @{$top_exclude_re_ref}))
{
my $from_spec = "$from_dir_spec/$item";
$NR_TOTAL_SCANNED++;
if (-d $from_spec)
{
if (!-l $from_spec)    # do not follow symbolic links
{
del_tree( $from_spec);
}

FILE_del_dirs( $from_spec);
} else
{

FILE_del_files( $from_spec);
}
} else
{

}
}

exit_recurse();

return (wantarray) ? ($NR_DIRS_DELETED, $NR_FILES_DELETED, $NR_TOTAL_SCANNED) : [ $NR_DIRS_DELETED, $NR_FILES_DELETED, $NR_TOTAL_SCANNED ];
}





sub del_tree($)
{
my ($from_dir_spec) = @_;

foreach my $item (SLURP_dir_all( $from_dir_spec, 1))
{
my $from_spec = "$from_dir_spec/$item";
$NR_TOTAL_SCANNED++;
if (-d $from_spec)
{
if (!-l $from_spec)    # do not follow symbolic links
{
del_tree( $from_spec);
}

FILE_del_dirs( $from_spec);
} else
{

FILE_del_files( $from_spec);
}
}
}




sub init_recurse($$$)
{
my ($func_ref,
$ignore_ref,
$verbose_level,
) = @_;

$FUNC_REF = $func_ref;
@IGNORES = ($ignore_ref) ? @{$ignore_ref} : ();
$VERBOSE_LEVEL = $verbose_level;

$RECURSING = 1;
$NR_DIRS_CREATED = 0;
$NR_FILES_COPIED = 0;
$NR_DIRS_DELETED = 0;
$NR_FILES_DELETED = 0;
$NR_TOTAL_SCANNED = 0;
}




sub must_ignore($)
{
my ($file_dir_spec) = @_;
my $ignore = 0;

if (@IGNORES)
{
my ($file) = ENV_split_spec_f( $file_dir_spec);
if (grep( $file =~ $_, @IGNORES))
{
ENV_say( 1, "Ignored $file_dir_spec")
if ($VERBOSE_LEVEL >= 1);
$ignore = 1;
}
}

return $ignore;
}




sub exit_recurse()
{
$FUNC_REF = undef;
@IGNORES = ();
$VERBOSE_LEVEL = undef;

$RECURSING = 0;
}




sub FILE_replicate($$)
{
my ($in_file,
$out_file,
) = @_;
my $replicated = 0;    	    # 0 = not replicated (already ok), 1 = replicated




if (!ENV_file_is_same( $in_file, $out_file))
{
ENV_copy_file( $in_file, $out_file, 1);   # $exit_on_error = 1
$replicated = 1;
}

return $replicated;
}




sub copy_file($$)
{
my ($in_file,
$out_file,
) = @_;

ENV_copy_file( $in_file, $out_file, 1);   # $exit_on_error = 1

$NR_FILES_COPIED++
if ($RECURSING);
}





sub FILE_move($$$)
{
my ($old_spec,
$new_spec,
$verbose_level,		# 0 = nothing, 1 = ignore, 2 = directories, 3 = files, 4 = dir + files, 5 = max
) = @_;

ENV_fatal( "MOVE: file/dir '$old_spec' does not exist")
if (!- $old_spec);




if (!rename( $old_spec, $new_spec))
{



my $is_copied = 0;
my $is_deleted = 0;
if (-f $old_spec)
{
$is_copied = FILE_replicate( $old_spec, $new_spec);
if ($is_copied)
{
$is_deleted = FILE_del_files( $old_spec);
}
} else
{
$is_copied = FILE_replicate_tree( $old_spec, $new_spec, undef, $verbose_level);
if ($is_copied)
{
$is_deleted = FILE_del_tree( $old_spec, undef, $verbose_level);
}
}
if (!$is_copied || !$is_deleted)
{
ENV_fatal( "Cannot move file/dir: '$old_spec'",
"                   -> '$new_spec'",
"- $!");
}
}
}




sub FILE_mkdir($;$)
{
my ($dir_spec_or_ref,
$mode,
) = @_;
my $count = 0;

$mode = $DEFAULT_MODE unless (defined $mode);

my $dir_ref = (ref $dir_spec_or_ref) ? $dir_spec_or_ref : [ $dir_spec_or_ref ];
foreach my $dir_spec (@{$dir_ref})
{
if (!-d $dir_spec)
{
ENV_mkdir( $dir_spec, $mode);
$count++;
}
}

$NR_DIRS_CREATED += $count
if ($RECURSING);

return $count;
}




sub FILE_mkpath($;$)
{
my ($path_or_ref,
$mode,
) = @_;
my $count = 0;

my $path_ref = (ref $path_or_ref) ? $path_or_ref : [ $path_or_ref ];
$mode = $DEFAULT_MODE unless (defined $mode);

foreach my $path (@{$path_ref})
{
next if (-d $path);

my ($parent) = $path =~ m!(.*)/!;
if (defined $parent)
{

$count += FILE_mkpath( $parent, $mode)
if (! -d $parent);

}
FILE_mkdir( $path, $mode);
$count++;
}

return $count;
}




sub FILE_del_dirs($)
{
my ($dir_spec_or_ref) = @_;
my $count = 0;

my @dirs = (ref $dir_spec_or_ref) ? @{$dir_spec_or_ref} : ($dir_spec_or_ref);
foreach my $dir_spec (@dirs)
{
if (rmdir $dir_spec)
{
$count++;
} else
{
chmod 0777, $dir_spec;
if (rmdir $dir_spec)
{
$count++;
} else
{
ENV_warn( "Unable to remove dir '$dir_spec'", "- $!");
}
}
}

$NR_DIRS_DELETED += $count
if ($RECURSING);

return $count;
}




sub FILE_del_files($)
{
my ($file_or_ref) = @_;
my $count = 0;

my @files = (ref $file_or_ref) ? @{$file_or_ref} : ($file_or_ref);

$count = unlink( @files);
if ($count != @files)
{
$count = 0;
foreach my $file (@files)
{
chmod 0777, $file;
if (unlink $file)
{
$count++;
} else
{
ENV_warn( "Unable to remove file '$file'", "- $!");
}
}
}

$NR_FILES_DELETED += $count
if ($RECURSING);

return $count;
}




sub FILE_copy_template($$;@)
{
my ($from_spec,	# template-file
$to_spec,	# out-file
@translate,	# ($token, $value) or ($token, [@insert_lines]), ...
) = @_;

my @lines = SLURP_file( $from_spec);

push @translate, ( '_FILE_', $to_spec );
while (@translate)
{



my $token = shift @translate;
my $value = shift @translate;

if (ref $value)
{




my $nr_value_lines = @{$value};
my $i = 0;
do
{
if ($lines[$i] eq "%$token%")
{
splice( @lines, $i, 1, @{$value});
$i += $nr_value_lines;
} else
{
$i++;
}
} while ( $i < @lines);
} else
{



map { s/%$token%/$value/g } @lines;
}
}

SPIT_file_nl( $to_spec, [ @lines ]);
}




sub FILE_copy_translate($$;@)
{
my ($from_spec,	# in-file
$to_spec,	# out-file
@translate,	# ($token, $value) or ($token, [@insert_lines]), ...
) = @_;

my @lines = SLURP_file( $from_spec);

push @translate, ( $from_spec, $to_spec );	    # substitute the file_spec
while (@translate)
{



my $token = shift @translate;
my $value = shift @translate;

my $q_token = quotemeta $token;
map { s/[\W]($q_token)[\W]/$value/g } @lines;
}

SPIT_file_nl( $to_spec, [ @lines ]);
}




sub FILE_lock_ex_wait($)
{
my ($file_spec) = @_;
my $fh;

my ($dir_spec, $name, $type) = ENV_split_spec_dnt( $file_spec);
$file_spec = "$dir_spec/.$name.lck"
if ($type eq '.lck' && substr( $name, 0, 1) ne '.');

open( $fh, ">>$file_spec") || ENV_fatal( "Cannot open/append '$file_spec'", "- $!");
flock( $fh, LOCK_EX) || ENV_warn( "Cannot flock/2 (EX) '$file_spec'", "- $!");
$LOCKS{$fh} = $file_spec;

return $fh;
}




sub FILE_lock_sh_wait($)
{
my ($file_spec) = @_;
my $fh;

my ($dir_spec, $name, $type) = ENV_split_spec_dnt( $file_spec);
$file_spec = "$dir_spec/.$name.lck"
if ($type eq '.lck' && substr( $name, 0, 1) ne '.');

open( $fh, ">>$file_spec") || ENV_fatal( "Cannot open/append '$file_spec'", "- $!");
flock( $fh, LOCK_SH) || ENV_warn( "Cannot flock/1 (SH) '$file_spec'", "- $!");
$LOCKS{$fh} = $file_spec;

return $fh;
}




sub FILE_unlock($)
{
my ($fh) = @_;

my $file_spec = $LOCKS{$fh};

flock( $fh, LOCK_UN) || ENV_warn( "Cannot flock/8 (UN) '$file_spec'", "- $!");
close( $fh) || ENV_fatal( "Cannot close '$file_spec'", "- $!");
delete( $LOCKS{$fh});
}




sub FILE_which($$)
{
my ($command,
$error_if_not_found) = @_;
my $file_spec;				# undef if not found

$file_spec = $COMMANDS{ $command};
if (!defined $file_spec)
{
get_command_types()
if (!@COMMAND_TYPES);
if (ENV_is_abs_path( $command))
{
$file_spec = match_type( $command);
ENV_error( "'$command' Not found")
if (!defined $file_spec && $error_if_not_found);
} else
{
my $cwd = ENV_cwd();
if ($command =~ m!/!)
{
$file_spec = match_type( "$cwd/$command");
ENV_error( "'$command' not found in CWD ($cwd)")
if (!defined $file_spec && $error_if_not_found);
} else
{
if (!@PATH)
{
foreach my $path (split( ($IS_WIN32) ? ';' : ':', $ENV{PATH}))
{
next if ($path eq '');
if (ENV_is_abs_path( $path))
{
if (-e $path)
{
if (-d $path)
{
push @PATH, ENV_perl_path_nospace( "\"$path\"");
} else
{
ENV_warn( "PATH contains non-Directory entry: '$path' - Skipped");
}
} else
{
ENV_warn( "PATH contains non-Existing entry: '$path' - Skipped");
}
} else
{
ENV_warn( "PATH contains non-AbsPath entry: '$path' - Skipped");
}
}
}

foreach my $dir_spec (@PATH)
{

my $this_dir_spec = (ENV_is_abs_path( $dir_spec)) ? $dir_spec : "$cwd/$dir_spec";
$file_spec = match_type( "$this_dir_spec/$command");
last if (defined $file_spec);
}
ENV_error( "'$command' (@COMMAND_TYPES) not found in PATH:", @PATH)
if (!defined $file_spec && $error_if_not_found);
}
}
}
if (defined $file_spec)
{
$file_spec = ENV_canonicalize_path($file_spec);
$COMMANDS{$command} = $file_spec;
}

return $file_spec;
}




sub FILE_which_path($@)
{
my ($command,
@paths,		    # abs paths
) = @_;
my $file_spec;	    # undef if not found

get_command_types()
if (!@COMMAND_TYPES);
foreach my $dir_spec (@paths)
{

$file_spec = match_type( "$dir_spec/$command");
last if (defined $file_spec);
}

return $file_spec;
}




sub get_command_types()
{
if ($IS_WIN32)
{
my $pathext = lc $ENV{PATHEXT};
my @pathext = ($pathext) ? ENV_split_path( $pathext) : qw( .com .exe .bat .cmd );
@COMMAND_TYPES = ( '', @pathext );
} else
{
@COMMAND_TYPES = ( '', qw( .sh ) );
}
}





sub match_type($)
{
my ($file_spec) = @_;
my $full_file_spec;     # undef = not found

foreach my $type (@COMMAND_TYPES)
{
my $ffs = "$file_spec$type";
if (-f $ffs)
{
$full_file_spec = $ffs;
last;
}
}
return $full_file_spec;
}

return 1;

