#===============================================================
#
#   genopt.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package glo::genopt;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
GENOPT_set_flag_prefix
GENOPT_set_optdefs
GENOPT_set_conflicts
GENOPT_set_envdefs
GENOPT_parse
GENOPT_get
GENOPT_get_all
GENOPT_get_changed
GENOPT_only_help_executed
GENOPT_makeopt
);
use strict;
use integer;

use glo::env;
use glo::slurp;
use glo::format;




sub GENOPT_set_flag_prefix($);
sub GENOPT_set_optdefs($$$$);
sub GENOPT_set_conflicts($);
sub GENOPT_set_envdefs($$$$);
sub GENOPT_parse();
sub GENOPT_get($);
sub GENOPT_get_all();
sub GENOPT_get_changed($);
sub GENOPT_only_help_executed();
sub GENOPT_makeopt($@);

sub full_opt($$);
sub do_semantics();
sub handle_opt_bool($);
sub handle_opt_string($);
sub handle_positional($);
sub do_item($$);
sub handle_envvar($);
sub handle_help($);
sub parse_time($);
sub validate_par($$$);
sub store_par($$$$);
sub syntax_error($;@);
sub option_error($$$);
sub print_usage();
sub print_args_usage(@);
sub parse_args(@);
sub unix_glob(@);
sub check_values($);
sub get_opt_values($);
sub put_opt_values($@);





my $IS_WIN32 = ENV_is_win32();

my $PROG_NAME;
my @ALIASES;
my $LONG_HELP_REF;
my $EXTRA_HELP_REF;

my $HELP_EXECUTED = 0;




my @ITEMS_ORDER;
my %ITEMS;



my %MNEMS;


my @M_ITEMS;	    # mandatory items

my @D_ITEMS;	    # items with default


my @CONFLICTS;






my %TYPES = ( b => "boolean",
i => "integer",
s => "string",
t => "time ([+=]hh:mm or [+=]mm)"
);
my $TYPES = 'b,i,s,t';
my %OCCUR = (
s => "scalar",
a => "array"
);
my $OCCUR = 's,a';
my %MANDA = (
m => "mandatory",
o => "optional"
);
my $MANDA = 'm,o';
my %CONSTR = (
n   => "",
r   => "range", 	    	    # array only
'>' => "greater",		    # value
'<' => "less",		    # value
s   => "choice:",		    # selection
e   => "existing-file/dir",
d   => "existing-dir",
f   => "existing-file",
x   => "non-existing-file/dir"
);
my $CONSTR = 'n,r,>,<,s,e,d,f,x';

my $FLAG_PREFIX = '--';
my $FLAG_PREFIX_L = length( $FLAG_PREFIX);

my $ENV_ALLOW = 0;
my $ENV_SET_HANDLER;
my $ENV_PRINT_HANDLER;
my $ENV_PREFIX;





my @ARGS;	# prepocessed @ARGV
my @ARG_REFS;





my $CURRENT_POS_PAR;

my %IN_OPTS;

my @IN_OPTS_ORDER;

my %IN_ENVS;

my @IN_ENVS_ORDER;

my %OPTS;



my @OPTS_TO_PRINT;
my $NR_ERRORS = 0;




sub GENOPT_set_flag_prefix($)
{
my ($prefix) = @_;

$FLAG_PREFIX = $prefix;
$FLAG_PREFIX_L = length( $FLAG_PREFIX);
}




sub GENOPT_set_optdefs($$$$)
{
my ($prog_name_or_ref,	# this is the name of the program, with aliasses
$items_ref,		# [ $mnem, $opt_name, $type_spec, $default, $item_help_or_ref ]
$help_text_or_ref,	# this is the general help-text. NOT optional
$extra_text_or_ref,	# this is the (optional) aditional help-text
) = @_;


















if (ref $prog_name_or_ref)
{
($PROG_NAME, @ALIASES) = @{$prog_name_or_ref};
} else
{
$PROG_NAME = $prog_name_or_ref;
}
$LONG_HELP_REF = (ref $help_text_or_ref) ? $help_text_or_ref : [ $help_text_or_ref ];
$EXTRA_HELP_REF = (ref $extra_text_or_ref) ? $extra_text_or_ref : [ $extra_text_or_ref ]
if (defined $extra_text_or_ref);




@M_ITEMS = ();
@D_ITEMS = ();

if (defined $items_ref)
{



foreach my $item_ref (@{$items_ref})
{
my @items =  @{$item_ref};
my ($mnem, $opt_name, $type_spec, $default, $item_help_or_ref) = @{ $item_ref };

my ($type, $occur, $manda) = split( '', substr( $type_spec, 0, 3));
my $constr = 'n';
my @constr_values=();
if (length $type_spec > 3)
{
$constr = substr( $type_spec, 3, 1);
@constr_values = split( ',',substr( $type_spec, 4));
}
ENV_fatal( "Invalid type '$type' ($TYPES)")
if (!exists $TYPES{$type});
ENV_fatal( "Invalid occur '$occur' ($OCCUR)")
if (!exists $OCCUR{$occur});
ENV_fatal( "Invalid manda '$manda' ($MANDA)")
if (!exists $MANDA{$manda});
ENV_fatal( "Invalid constr '$constr' ($CONSTR)")
if (!exists $CONSTR{$constr});

my $type_spec_ref = [$type, $occur, $manda, $constr, @constr_values];

push @M_ITEMS, $mnem if ($manda eq 'm');

if ($occur eq 's')
{
$OPTS{$opt_name} = undef;
} else  # array
{
$OPTS{$opt_name} = [];
}

if (defined $default)
{
push @D_ITEMS, $mnem;
} else
{
$default = "";
}

my $full_name;
my $full_spec;
if (substr( $mnem, 0, 1) eq '<')	# <n>, <*>
{
$full_name = "<$opt_name>";
$full_spec = $full_name;
} else
{
$full_name = "$FLAG_PREFIX$mnem";
if ($type eq 'b')
{
$full_spec = "$full_name\[+|-]";
} else
{
$full_spec = "$full_name=<$opt_name>";
}
}

my $item_help_ref;
$item_help_ref = (ref $item_help_or_ref) ? $item_help_or_ref : [ $item_help_or_ref ]
if (defined $item_help_or_ref);
my $items_refs = [ $mnem, $opt_name, $type_spec_ref, $default, $item_help_ref, $full_name, $full_spec ];
$ITEMS{$mnem} = $items_refs;

$MNEMS{$opt_name} = $mnem;

push @ITEMS_ORDER, $mnem;
}
}
}




sub GENOPT_set_conflicts($)
{
my ($con_refs_ref) = @_;

@CONFLICTS = ();
foreach my $con_ref (@{$con_refs_ref})
{
my ($l_arg_ref, @r_args) = @{$con_ref};
ENV_fatal( "'l_arg' must be defined")
if (!defined $l_arg_ref);
$l_arg_ref = [ $l_arg_ref ] unless (ref $l_arg_ref);

my $l_arg = $l_arg_ref->[0];
ENV_fatal( "No such mnem '$l_arg'")
if (!exists $ITEMS{$l_arg});

my @conflicts = ($l_arg_ref);
while (@r_args)
{
my $operator = shift @r_args;
ENV_fatal( "'operator' must be '=' or '!'")
if ($operator ne '=' && $operator ne '!');

my $r_arg_ref = shift @r_args;
ENV_fatal( "'r_arg' must be defined")
if (!defined $r_arg_ref);
$r_arg_ref = [ $r_arg_ref ] unless (ref $r_arg_ref);
my $r_arg = $r_arg_ref->[0];
ENV_fatal( "No such mnem '$r_arg'")
if (!exists $ITEMS{$r_arg});
push @conflicts, ($operator, $r_arg_ref);
}
push @CONFLICTS, [ @conflicts ];
}
}




sub GENOPT_set_envdefs($$$$)
{
(	$ENV_SET_HANDLER,
$ENV_PRINT_HANDLER,
$ENV_PREFIX,
my $envvars_ref,
) = @_;

$ENV_PREFIX = ''
if (!defined $ENV_PREFIX);
if (defined $envvars_ref)
{
foreach my $name (@{$envvars_ref})
{

&$ENV_SET_HANDLER( $name, undef);
}
$ENV_ALLOW = 1;
} else
{
$ENV_ALLOW = 0;
}
}




sub GENOPT_parse()
{
$NR_ERRORS = 0;
@OPTS_TO_PRINT = ();






parse_args( @ARGV);






do_semantics();




$LONG_HELP_REF = undef;
$EXTRA_HELP_REF = undef;
@ITEMS_ORDER = ();

@M_ITEMS = ();
@D_ITEMS = ();
@CONFLICTS = ();
@ARG_REFS =();

return;
}




sub GENOPT_get($)
{
my ($opt_name) = @_;

my $val_or_ref = $OPTS{$opt_name};
ENV_fatal( "No such opt '$opt_name'")
if (!defined $val_or_ref);
if (wantarray)
{
ENV_fatal( "Not list: $opt_name")
if (!ref $val_or_ref);
return @{$val_or_ref};
} else
{
ENV_fatal( "Is list: $opt_name")
if (ref $val_or_ref);
return $val_or_ref;
}
}




sub GENOPT_get_all()
{
return @ARGS;
}




sub GENOPT_get_changed($)
{
my ($excludes_ref) = @_;
my @opts;			    # items that do not have a default value

my @excludes = @{$excludes_ref};
foreach my $mnem (@IN_OPTS_ORDER)
{
my $opt_name = $ITEMS{$mnem}->[1];
if (! grep( $opt_name eq $_, @excludes))
{
my $value = get_opt_values( $opt_name);
if ($value ne $ITEMS{$mnem}->[3])   # default
{
push @opts, full_opt( $mnem, $value);
}
}
}

foreach my $name (@IN_ENVS_ORDER)
{
my $value = $IN_ENVS{$name};
push @opts, "$name=$value";
}

return @opts;
}





sub GENOPT_only_help_executed()
{
return $HELP_EXECUTED;
}




sub GENOPT_makeopt($@)
{
my ($opt_name,
@values,
) = @_;

my $value = join( ',', @values);
if ($IS_WIN32)
{
$value =~ s/ /\$_/g;
} else
{
$value =~ s/ /\%_/g;
}

my $mnem = $MNEMS{$opt_name};
ENV_fatal( "No such menmonic '$opt_name=$value'")
if (!$mnem);

if ($value eq $ITEMS{$mnem}->[3])	#$default
{
return ();
} else
{

return full_opt( $mnem, $value);
}
}




sub full_opt($$)
{
my ($mnem,
$value,
) = @_;
my $opt;

my $type = $ITEMS{$mnem}->[2]->[0];     # $type_spec->$type

if (substr( $mnem, 0, 1) eq '<')
{
$opt = $ITEMS{$mnem}->[5];  	    # $full_name
if ($value)
{
$opt .= "=$value";
}
} else
{
$opt = "$FLAG_PREFIX$mnem";
if ($type eq 'b')
{
if ($value eq '' || $value eq '+' || $value eq '1')
{
$value = '+';
} else
{
$value = '-';
}
$opt .= $value;
} else
{
if (defined $value)
{
$opt .= "=$value";
}
}
}
return $opt;
}




sub do_semantics()
{
$CURRENT_POS_PAR = 1;

%IN_OPTS = ();
@IN_OPTS_ORDER = ();

foreach my $ref (@ARG_REFS)
{




my $arg_type = $ref->[1];

if ($arg_type eq 'b')		    # opt_bool
{
handle_opt_bool( $ref);
} elsif ($arg_type eq 's')	    # opt_string
{
handle_opt_string( $ref);
} elsif ($arg_type eq '*')	    # positional
{
handle_positional( $ref);
} elsif ($arg_type eq '=')	    # EnvVar
{
handle_envvar( $ref);
} elsif ($arg_type eq 'h')	    # opt_help
{
handle_help( $ref);
exit 0;
} else
{
ENV_fatal ("Impossible else (arg_type=$arg_type)");
}
}




foreach my $mnem (@D_ITEMS)
{
my $opt_name = $ITEMS{$mnem}->[1];  # $opt_name

my $value_or_ref = $OPTS{$opt_name};
if (ref $value_or_ref)
{
if (!@{$value_or_ref})
{
$OPTS{$opt_name} = [ split( /,/, $ITEMS{$mnem}->[3]) ]; # $default
}
} else
{
if (!defined $value_or_ref)
{
$OPTS{$opt_name} = $ITEMS{$mnem}->[3];	# $default
}
}
}




foreach my $mnem (@M_ITEMS)
{

my $opt_name = $ITEMS{$mnem}->[1];  # $opt_name

my @values = get_opt_values( $opt_name);
if ((!@values) || (!defined $values[0]) || ($values[0] eq ''))
{
my $full_name = $ITEMS{$mnem}->[5]; # $full_name
syntax_error( "Mandatory '$full_name' missing", $mnem);
}
}




foreach my $c_ref (@CONFLICTS)
{
my ($l_arg_ref, @r_args) = @{$c_ref};
if (check_values( $l_arg_ref))
{
while (@r_args)
{
my $operator = shift @r_args;
my $r_arg_ref = shift @r_args;
my $result = check_values( $r_arg_ref);
if ($operator eq '=' xor !$result)
{
my $l_mnem = $l_arg_ref->[0];
my $l_name = $ITEMS{$l_mnem}->[1];	# $opt_name
my $full_name = full_opt($l_mnem, get_opt_values( $l_name));

my ($r_mnem, @values) = @{$r_arg_ref};
my $conflict = full_opt( $r_mnem, join( ',', @values));
syntax_error( "Combination of '$full_name' with $conflict is not possible",
($l_mnem, $r_mnem));

}
}
}
}




if ($NR_ERRORS > 0)
{
print_usage();
print_args_usage( @OPTS_TO_PRINT);
ENV_say( 1, "Command-line error(s)",
"$PROG_NAME @ARGV");
exit 1;
}
}




sub handle_opt_bool($)
{
my ($arg_ref,   # [ $original_arg, $arg_type, $key, $value ]
) = @_;

my ($original_arg, $arg_type, $key, $value) = @{$arg_ref};


if (exists $ITEMS{$key})
{
my $item_ref = $ITEMS{$key};
my $item_type = $item_ref->[2]->[0];	#   $type_spec_ref -> $type

if ($item_type eq 'b')
{
do_item( $arg_ref, $item_ref);
} else
{
option_error( $original_arg, $key, "Not a bool-type option: =<value> required");
}
} else
{
syntax_error( "No such option '$original_arg'");
}
}




sub handle_opt_string($)
{
my ($arg_ref,   # [ $original_arg, $arg_type, $key, $value ]
) = @_;

my ($original_arg, $arg_type, $key, $value) = @{$arg_ref};


if (exists $ITEMS{$key})
{
my $item_ref = $ITEMS{$key};
my $item_type = $item_ref->[2]->[0];	#   $type_spec_ref -> $type

if ($item_type ne 'b')	    # i, s, t
{
do_item( $arg_ref, $item_ref);
} else
{
option_error( $original_arg, $key, "Not a string-type option: + or - required");
}
} else
{
syntax_error( "No such option '$original_arg'");
}
}




sub handle_positional($)
{
my ($arg_ref,   # [ $original_arg, $arg_type, $key, $value ]
) = @_;

my ($original_arg, undef, undef, $value) = @{$arg_ref};


my $key = "<$CURRENT_POS_PAR>";
if (exists $ITEMS{$key})
{
my $item_ref = $ITEMS{$key};
$arg_ref->[2] = $key;
do_item( $arg_ref, $item_ref);
$CURRENT_POS_PAR++;
} else
{
$key = '<*>';
if (exists $ITEMS{$key})
{
my $item_ref = $ITEMS{$key};
$arg_ref->[2] = $key;
do_item( $arg_ref, $item_ref);
} else
{
syntax_error( "Invalid Positional Parameter $key '$original_arg'");
}
}
}




sub do_item($$)
{
my ($arg_ref,   # [ $original_arg, $arg_type, $key, $value ]
$item_ref,  # [ $mnem, $opt_name, $type_spec_ref, $default, $help_text, $full_name, $full_spec]

) = @_;

my ($original_arg, undef, $key, $value) = @{$arg_ref};
my (undef, $opt_name, $type_spec_ref) = @{$item_ref};
my ($item_type, $occur) = @{$type_spec_ref};

$IN_OPTS{$key} = $value;
push @IN_OPTS_ORDER, $key;




my @values = ($occur eq 'a') ? split( ',', $value) : ($value);
if ($item_type eq 'b')
{
foreach my $value (@values)
{
$value = ($value eq '-') ? 0 : 1;
}
} elsif ($item_type eq 'i')
{
foreach my $value (@values)
{
if ($value =~ /^[-+]?\d+$/)		# numeric
{
$value *= 1;
} else
{
$value = undef;
option_error( $original_arg, $key, "Value must be an integer");
}
}
} elsif ($item_type eq 't')
{
foreach my $value (@values)
{
$value = parse_time( $value);
option_error( $original_arg, $key, "Value must have time format ([+|=][[H]H:][M]M or 'now'")
if (!defined $value);
}
}




foreach my $value (@values)
{
if (defined $value && validate_par( $original_arg, $key, $value))
{
store_par( $occur, $opt_name, $value, $key);
}
}
}




sub handle_envvar($)
{
my ($arg_ref,   # [ $original_arg, $arg_type, $key, $value ]
) = @_;

my ($original_arg, undef, $key, $value) = @{$arg_ref};


if ($ENV_ALLOW)
{
my $new_value = &$ENV_SET_HANDLER( $key, $value);
syntax_error( "'$original_arg' Not accepted")
if (!defined $new_value);
$IN_ENVS{ $key} = $value;
push @IN_ENVS_ORDER, $key;
} else
{
syntax_error( "Cannot specify EnvVar Arguments ($original_arg)");
}
}




sub handle_help($)
{
my ($arg_ref,   # [ $original_arg, $arg_type, $key, $value ]
) = @_;

my ( undef, undef, $key, @values ) = @{$arg_ref};


if ($key eq 'help')
{
if (@ALIASES)
{
ENV_say( 0, "$PROG_NAME (@ALIASES):");
} else
{
ENV_say( 0, "$PROG_NAME:");
}
map { ENV_say( 0, "  $_") } @{$LONG_HELP_REF};

print_usage();

my @usages = (@values) ? @values : @ITEMS_ORDER;
print_args_usage( @usages);

if ($ENV_ALLOW)
{
&$ENV_PRINT_HANDLER();
} else
{
ENV_say( 0, "Environment Variables:");
ENV_say( 0, "  None");
}

if ($EXTRA_HELP_REF)
{
ENV_say( 0, "Note:");
map { ENV_say( 0, "  $_") } @{$EXTRA_HELP_REF};
}

my $app_name_lc = lc ENV_get_application_name();
my $help_command = "${app_name_lc}help";
ENV_say( 0, "Help:");
ENV_say( 0, "  Extensive help (html) can be obtained with command '$help_command'")
unless( $PROG_NAME eq $help_command);
} else  # $help eq '--h'
{
print_usage();
print_args_usage( @values);
}
$HELP_EXECUTED = 1;
}




sub parse_time($)
{
my ($time) = @_;
my $new_time = undef;

if ($time eq 'now')
{
$new_time = 'now';
} elsif ($time =~ /^([=\+]?)(\d\d?)(:\d\d?)?$/)
{
my $delta = $1; $delta = ''
if (! defined $delta);
my $hours = $2; $hours = ''
if (! defined $hours);
my $minutes = $3; $minutes = ''
if (! defined $minutes);

$delta = '='
if ($delta eq '');
if ($minutes eq '')
{
$minutes = $hours;
$hours = '00';
} else
{
$minutes = substr( $minutes, 1);
}
$new_time = "$delta$hours:$minutes";
$new_time = undef
if ($hours * 1 > 24);
$new_time = undef
if ($minutes * 1 >= 60);
}

return $new_time;
}




sub validate_par($$$)
{
my ($arg,
$mnem,
$par,
) = @_;
my $val_ok = 1;

my $type_spec_ref = $ITEMS{$mnem}->[2];
my ($d_type, $d_occur, $d_man, $constr_type, @constr_values) = @{$type_spec_ref};
if ($constr_type ne "n")
{

if ($constr_type eq 'r')		    # Range
{
$val_ok = 0;
foreach my $range (@constr_values)
{
my ($r1, $r2) = split( /\.\./, $range);
$r2 = $r1 if (!defined $r2);
if ($par >= $r1 && $par <= $r2)
{
$val_ok = 1;
last;
}
}
option_error( $arg, $mnem, "Value ($par) not in Range (@constr_values)") if (!$val_ok);
} elsif ($constr_type eq '>')
{
my $constr_value =  $constr_values[0];
$val_ok = ($par > $constr_value);
option_error( $arg, $mnem, "Value ($par) must be > $constr_value") if (!$val_ok);
} elsif ($constr_type eq '<')
{
my $constr_value =  $constr_values[0];
$val_ok = ($par < $constr_value);
option_error( $arg, $mnem, "Value ($par) must be < $constr_value") if (!$val_ok);
} elsif ($constr_type eq 's')
{
$val_ok = grep( $par eq $_, @constr_values);
option_error( $arg, $mnem, "Value ($par) not in Selection (@constr_values)") if (!$val_ok);
} elsif ($constr_type eq 'e')
{
$val_ok = -e $par;
option_error( $arg, $mnem, "$par Must Exist") if (!$val_ok);
} elsif ($constr_type eq 'd')
{
$val_ok = -d $par;
option_error( $arg, $mnem, "$par Must be Directory") if (!$val_ok);
} elsif ($constr_type eq 'f')
{
$val_ok = -f $par;
option_error( $arg, $mnem, "$par Must be File") if (!$val_ok);
} elsif ($constr_type eq 'x')
{
$val_ok = !-e $par;
option_error( $arg, $mnem, "$par Must Not Exist") if (!$val_ok);
} else
{
ENV_fatal( "Falling through if-else '$constr_type' (@constr_values)");
}
}

return $val_ok;
}




sub store_par($$$$)
{
my ($occur,
$opt_name,
$par,
$opt,
) = @_;


if ($occur eq 's')
{
my $prev_val = $OPTS{$opt_name};
ENV_say( 1, "$opt=$par replaced $prev_val")
if (defined $prev_val);
}
put_opt_values( $opt_name, $par);
}




sub syntax_error($;@)
{
my ($texts_ref,
@opts,
) = @_;

my @texts = (ref $texts_ref) ? @{$texts_ref} : ($texts_ref);
ENV_error( @texts);
foreach my $opt (@opts)
{
push @OPTS_TO_PRINT, $opt
if (!grep( $opt eq $_, @OPTS_TO_PRINT));
}
$NR_ERRORS++;
}




sub option_error($$$)
{
my ($original_arg,
$opt,
$text,
) = @_;

ENV_error( "Option '$original_arg'", $text);
push @OPTS_TO_PRINT, $opt
if (!grep( $opt eq $_, @OPTS_TO_PRINT));
$NR_ERRORS++;
}




sub print_usage()
{
ENV_say( 0, "Usage:");
my @items;
foreach my $mnem (@ITEMS_ORDER)
{
my $opt_name = $ITEMS{$mnem}->[1];
my $type_spec_ref = $ITEMS{$mnem}->[2];
my $full_opt_name = $ITEMS{$mnem}->[6];;
my ($type, $occur, $man) = @{$type_spec_ref};

$full_opt_name = "[$full_opt_name]" if ($man eq 'o');
$full_opt_name = "$full_opt_name..." if ($occur eq 'a');

push @items, $full_opt_name;
}

ENV_say( 0, FORMAT_table( 0, 2, ' ', undef, [ $PROG_NAME, "@items" ]));
}




sub print_args_usage(@)
{
my @args = @_;

if (@args)
{
ENV_say( 0, "Args:");

my @line_refs;
foreach my $mnem (@args)
{

if (exists $ITEMS{$mnem})
{
my (undef, $opt_name, $type_spec_ref, $default, $item_help_ref, $full_name, $full_spec) = @{$ITEMS{$mnem}};
my ($type, $occur, $man, $constr, @constr_vals) = @{$type_spec_ref};

my $extra = (substr( $mnem, 0, 1) eq '<') ? $mnem : ($type eq 'b') ? $opt_name : '';
$extra = "($extra) " if ($extra ne '');
my $text = "$extra$MANDA{$man} $TYPES{$type} $OCCUR{$occur} $CONSTR{$constr} @constr_vals";
push @line_refs, [ $full_spec, $text ];

push @line_refs, [ '', "default=$default" ] if ($default ne '');

if (defined $item_help_ref)
{
push @line_refs, map { [ '', $_ ] } @{$item_help_ref};
}
} else
{
ENV_say( 1, "No such mnemonic '$mnem'",
"Available mnemonics: = @ITEMS_ORDER\n");
exit 9;
}
}
ENV_say( 0, FORMAT_table( 0, 2, '  ', undef, @line_refs));
}
}




sub parse_args(@)
{
my @args = @_;
my ($handle_args, $handle_flags) = (1 ,1);






if ($IS_WIN32)
{
@args = unix_glob( @args);
map { s/\$_/ /g } @args;
} else
{
map { s/\%_/ /g } @args;
}
map { s/%20/ /g } @args;

@args = ()
if ("@args" eq '');

while ($handle_args && @args)
{
my $arg = shift @args;

if ($arg eq '---')	# '---' means: stop handling of arguments
{
$handle_args = 0;
} elsif ($arg eq '--')	# '--' means: stop handling of $FLAG_PREFIX arguments
{
$handle_flags = 0;
} else
{
my ( $original_arg, $arg_type, $key, $value );




$original_arg = $arg;
if ($handle_flags && $arg =~ /^$FLAG_PREFIX(.*)/)	# --opt[+|-], --opt=<value>
{
my $opt = $1;	    # remainder
($key, $value) = $opt =~ /^([a-z]\w*)(.*)/;	# starts with lowercase alpha char followed by word characters
if (defined $key)
{
if ($value eq '' || $value eq '+' || $value eq '-')
{
$arg_type = 'b';
} elsif (substr( $value, 0, 1) eq '=')
{
$arg_type = 's';
$value = substr( $value, 1);
my $fc = substr( $value, 0, 1);
if ($fc eq '"' || $fc eq "'")
{
while (substr( $value, -1, 1) ne $fc && @args)
{
my $next_arg = shift @args;
$value .= ' ' . $next_arg;
$original_arg .= ' ' . $next_arg;
}
syntax_error( "Terminating quote ($fc) not found ($original_arg)")
if (substr( $value, -1, 1) ne $fc);
$value = substr( $value, 1, -1);	    # remove first and last quote

}
} else
{
syntax_error( [ "Invalid value syntax for option '$key' ($arg).",
"Must be either <empty>, - , + or =<value>" ]);
$arg_type = '?';
}
} else
{
syntax_error( "Invalid option name for '$arg'");
$arg_type = '?';
}
} elsif ($arg =~ /^([A-Za-z][A-Za-z0-9_]*)=(.*)/)	# EnvVar (starts with an alphabethical character and contains '=')
{
$arg_type = '=';
$key = $1;
$value = $2;	# remainder
} else						# positional
{
$arg_type = '*';
$key = '';
$value = $arg;
}

if (($key eq 'h' || $key eq 'help') && $arg_type eq 'b')
{
$arg_type = 'h';
my @values;
while (@args)
{
my $this_arg = shift @args;
if ($this_arg =~ /^$FLAG_PREFIX(.*)/)
{
push @values, $1 if ($1 ne 'help' && $1 ne 'h');
}
}

push @ARGS, $arg;

push @ARG_REFS, [ $original_arg, $arg_type, $key, @values ];
} elsif ($arg_type eq 's' && $key eq 'inc')
{



my $file = $value;
ENV_error_exit( "No such '${FLAG_PREFIX}inc' file '$file'")
if (!-e $file);
my @lines = SLURP_file( $file);
@lines = grep !($_ eq '' || substr($_,0,1) eq '#'), @lines;	    # skip empty and comment lines
($handle_args, $handle_flags) = parse_args( @lines);		    # RECURSE! One item per line
} else
{

if ($arg_type ne '?')
{
push @ARGS, $arg;
push @ARG_REFS, [ $original_arg, $arg_type, $key, $value ];
}
}
}
}
ENV_say( 1, "Ignored '@args' after '---'")
if (!$handle_args);

return ($handle_args, $handle_flags);
}




sub unix_glob(@)
{
my @args = @_;
my @new_args;

foreach my $arg (@args)
{
my @files = ENV_glob( $arg);

push @new_args, (@files) ? @files : $arg;
}
return @new_args;
}





sub check_values($)
{
my ($arg_ref) = @_;
my $result; 		# bool

my ($mnem, @values) = @{$arg_ref};

my $opt_name = $ITEMS{$mnem}->[1];

my $current_value = get_opt_values($opt_name);




if (exists $IN_OPTS{$mnem})
{

if (! defined $values[0])
{



$result = ($current_value ne '');
} elsif (!@values)
{



$result = ($current_value eq '')
} else
{



$result = 0;
foreach my $value (@values)
{
if ($current_value eq $value)
{
$result = 1;
last;
}
}
}

} else
{
$result = 0;

}

return $result;
}




sub get_opt_values($)
{
my ($opt_name) = @_;

ENV_fatal( "No such option '$opt_name'")
if (!exists $OPTS{$opt_name});
my $value = $OPTS{$opt_name};
if (wantarray)
{
if (ref $value)
{
return @{$value};
} else
{
return ($value);
}
} else
{
if (ref $value)
{
return join( ',', @{$value});
} else
{
return $value;
}
}
}




sub put_opt_values($@)
{
my ($opt_name,
@values,
) = @_;

ENV_fatal( "No such option '$opt_name'")
if (!exists $OPTS{$opt_name});
if (ref $OPTS{$opt_name})
{
push @{$OPTS{$opt_name}}, @values;

} else
{
$OPTS{$opt_name} = $values[0];

}
}

return 1;
