#=================================================
#
#   proc.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package glo::proc;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
PROC_at
PROC_at_show
PROC_spawn_detached
PROC_spawn_xterm
PROC_spawn
PROC_wait
PROC_waitpid
PROC_waitpid_test
PROC_kill
);

use strict;
use integer;

use POSIX qw(:errno_h :sys_wait_h);
BEGIN
{
if ($^O eq 'MSWin32')
{
require glo::proc_win32;
import  glo::proc_win32;
}
}

use glo::env;
use glo::shell;
use glo::file;




sub PROC_at($$$$);
sub PROC_at_show();
sub PROC_spawn_detached($);
sub PROC_spawn_xterm($$$$$$$);
sub PROC_spawn($$);
sub PROC_wait();
sub PROC_waitpid($);
sub PROC_waitpid_test($);
sub PROC_kill($);




my $WNOHANG = 1;
my $IS_WIN32 = ENV_is_win32();




sub PROC_at($$$$)
{
my ($command,
$name,			# jobname
$delay,			# now | [=+]hh:mm
$must_mail		# bool
) = @_;

my $os_command = SHELL_at( $command, $name, $delay, $must_mail);
if ($IS_WIN32)
{
$os_command = "start \"$name\" /I /MIN $os_command";
} else
{

}
ENV_system( $os_command, 0 , undef);
}





sub PROC_at_show()
{
if ($IS_WIN32)
{

} else
{
my $rc = ENV_system( SHELL_at_show(), [0, 1], undef);
ENV_warn( "'\\atq' returned rc=$rc. This seems to happen sometimes without known reason...")
if ($rc != 0);
}
}





sub PROC_spawn_detached($)
{
my ($command_ref,		# ADT
) = @_;

ENV_debug( 'PROC_spawn_detached:');

if ($IS_WIN32)
{
my $command_line = ENV_command_line( $command_ref);
my ($type, $prefix, $os_exec, @args) = ENV_get_command_ref_data( $command_ref);
if ($prefix eq 'perl')
{
my $command_spec = ENV_get_perl_cmd();
if ($command_spec =~ /^perl(\s|$)/)
{
$command_spec = FILE_which( ENV_perl_path_nospace( 'perl'), 1);	    # $error_if_not_found
ENV_fatal( "Cannot find 'perl'")
if (!defined $command_spec);
$command_spec =~ tr%/%\\%;	# os_filespec
}
my ($pid, $proc) = PROC_WIN32_create_detached( $command_spec, $command_line, '.');
} else
{
$command_line = 'start ' . $command_line
if ($command_line !~ /^start /);

ENV_system( $command_line, 0, undef);
}
} else #assume LUNIX
{
my $command_line = ENV_command_line( $command_ref);

my $logfile = '/dev/null';
$logfile = ENV_get_tmp_spec( 'nohup.out')
if (ENV_is_debug());

ENV_system( "nohup $command_line 1>> $logfile 2>&1 < /dev/null &", 0, undef);
}
}




sub PROC_spawn_xterm($$$$$$$)
{
my ($command_ref,		# ADT
$xtitle,
$fg,
$bg,
$cols,
$rows,
$icon			# file_spec
) = @_;

ENV_debug( "PROC_spawn_xterm:");

if ($IS_WIN32)
{
my $command_line;
my ($type, $prefix, $os_exec, @args) = ENV_get_command_ref_data( $command_ref);
$prefix = ''
if ($prefix eq 'start');
$command_line =  "start \"$xtitle\" " . ENV_command_line( [ $type, $prefix, $os_exec, @args ]);

ENV_system( $command_line, 0, undef);

} else #assume LUNIX
{
my $command_line = ENV_command_line( $command_ref);

$command_line = "xterm -T \"$xtitle\" -fg blue -bg white -geometry 100x40 -e " . $command_line;


my $logfile = '/dev/null';
$logfile = ENV_get_tmp_spec( 'nohup.out') if (ENV_is_debug());

ENV_system( "nohup $command_line 1>> $logfile 2>&1 < /dev/null &", 0, undef);
}
}






sub PROC_spawn($$)
{
my ($command_ref,		# ADT
$log_file,		# undef = do redirect
) = @_;
my $pid = 0;

ENV_debug( "PROC_spawn:");

do
{
$pid = fork;
if (defined $pid)
{
if ($pid == 0)
{



ENV_debug( "PROC_spawn: fork child");
ENV_exec( $command_ref, $log_file);
} else
{



ENV_debug( "PROC_spawn: fork parent: child_pid=$pid");
}
} else
{
if ($! == EAGAIN)
{



my $command_line = ENV_command_line( $command_ref);
ENV_warn( "fork 'exec $command_line' failed",
"- $!",
"- Retry after 5 seconds");
sleep( 5);
} else
{
ENV_fatal( "Unable to fork: $!");
}
}
} while (!defined $pid);

return $pid;
}






sub PROC_wait()
{
my ($r_pid, $rc);

ENV_debug( "PROC_wait");
$r_pid = wait();
ENV_debug( "- r_pid=$r_pid");

if ($r_pid != -1)
{
$rc = ENV_rc_status( $?, 'wait');
} else
{
$rc = -1;
}

return ($r_pid, $rc);
}






sub PROC_waitpid($)
{
my ($pid) = @_;	    # -1 == wait for any (is equal to PROC_Wait)
my $rc = 0;

ENV_debug( "PROC_waitpid pid=$pid");
my $r_pid = waitpid( $pid, 0);
ENV_debug( "- r_pid=$r_pid");

if ($r_pid != -1)
{
$rc = ENV_rc_status( $?, "$r_pid=waitpid $pid, 0");
} else
{
$rc = -1;
}

return (wantarray ? ($r_pid, $rc) : $rc);
}







sub PROC_waitpid_test($)
{
my ($pid) = @_;	# -1 == wait for any.
my $rc = 0;

ENV_debug( "PROC_waitpid_test pid=$pid");
my $r_pid = waitpid( $pid, WNOHANG);
ENV_debug( "- r_pid=$r_pid");

if ($r_pid != -1 && $r_pid != 0)
{
$rc = ENV_rc_status( $?, "waitpid_test $r_pid, " . WNOHANG);
} else
{
$rc = -1;
}

return (wantarray ? ($r_pid, $rc) : $rc);
}





sub PROC_kill($)
{
my ($pid) = @_;

ENV_debug( "$$ PROC_kill pid=$pid");

if ($IS_WIN32)
{
my $rc = ENV_system( "TASKKILL /F /T /PID $pid", undef, undef);
ENV_debug( "PROC_kill rc = $rc");
} else
{
my $count = kill( 9, $pid);
ENV_debug( "PROC_kill count=$count");
}
}

return 1;
