#===============================================================
#
#   scmsvn.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================
#   File states:
#    -2 => 'NO-FILE',
#    -1 => 'NOT-CCM',
#     0 => 'CHECKED-IN',
#     1 => 'CHECKED-OUT',
#===============================================================

package glo::scmsvn;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
SCMSVN_init
SCMSVN_connect
SCMSVN_cache_states
SCMSVN_get_states
SCMSVN_checkout
SCMSVN_uncheckout
SCMSVN_checkin
SCMSVN_remove
SCMSVN_move
SCMSVN_add_dirs
SCMSVN_add_files
SCMSVN_set_ignore
SCMSVN_setup
);

use strict;
use integer;

use File::Path;
use glo::env;
use glo::spit;
use glo::ask;

ENV_say( 1, "Loading SubVersioN Plugin...");




sub SCMSVN_init($$);
sub SCMSVN_connect($$$);
sub SCMSVN_cache_states($);
sub SCMSVN_get_states($);
sub SCMSVN_checkout($);
sub SCMSVN_uncheckout($);
sub SCMSVN_checkin($);
sub SCMSVN_remove($);
sub SCMSVN_move($);
sub SCMSVN_add_dirs($);
sub SCMSVN_add_files($);
sub SCMSVN_set_ignore($$$);
sub SCMSVN_setup($$$);

sub check_repos($$$);
sub exec_version();
sub exec_info();
sub exec_status($);
sub exec_full_status_recurse($);
sub exec_commit($);
sub exec_revert($);
sub exec_delete($);
sub exec_move($$);
sub exec_add($);
sub exec_propset_ignore($$$);
sub exec_list($);

sub extract_state($$);




my $APP_NAME;
my $APP_ROOT_PATH;
my @SKIPTYPES = qw( ^\.svn$ );

my $EXEC_OS = 'svn';		# Keep in synch with %PREINFO_DATA in scm.pm

my %FSTATES;


my $CACHE_TOP_LEVEL_DIR;	# ends with '/'
my $CACHE_TOP_LEVEL_DIR_L;




sub SCMSVN_init($$)
{
($APP_NAME,
$EXEC_OS) = @_;
}





sub SCMSVN_connect($$$)
{
(undef,		    # REPOSITORY
$APP_ROOT_PATH,
undef) = @_;	    # DATA
my ($repos_root_path, $internal_path, $connect_texts_ref);

$connect_texts_ref = exec_version();

($repos_root_path, $internal_path) = exec_info();


%FSTATES = ();	    # clear previous file-states
$CACHE_TOP_LEVEL_DIR = undef;
$CACHE_TOP_LEVEL_DIR_L = undef;

return ($repos_root_path, $internal_path, $connect_texts_ref);
}




sub SCMSVN_cache_states($)
{
my ($dir_spec) = @_;

foreach my $ref (exec_full_status_recurse( $dir_spec))
{
my ($states_6, $file_spec) = @{$ref};
my $scm_state = extract_state( $states_6, $file_spec);
$FSTATES{$file_spec} = $scm_state;
}
$CACHE_TOP_LEVEL_DIR = "$dir_spec/";
$CACHE_TOP_LEVEL_DIR_L = length $CACHE_TOP_LEVEL_DIR;
}








sub SCMSVN_get_states($)
{
my ($spec_refs_ref,	    # [ [ $spec, $state ], ... ]
) = @_;

my @specs;
foreach my $spec_ref (@{$spec_refs_ref})
{
my ($spec, $state) = @{$spec_ref};

if (defined $state)	    # $state may be -2 (no such file/dir)
{

} else
{

$state = $FSTATES{$spec};
if (defined $state)
{

$spec_ref->[1] = $state;	# $state
} else
{
if (defined $CACHE_TOP_LEVEL_DIR_L &&
substr( $spec, 0, $CACHE_TOP_LEVEL_DIR_L) eq $CACHE_TOP_LEVEL_DIR )
{



$spec_ref->[1] = -1;	# $state = Not SCM

} else
{
push @specs, $spec;
}
}
}
}

if (@specs)
{



foreach my $ref (exec_status( \@specs))
{
my ($states_6, $file_spec) = @{$ref};
my $scm_state = extract_state( $states_6, $file_spec);
my $state = substr( $states_6, 0, 1);
$FSTATES{$file_spec} = $scm_state;
}
}




foreach my $spec_ref (@{$spec_refs_ref})
{
my ($spec, $state) = @{$spec_ref};
$spec_ref->[1] = $FSTATES{$spec}
if (!defined $state)
}
}




sub SCMSVN_checkout($)
{
my ($specs_ref) = @_;

ENV_chmod( 'u+w', @{$specs_ref});

map { $FSTATES{$_} = 1 } @{$specs_ref}; # checkout
}




sub SCMSVN_checkin($)
{
my ($specs_ref) = @_;

exec_commit( $specs_ref);
ENV_chmod( 'ugo-w', @{$specs_ref});
map { $FSTATES{$_} = 0 } @{$specs_ref}; # checkin
}




sub SCMSVN_uncheckout($)
{
my ($specs_ref) = @_;

exec_revert( $specs_ref);
ENV_chmod( 'ugo-w', @{$specs_ref});
map { $FSTATES{$_} = 0 } @{$specs_ref}; # checkin
}




sub SCMSVN_remove($)
{
my ($specs_ref) = @_;

ENV_chmod( 'u+w', @{$specs_ref});
exec_delete( $specs_ref);
map { rmtree( $_, 0, 1) } @{$specs_ref};
map { $FSTATES{$_} = -1 } @{$specs_ref}; # notscm
}




sub SCMSVN_move($)
{
my ($pair_refs_ref) = @_;

foreach my $ref (@{$pair_refs_ref})
{
my ($old_spec, $new_spec) = @{$ref};
ENV_chmod( 'u+w', $old_spec);
exec_move( $old_spec, $new_spec);
$FSTATES{$old_spec} = -1;	# notscm
$FSTATES{$new_spec} = 1;    	# checkout
}
}




sub SCMSVN_add_dirs($)
{
my ($dir_refs_ref) = @_;	    #	[ $dir, $ignores_ref ]

my %ignores;

my @dir_specs;
foreach my $ref (@{$dir_refs_ref})
{
my ($dir, $ignores_ref) = @{$ref};
ENV_mkdir( $dir, 0777)
if (!-d $dir);
push @dir_specs, $dir;
if (@{$ignores_ref})
{
my $ignores = "@{$ignores_ref}";
push @{$ignores{$ignores}}, $dir;
}
}

exec_add( \@dir_specs);
foreach my $ignores (keys %ignores)
{
my $dirs_ref = $ignores{$ignores};
my @ignores = split( ' ', $ignores);
exec_propset_ignore( $dirs_ref, [ @ignores ], 0);   # $must_recurse
}

map { $FSTATES{$_} = 1 } @dir_specs; # checkout
}




sub SCMSVN_add_files($)
{
my ($files_ref) = @_;

map { ENV_touch_file( $_) if (!-f $_) } @{$files_ref};
exec_add( $files_ref);
map { $FSTATES{$_} = 1 } @{$files_ref}; # checkout
}




sub SCMSVN_set_ignore($$$)
{
my ($dirs_ref,
$ignores_ref,
$must_recurse,	# Bool. propagate
) = @_;

exec_propset_ignore( $dirs_ref, $ignores_ref, $must_recurse);
}




sub SCMSVN_setup($$$)
{
my (undef,				# $app_root_path
$default_scms_repository_path,
$default_scms_data,
) = @_;
my ($scms_repository_path, $scms_data);

my $default_repository = $default_scms_repository_path;
$default_repository = '' if (!defined $default_repository);
ENV_say( 1, "Repository access URLs:");
ENV_say( 0, "  file:///path/to/repos         | Direct disk access (Unix)");
ENV_say( 0, "  file:///X:/path/to/repos      | Direct disk access (Windows)");
ENV_say( 0, "  http://<repos>                | WebDAV for Apache server");
ENV_say( 0, "  https://<repos>               | Same as http: but with SSL encryption");
ENV_say( 0, "  svn://<repos>                 | Access via svnserv server");
ENV_say( 0, "  svn+ssh://<repos>             | Same as svn, but through SSH tunnel");
$scms_repository_path = ASK_text( 'Enter SVN Repository URL', $default_repository, -1, [ \&check_repos, undef ]);

my $default_data = $default_scms_data;
$default_data = '' if (!defined $default_data);
$scms_data = ASK_text( 'Enter SVN Internal Path', $default_data, -1);

return ( $scms_repository_path, $scms_data );
}




sub check_repos($$$)
{
my ($values_ref,
$default_ref,
$args_ref,
) = @_;
my $error_txt = '';

$values_ref->[0] =~ s!\\!/!g;
my $scms_repository_path = $values_ref->[0];

if ($scms_repository_path =~ /\s/)
{
$error_txt = "No white space allowed in '$scms_repository_path'";
} else
{
my $rc = exec_list( $scms_repository_path);
$error_txt = "Not a valid SVN Repository URL"
if ($rc == 1);
}

return $error_txt;
}
























sub exec_version()
{
my @connect_texts;

my @stdout;
my $command = "$EXEC_OS --version";
ENV_run3( $command, undef, [0], \@stdout, undef);
if ($stdout[0] !~ /^svn, version/)
{
ENV_say( 0, @stdout);
ENV_fatal( "Failed to start SVN session");
}
@connect_texts = ( join( ',', @stdout[0..1]) );

return [ @connect_texts];
}



















sub exec_info()
{
my ($repos_root_path, $internal_path);

my $stdout;
my $stderr;
my $command = "$EXEC_OS info $APP_ROOT_PATH";
my $rc = ENV_run3( $command, undef, [0,1], \$stdout, \$stderr);
if ($rc == 0)
{
my ($url) = $stdout =~ /^URL: (..*)$/m;
($repos_root_path) = $stdout =~ /^Repository Root: (..*)$/m;
if (!defined $url || !defined $repos_root_path)
{
print $stdout;
ENV_error_exit( "'svn info' returned unexpected result\n");
}
$internal_path = $url;
$internal_path =~ s!^$repos_root_path/!!;
} else # $rc ==1
{
print $stdout;
print $stderr;
ENV_error_exit( "'svn info' returned non-zero result ($rc)\n");
}

return ($repos_root_path, $internal_path);
}









sub exec_full_status_recurse($)
{
my ($dir_spec) = @_;
my (@status_refs);	    # ( [ $states_6, $file_spec ])

my @stdout;
my $stderr;
my $command = "$EXEC_OS status -v $dir_spec";
my $rc = ENV_run3( $command, undef, [0], \@stdout, \$stderr);
print "STDERR: $stderr"
if ($stderr);







if (@stdout)
{
my $os_dir_spec_qm = quotemeta( ENV_os_path( $dir_spec));
foreach my $line (@stdout)
{

my $states_6 = substr( $line, 0, 6);
my ($file_spec) = substr( $line, 41) =~ /($os_dir_spec_qm.*)/;
if (defined $file_spec)
{
$file_spec = ENV_perl_paths( $file_spec);
push @status_refs, [ $states_6, $file_spec ];
} else
{
ENV_error( "==>$line", 'svn status: Filespec cannot be determined');
}
}
}


return @status_refs;
}















sub exec_status($)
{
my ($specs_ref) = @_;   # files/dirs MUST exist!
my (@status_refs);	    # ( [ $states_6, $file_spec ])




my @stdout;
my $stderr;

my $command = "$EXEC_OS status -N";
my $rc = ENV_run3( $command, $specs_ref, [0], \@stdout, \$stderr);
print "STDERR: $stderr"
if ($stderr);

if (@stdout)
{
my %states = map { $_ => '      ' } @{$specs_ref};

foreach my $line (@stdout)
{

my $states_6 = substr( $line, 0, 6);
my $file_spec = substr( $line, 7);
$file_spec =~ s/^\s+//;		    # 7 char-state
$file_spec = ENV_perl_path_nospace( $file_spec);
$states{$file_spec} = $states_6;

}
foreach my $file_spec (@{$specs_ref})
{
my $states_6 = $states{$file_spec};
push @status_refs, [ $states_6, $file_spec ];

}
} else
{



foreach my $spec (@{$specs_ref})
{
push @status_refs, [ '      ', $spec ];
}
}


return @status_refs;
}

















sub exec_commit($)
{
my ($specs_ref) = @_;



my $command = "$EXEC_OS commit -q -N -m \"Checkin by $APP_NAME\"";
ENV_run3( $command, $specs_ref, [0], undef, undef);
}










sub exec_revert($)
{
my ($specs_ref) = @_;

my $command = "$EXEC_OS revert -q";
ENV_run3( $command, $specs_ref, [0], undef, undef);
}










sub exec_delete($)
{
my ($specs_ref) = @_;


my $command = "$EXEC_OS delete -q --force";
ENV_run3( $command, $specs_ref, [0], undef, undef);
}










sub exec_move($$)
{
my ($old_spec,
$new_spec,
) = @_;


my $command = "$EXEC_OS move -q";
ENV_run3( $command, [$old_spec, $new_spec], [0], undef, undef);
}










sub exec_add($)
{
my ($specs_ref) = @_;

my $command = "$EXEC_OS add -N";
ENV_run3( $command, $specs_ref, [0], undef, undef);
}










sub exec_propset_ignore($$$)
{
my ($dirs_ref,
$ignores_ref,
$must_recurse,
) = @_;

my @ignores = @{$ignores_ref};
my $command = "$EXEC_OS propset -q svn:ignore";


my $tmp_file = SPIT_tmp_file( 'svn_ignore', map { "$_\n" } @ignores);
$command .= " -F $tmp_file";
$command .= ' -R'
if ($must_recurse);

ENV_run3( $command, $dirs_ref, [0], undef, undef);
unlink $tmp_file;
}







sub exec_list($)
{
my ($repository) = @_;
my $rc;

my $command = "$EXEC_OS list \"$repository\"";
$rc = ENV_run3( $command, undef, [0, 1], undef, undef);

return $rc;
}




















































sub extract_state($$)
{
my ($state_6,	    #
$file_spec,
) = @_;
my $scm_state;	    # -1 = NotSCM, 0 = Checked_in, 1 = CheckedOut

my $state = substr( $state_6, 0, 1);
if (index( 'D?I', $state) >= 0)
{
$scm_state = -1;			# Not SCM
} elsif (index( 'AM', $state) >= 0)
{
$scm_state = 1;				# Checked Out;
} else
{
$scm_state = (-w $file_spec) ? 1 : 0;	# Checked Out : Checked In
}

return $scm_state;
}

return 1;

