#===============================================================
#
#   scmvss.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================
#   File states:
#    -2 => 'NO-FILE',
#    -1 => 'NOT-CCM',
#     0 => 'CHECKED-IN',
#     1 => 'CHECKED-OUT',
#===============================================================

package glo::scmvss;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
SCMVSS_init
SCMVSS_connect
SCMVSS_get_states
SCMVSS_checkout
SCMVSS_uncheckout
SCMVSS_checkin
SCMVSS_remove
SCMVSS_move
SCMVSS_add_dirs
SCMVSS_add_files
SCMVSS_setup
);

use strict;
use integer;

use File::Path;
use glo::env;
use glo::ask;

ENV_say( 1, "Loading VSS Plugin...");




sub SCMVSS_init($$);
sub SCMVSS_connect($$$);
sub SCMVSS_get_states($);
sub SCMVSS_checkout($);
sub SCMVSS_uncheckout($);
sub SCMVSS_checkin($);
sub SCMVSS_remove($);
sub SCMVSS_move($);
sub SCMVSS_add_dirs($);
sub SCMVSS_add_files($);
sub SCMVSS_setup($$$);

sub exec_about();
sub exec_status($);
sub exec_checkout($);
sub exec_checkin($);
sub exec_undocheckout($);
sub exec_destroy($);
sub exec_move($$);
sub exec_create($);
sub exec_add($);
sub exec_cp($);
sub vss_spec(@);














my $APP_NAME;
my $REPOSITORY_OS;
my $VSS_ROOT;
my $APP_ROOT_PATH;
my $WA_PATH;
my $WA_PATH_L;
my @SKIPTYPES = qw( \.scc$ );	# Keep in synch with %PREINFO_DATA in scm.pm

my $EXEC_OS = 'ss';

my %FSTATES;





sub SCMVSS_init($$)
{
($APP_NAME,
$EXEC_OS) = @_;
}





sub SCMVSS_connect($$$)
{
($REPOSITORY_OS,
$APP_ROOT_PATH,
$VSS_ROOT,
) = @_;	    # DATA
my $connect_texts_ref;

ENV_fatal( "Unable to connect",
"No such Repository '$REPOSITORY_OS'")
if (!-e $REPOSITORY_OS);

my $old_env = ENV_getenv( 'SSDIR');
if ($old_env ne $REPOSITORY_OS)
{
ENV_setenv( SSDIR => $REPOSITORY_OS);
ENV_say( 1, "'SSDIR' set to $REPOSITORY_OS");
}

$WA_PATH = ENV_parent_path( $APP_ROOT_PATH);
$WA_PATH_L = length $WA_PATH;
$VSS_ROOT .= '/' if ($VSS_ROOT ne '' && substr( $VSS_ROOT, -1, 1) ne '/');

$connect_texts_ref = exec_about();
%FSTATES = ();	    # clear previous file-states

return ( undef, undef, $connect_texts_ref );
}








sub SCMVSS_get_states($)
{
my ($spec_refs_ref) = @_;

my @specs;
foreach my $spec_ref (@{$spec_refs_ref})
{
my ($spec, $state) = @{$spec_ref};
if (!defined $state)	    # $state may be -2 (no such file/dir)
{
$state = $FSTATES{$spec};
if (defined $state)
{
$spec_ref->[1] = $state;
} else
{
push @specs, $spec;
}
}
}

if (@specs)
{
exec_status( \@specs);
foreach my $spec_ref (@{$spec_refs_ref})
{
my ($spec, $state) = @{$spec_ref};
$spec_ref->[1] = $FSTATES{$spec}
if (!defined $state);
}
}
}




sub SCMVSS_checkout($)
{
my ($specs_ref) = @_;

ENV_pushd( undef);
foreach my $spec (@{$specs_ref})
{
if (-d $spec)
{
ENV_say( 1, "No CheckOut for directories",
"- $spec\n");
} else
{
exec_checkout( $spec);	    # changes cwd
}
}
ENV_popd();

map { $FSTATES{$_} = 1 } @{$specs_ref}; # checkout
}




sub SCMVSS_checkin($)
{
my ($specs_ref) = @_;

ENV_pushd( undef);
foreach my $spec (@{$specs_ref})
{
if (-d $spec)
{
ENV_say( 1, "No CheckIn for directories",
"- $spec\n");
} else
{
exec_checkin( $spec);
}
}
ENV_popd();

map { $FSTATES{$_} = 0 } @{$specs_ref}; # checkin
}




sub SCMVSS_uncheckout($)
{
my ($specs_ref) = @_;

ENV_pushd( undef);
foreach my $spec (@{$specs_ref})
{
if (-d $spec)
{
ENV_say( 1, "No UndoCheckOut for directories",
"- $spec\n");
} else
{
exec_undocheckout( $spec);
}
}
ENV_popd();

map { $FSTATES{$_} = 0 } @{$specs_ref}; # checkin
}




sub SCMVSS_remove($)
{
my ($specs_ref) = @_;

foreach my $spec (@{$specs_ref})
{
exec_destroy( $spec);
rmtree( $spec, 0, 1);
}
map { $FSTATES{$_} = -1 } @{$specs_ref}; # notscm
}




sub SCMVSS_move($)
{
my ($pair_refs_ref) = @_;

foreach my $ref (@{$pair_refs_ref})
{
my ($old_spec, $new_spec) = @{$ref};
ENV_chmod( 'u+w', $old_spec);
exec_move( $old_spec, $new_spec);
$FSTATES{$old_spec} = -1;	# notscm
$FSTATES{$new_spec} = 1;    	# checkout
}
}




sub SCMVSS_add_dirs($)
{
my ($dir_refs_ref) = @_;	#   [ $dir, $ignores_ref ]

my @dir_specs;
foreach my $ref (@{$dir_refs_ref})
{
my ($dir, undef) = @{$ref};
ENV_mkdir( $dir, 0777)
if (!-d $dir);
exec_create( $dir);
push @dir_specs, $dir;
}

map { $FSTATES{$_} = 1 } @dir_specs; # checkout
}




sub SCMVSS_add_files($)
{
my ($files_ref) = @_;

ENV_pushd( undef);
foreach my $file (@{$files_ref})
{
ENV_touch_file( $file)
if (!-f $file);
exec_add( $file);
}
ENV_popd();

map { $FSTATES{$_} = 1 } @{$files_ref}; # checkout
}




sub SCMVSS_setup($$$)
{
my (undef,				    # $app_root_path
$default_scms_repository_path,
$default_scms_data,
) = @_;
my ($scms_repository_path, $scms_data);

ENV_say( 1, "VSS Layout:");
my @texts = (
'    REPOSITORY:---VSS_PATH--+              +----',
'                             \            /     ',
'    DISK:----------WA_PATH----+--SYSTEM--+------',
'                                          \     ',
"    DISK:---------${APP_NAME}_ROOT-------------->  +----",
);
ENV_say( 0, @texts);

my $vss_path = (defined $default_scms_data) ? $default_scms_data : '';
$vss_path = ASK_dir( 'Enter possible VSS_ROOT', $vss_path, '', 0, 0);

if (!$default_scms_repository_path)
{
$default_scms_repository_path = ENV_getenv( 'SSDIR');
ENV_say( 1, "SSDIR = $default_scms_repository_path");
$default_scms_repository_path = '' if (! -d $default_scms_repository_path);
}
$scms_repository_path = ASK_path( 'Enter Repository (SSDIR)', $default_scms_repository_path, 1, -1);
$scms_data = $vss_path;

return ( $scms_repository_path, $scms_data );
}













sub exec_about()
{
my @connect_texts;

my $stdout;
my $command = "$EXEC_OS about";
ENV_run3( $command, undef, [0], \$stdout, undef);
if ($stdout !~ /Microsoft \(R\) Visual SourceSafe/)
{
print $stdout;
ENV_fatal( "Failed to start VSS session");
}
@connect_texts = (split( '\n', $stdout))[0..1];

return [ @connect_texts ];
}



























sub exec_status($)
{
my ($specs_ref) = @_;

my $file_count = 0;
my $stdout;
my $stderr;
foreach my $spec (@{$specs_ref})
{
my $vss_spec = vss_spec( $spec);
my $command = "$EXEC_OS Status $vss_spec -U";
my $rc = ENV_run3( $command, undef, [0, 1, 100], \$stdout, \$stderr);



my $state;
if (-d $spec)
{
if ($rc == 100)
{
$state = -1;	    # not scm
} else # rc == 0 or 1
{
$state = 1;	    # assume directories are always checkedout
}
} else
{
if ($rc == 0)
{
$state = 0;	    # checkedin
} elsif ($rc == 1)
{
$state = 1;	    # checkedout
} else  #rc == 100
{
$state = -1;	    # not scm
}
}
$FSTATES{$spec} = $state;

$file_count++;
}

if ($file_count != @{$specs_ref})
{
ENV_fatal( "Unexpected return text with 'vss status ...'",
"STDOUT: $stdout",
"STDERR: $stderr");
}


}







sub exec_checkout($)
{
my ($spec) = @_;

my $dir = ENV_split_spec_d( $spec);

ENV_chdir( $dir);
my $vss_spec = vss_spec( $spec);
my $command = "$EXEC_OS Checkout $vss_spec \"-CCheckout by $APP_NAME\"";
ENV_run3( $command, undef, [0], undef, '');
}







sub exec_checkin($)
{
my ($spec) = @_;

my $dir = ENV_split_spec_d( $spec);

ENV_chdir( $dir);
my $vss_spec = vss_spec( $spec);
my $command = "$EXEC_OS Checkin $vss_spec -P \"-CCheckin by $APP_NAME\"";
ENV_run3( $command, undef, [0], undef, '');
}






sub exec_undocheckout($)
{
my ($spec) = @_;

my $dir = ENV_split_spec_d( $spec);

ENV_chdir( $dir);
my $vss_spec = vss_spec( $spec);
my $command = "$EXEC_OS UndoCheckout $vss_spec";
ENV_run3( $command, undef, [0], undef, '');
}






sub exec_destroy($)
{
my ($spec) = @_;

my $vss_spec = vss_spec( $spec);
my $command = "$EXEC_OS Destroy $vss_spec -S -I-";
ENV_run3( $command, undef, [0], undef, '');
}






sub exec_move($$)
{
my ($old_spec,
$new_spec,
) = @_;

ENV_fatal( "MOVE not implemented!");    # TBS
}






sub exec_create($)
{
my ($spec) = @_;

my $vss_spec = vss_spec( $spec);
my $command = "$EXEC_OS Create $vss_spec \"-CCreated by $APP_NAME\" -I-";
ENV_run3( $command, undef, [0], undef, '');
}






sub exec_add($)
{
my ($spec) = @_;

exec_cp( $spec);

my $file = ENV_split_spec_f( $spec);
my $command = "$EXEC_OS Add $file -K \"-CCreated by $APP_NAME\"";
ENV_run3( $command, undef, [0], undef, '');
}






sub exec_cp($)
{
my ($spec) = @_;

my $dir = ENV_split_spec_d( $spec);
ENV_chdir( $dir);
my $vss_spec = vss_spec( $dir);
my $command = "$EXEC_OS Cp $vss_spec";
ENV_run3( $command, undef, [0], undef, undef);
}




sub vss_spec(@)
{

my @vss_specs;

foreach my $spec (@_)
{
my $vss_spec;
if ($spec eq $WA_PATH)
{
$vss_spec = '';
} else
{
$vss_spec = $VSS_ROOT . substr( $spec, $WA_PATH_L + 1);
}

push @vss_specs, "\$/$vss_spec";
}

return (wantarray) ? @vss_specs : $vss_specs[0];
}
return 1;

