#===============================================================
#
#   mm.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package mm;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
MM_makemake
);

use strict;
use integer;

use glo::env;
use glo::path;
use glo::depend;
use gbsenv;
use gbsglo;
use gbsfileglo;
use flincs;
use glkb;
use target;
use fspec;




sub MM_makemake(;@);

sub makecompmake($);
sub get_file_dependencies();
sub depend_glkb();
sub get_gbs_dependencies();
sub print_rule($$$$$);
sub heading($@);
sub set_includes();
sub find_file_rel($);
sub warning($$@);
sub error($$@);




my $GBS_SCRIPTS_PATH = GBSENV_scripts_path();
my $GBS_ROOT_PATH = GBSENV_gbs_root_path();;
my $GBS_SYSBUILD_PATH = "$GBS_ROOT_PATH/sysbuild";
my $GBS_SUBSYS = ENV_getenv( 'GBS_SUBSYS');
my $GBS_SUBSYS_PATH = "$GBS_ROOT_PATH/dev/$GBS_SUBSYS";
my $GBS_COMP_PATH = "$GBS_ROOT_PATH/dev/$GBS_SUBSYS/comp";




my $GBS_TARGET; 	#do not initialize!
my @STUBS_PATHS;

my %SRCS;

my %BLDS;


my %SRC_TYPES_DATA;










my $COMPONENT;
my $COMPONENT_PATH;
my $CWD;
my %GBS_DEPEND;


my %GBSFILES;	    # cache for processed gbsfile-contents


my %ABS_INCPATH_CACHE;  # cache for abs include-paths (used by DEPEND_.. functions)


my %INCPATHS_CACHE;	    # cache for include-paths




my %FOUND_FILES;     # cache for found includes






my @INCPATH_REFS;	# from %INCPATHS_CACHE: contains the include-path for the current file [ type, name ]
my @ABS_INCPATHS;	# from %ABS_INCPATH_CACHE: contains the include-path for the current file

my $SRC_FILE;		# name.type
my $SRC_TYPE;		# .type
my $SRC_GEN_TYPE;	# gen | asm | c | glb | glk | glt
my $SRC_FILE_SPEC;	# $CWD/$SRC_FILE
my $SRC_SHORT_FILE_SPEC;

my %EXTERNALS;


my %UNRESOLVED;





my $ERROR_COUNT;
my $WARNING_COUNT;




my $COMMENT_RE_ASM = qr!^\s*[;*/#].*!;
my $INCLUDE_RE_ASM = qr!^\s*#\s*include\s+([<"]?)(.+)[>"]?!;
my %DEFAULT_RE = (

asm =>  [ $INCLUDE_RE_ASM, $COMMENT_RE_ASM],
c	=>  [ '',	       '' ],
gen =>  [ '',	       '' ],
glb =>  [ '',	       '' ],
glk =>  [ '',	       '' ],
glt =>  [ '',	       '' ],
);




sub MM_makemake(;@)
{
my @components = @_;
my $rc = 0;

$GBS_TARGET = ENV_getenv( 'GBS_TARGET');
%SRCS = ();
%BLDS = ();
$CWD = ENV_cwd();
$ERROR_COUNT = 0;
$WARNING_COUNT = 0;
%SRC_TYPES_DATA = ();

@STUBS_PATHS = FLINCS_get_stubs_incs();

TARGET_set_envs( $GBS_TARGET); # set all GBS_BLD_...

foreach my $src_type (TARGET_get_src_types( $GBS_TARGET))
{
my ($type, $include_re, $comment_re) = TARGET_get_src_items( $GBS_TARGET, $src_type, 'TYPE');
$include_re = $DEFAULT_RE{$type}->[0]
if ($include_re eq '');
$comment_re = $DEFAULT_RE{$type}->[1]
if ($comment_re eq '');
$SRC_TYPES_DATA{$src_type}->{TYPE} = $type;
$SRC_TYPES_DATA{$src_type}->{INCLUDE_RE} = $include_re;
$SRC_TYPES_DATA{$src_type}->{COMMENT_RE} = $comment_re;
$SRC_TYPES_DATA{$src_type}->{OUT_TYPES} = TARGET_get_src_items( $GBS_TARGET, $src_type, 'OUT_TYPES');
$SRC_TYPES_DATA{$src_type}->{OUT_FILES} = TARGET_get_src_items( $GBS_TARGET, $src_type, 'OUT_FILES');
$SRC_TYPES_DATA{$src_type}->{UNIX_TYPE_SEARCH} =
(TARGET_get_src_items( $GBS_TARGET, $src_type, 'INC_SEARCH_STYLE') eq 'UNIX') ? 1 : 0;
}

my @allcomp = GBSGLO_components_for_target( $GBS_SUBSYS, $GBS_TARGET);
my %selcomp;
if (@components)
{
%selcomp = map { $_ => 1 } @components;
ENV_say( 1, "MakeMake SubSystem '${GBS_SUBSYS}::(@components)' for Target '$GBS_TARGET'...");
} else
{
%selcomp = map { $_ => 1 } @allcomp;
ENV_say( 1, "MakeMake SubSystem '$GBS_SUBSYS' for Target '$GBS_TARGET'...");
}

if (%selcomp)
{



my $saved_cwd = $CWD;
foreach my $component (@allcomp)
{
$COMPONENT = $component;
$COMPONENT_PATH = "$GBS_COMP_PATH/$COMPONENT";
$CWD = ENV_chdir( "$COMPONENT_PATH/src");
my @all_src_files = sort( glob( '*.*'));

my @src_file_refs;
foreach my $file (GBSGLO_select_specs( @all_src_files))
{

if ($file =~ /\s/)
{
ENV_warn( "File: '$COMPONENT:$file': No spaces in filename allowed. File Skipped");
next;
}
my ($file_name, $file_type) = ENV_split_spec_nt( $file);
$SRC_FILE = "$file_name$file_type";
$SRC_TYPE = $file_type;
$SRC_GEN_TYPE = undef;
$SRC_FILE_SPEC = "$CWD/$SRC_FILE";
$SRC_SHORT_FILE_SPEC = GBSGLO_short_filespecs( $SRC_FILE_SPEC);
if (exists $SRC_TYPES_DATA{$file_type})
{
push @src_file_refs, [ $file_name, $file_type ];
foreach my $out_type (@{$SRC_TYPES_DATA{$file_type}->{OUT_TYPES}})
{
my $bld_file = $file_name . $out_type;
$BLDS{$COMPONENT}->{$bld_file} = 1;

}
foreach my $bld_file (@{$SRC_TYPES_DATA{$file_type}->{OUT_FILES}})
{
$BLDS{$COMPONENT}->{$bld_file} = 1;

}
} else
{
error( undef, undef,
"Unknown file-type ($file_type) - Skipped");
}
}
$SRCS{$component} = [ @src_file_refs ];
}





my @allgencomp;
my @compmakefiles;
foreach my $component (@allcomp)
{
$COMPONENT = $component;





my $compmakefile = "$GBS_COMP_PATH/$COMPONENT/bld/$GBS_TARGET.mk";
if (exists $selcomp{$COMPONENT} || ! -f $compmakefile)
{
$COMPONENT_PATH = "$GBS_COMP_PATH/$COMPONENT";
$CWD = ENV_chdir( "$COMPONENT_PATH/src");
makecompmake( $compmakefile);
}
push @allgencomp, $COMPONENT;
push @compmakefiles, $compmakefile;
}
ENV_chdir( $saved_cwd)
if ($CWD ne $saved_cwd);




my $makefile = FSPEC_make_file( "$GBS_SUBSYS_PATH/build", $GBS_TARGET);
open( my $gen_out, ">$makefile") || ENV_fatal( "Unable to create $makefile", "- $!");




my $now = localtime();
heading( $gen_out,
"File:\t$makefile",
"SubSys:\t$GBS_SUBSYS",
"Generated:\t$now", '');
print( $gen_out "\n");




map { print( $gen_out "include $_\n") } @compmakefiles;




print ($gen_out "\n");
heading( $gen_out, "All components");
print_rule( $gen_out, [ 'ALL' ], '', \@allgencomp, 0);




print ($gen_out "\n");
heading( $gen_out, "Dummy Make-Target");
print_rule( $gen_out, [ 'ALLWAYS' ], '', [], 0);




print( $gen_out "\n### EOF ###\n");
close( $gen_out) || ENV_fatal( "Unable to close $gen_out", "- $!");
} else
{
ENV_warn( "No components for this Target - Nothing generated");
}

print "\n"
if (%UNRESOLVED || %EXTERNALS);

if (%UNRESOLVED)
{
ENV_say( 1, "Unresolved:");
foreach my $file (sort( keys %UNRESOLVED))
{
my $count = $UNRESOLVED{$file};
print "    $file: $count ref(s)\n";
}
}
if (%EXTERNALS)
{
ENV_say( 1, "Unresolved externals:");
foreach my $file (sort( keys %EXTERNALS))
{
my $count = $EXTERNALS{$file};
print "    $file: $count ref(s)\n";
}
}
if ($WARNING_COUNT > 0)
{
ENV_say( 1, "Nr of Warnings: $WARNING_COUNT");
}
if ($ERROR_COUNT > 0)
{
ENV_say( 1, "Nr of Errors: $ERROR_COUNT");
$rc = 1 if ($rc < 1);
}

return $rc;
}




sub makecompmake($)
{
my ($compmakefile) = @_;

ENV_say( 1, "MakeCompMake '$COMPONENT'...");
my $src_path = "$COMPONENT_PATH/src";
my $bld_path = "$COMPONENT_PATH/bld/$GBS_TARGET";




%GBS_DEPEND = ();
%GBSFILES = ();
%INCPATHS_CACHE = ();
%ABS_INCPATH_CACHE = ();
%FOUND_FILES = ();




my $reffile = "$COMPONENT_PATH/bld/$GBS_TARGET.ref";
open( my $out_mk, ">$compmakefile") || ENV_fatal( "Unable to create $compmakefile", "- $!");
open( my $out_xref, ">$reffile") || ENV_fatal( "Unable to create $reffile", "- $!");
my $now = localtime();
heading( $out_mk,
"File:\t$GBS_TARGET.mk",
"Component:\t$COMPONENT",
"SubSys:\t$GBS_SUBSYS",
"Generated:\t$now", '');

my %dup_mk_targets;
my @all_out_files;
foreach my $src_file_ref (@{$SRCS{$COMPONENT}})
{
my ($file_name, $file_type) = @{$src_file_ref};
$SRC_FILE = "$file_name$file_type";
$SRC_TYPE = $file_type;
$SRC_GEN_TYPE = $SRC_TYPES_DATA{$SRC_TYPE}->{TYPE};
$SRC_FILE_SPEC = "$CWD/$SRC_FILE";
$SRC_SHORT_FILE_SPEC = GBSGLO_short_filespecs( $SRC_FILE_SPEC);




my @out_types = @{$SRC_TYPES_DATA{$SRC_TYPE}->{OUT_TYPES}};
my @out_files = @{$SRC_TYPES_DATA{$SRC_TYPE}->{OUT_FILES}};
my $out_type = $out_types[0];





print( $out_mk "\n");
my @mk_target_names = map { "$file_name$_" } @out_types;
push @mk_target_names, @out_files;
foreach my $mk_target_name (@mk_target_names)
{
if (exists $dup_mk_targets{$mk_target_name})
{
error( undef, undef,
"Duplicate Make-Target: $mk_target_name");
} else
{
$dup_mk_targets{$mk_target_name} = 1;
}
}
heading( $out_mk, "$SRC_FILE -> @mk_target_names");




my $gen_command = "gbsgen $COMPONENT:$SRC_FILE";




my @depend_line_refs = get_file_dependencies();





my @depend_lines = map { $_->[1] } @depend_line_refs;	    # $file




@mk_target_names = map { "$bld_path/$_" } @mk_target_names;
print_rule( $out_mk, [ @mk_target_names ], $gen_command, \@depend_lines, 1);
push @all_out_files, "$bld_path/$file_name$out_type";




foreach my $ref (@depend_line_refs)
{
my ($inc_level, $file) = @{$ref};
next if ($inc_level < 0);			# comment
print( $out_xref "$inc_level $file\n");
}
}

print( $out_mk "\n");
heading( $out_mk, "All files for component $COMPONENT");
print_rule( $out_mk, [ $COMPONENT ], '', \@all_out_files, 0);

print( $out_mk "\n### EOF ###\n");
close( $out_xref) || ENV_fatal( "Unable to close $out_xref", "- $!");
close( $out_mk) || ENV_fatal( "Unable to close $out_mk", "- $!");

return;
}





sub get_file_dependencies()
{
my ($file) = @_;
my @file_refs;





set_includes();	# @INCPATH_REFS & @ABS_INCPATHS

push @file_refs, [ 0, $SRC_FILE_SPEC ];
my ($nr_errors, $nr_warnings, @depend_refs);
if (TARGET_is_glkb_type( $SRC_GEN_TYPE))	# glb glk glt
{
($nr_errors, $nr_warnings, @depend_refs) = depend_glkb();
} elsif ($SRC_GEN_TYPE eq 'c')		# c
{
my $is_unix_type_search = $SRC_TYPES_DATA{$SRC_TYPE}->{UNIX_TYPE_SEARCH};
($nr_errors, $nr_warnings, @depend_refs) = DEPEND_c( $SRC_FILE_SPEC, $is_unix_type_search,
\@ABS_INCPATHS, \@STUBS_PATHS, undef);
} else					# gen asm
{
my $comment_re = $SRC_TYPES_DATA{$SRC_TYPE}->{COMMENT_RE};
my $include_re = $SRC_TYPES_DATA{$SRC_TYPE}->{INCLUDE_RE};
($nr_errors, $nr_warnings, @depend_refs) = DEPEND_line( $comment_re, $include_re, $SRC_FILE_SPEC, \@ABS_INCPATHS, \@STUBS_PATHS, undef);
}

if ($nr_errors > 0)
{
ENV_say( 1, "$nr_errors error(s) found while parsing '$COMPONENT:$SRC_FILE'");
push @file_refs, [ 1, 'ALLWAYS' ];
push @file_refs, [ -3, "# ALLWAYS: $SRC_FILE - parser found $nr_errors error(s)" ];
$ERROR_COUNT += $nr_errors;
}

if ($nr_warnings > 0)
{
foreach my $ref (@depend_refs)
{
my ($level, $found, $line_nr, $inc_type, $inc_name, $inc_file_spec) = @{$ref};

if ($found eq 'N')
{
($found, $inc_file_spec) = find_file_rel( $inc_name);
if ($found ne 'N')
{

$ref->[1] = $found;
$ref->[5] = $inc_file_spec;
$nr_warnings--;
}
}
}
ENV_say( 1, "$nr_warnings warning(s) found while processing '$COMPONENT:$SRC_FILE'")
if ($nr_warnings > 0);
}

foreach my $ref (@depend_refs)
{
my ($level, $found, $line_nr, $inc_type, $inc_name, $inc_file_spec) = @{$ref};

if ($found eq 'F')
{


push @file_refs, [ $level, $inc_file_spec ];
} elsif ($found eq 'S')
{

my $short_inc_file_spec = GBSGLO_short_filespecs( $inc_file_spec);
push @file_refs, [ -1, "# $inc_name found in $short_inc_file_spec" ];
} elsif ($found eq 'N')
{

my $calling_file_spec = $inc_file_spec;
my $short_calling_file_spec = GBSGLO_short_filespecs( $calling_file_spec);






my $go_soft = ((!ENV_is_in_path( $calling_file_spec, $GBS_ROOT_PATH)) ||
ENV_is_in_path( $calling_file_spec, "$GBS_ROOT_PATH/ext"));
if ($go_soft)
{
if (exists $EXTERNALS{$inc_name})
{
$EXTERNALS{$inc_name}++;
} else
{
$EXTERNALS{$inc_name} = 1;
warning( $calling_file_spec, $line_nr,
"gen: Unable to locate '$inc_name' (external)",
'File skipped',
'Further messages suppressed');
}

push @file_refs, [ -1, "# Unable to locate (external) '$inc_name' from $short_calling_file_spec($line_nr)" ];
} else
{
$UNRESOLVED{$inc_name}++;
warning( $calling_file_spec, $line_nr,
"gen: Unable to locate '$inc_name'",
'Possibly a conditional or generated file',
'Source will regenerate ALLWAYS');
push @file_refs, [ 1, 'ALLWAYS' ];

push @file_refs, [ -2, "# ALLWAYS: Unable to locate: '$inc_name' from $short_calling_file_spec($line_nr)" ];
}
} else
{
ENV_fatal( "Invalid 'found' ($found'");
}
}

return @file_refs;
}




sub depend_glkb()
{
my ($nr_errors, $nr_warnings, @depend_refs);







GLKB_parse( $SRC_FILE, $GBS_TARGET, $COMPONENT);
( $nr_errors, $nr_warnings, my @in_line_refs ) = GLKB_get_dependencies();




foreach my $ref (@in_line_refs)
{
my ($inc_level, $file, $org_file, $line_ref) = @{$ref};

my (undef, $inc_file, $line_nr, $line) = GBSFILEGLO_get_line_info( $line_ref);


my ($found, $spec);
if (ENV_is_abs_path_perl( $file))	    # ABS path
{




$found = 'F';
$spec = $file;
} elsif ($file !~ m!/!) 		    # pure file (no dirspec)
{





($found, $spec) = find_file_rel( $file);
} else					    # other relative path
{



print "=====>SEARCH $file\n";
my $found_file = '';
foreach my $inc_path_ref (@INCPATH_REFS)
{
if ($inc_path_ref->[0] eq 'P')
{
my $dir = $inc_path_ref->[1];

my $search_file = "$dir/$file";
if (PATH_file_exists( $search_file))
{
$found_file = $search_file;
last;
}
}
}
if ($found_file ne '')
{
$found = 'F';
$spec = $found_file;
} else
{



if ($file =~ m!^\.\.?/!)	# starts with ./ or ../
{
$found = 'F';
$spec = ENV_canonicalize_path( "$CWD/$file");
} else
{
$found = 'N';
$spec = $inc_file;
}
}
}

push @depend_refs, [ $inc_level, $found, $line_nr, '', $org_file, $spec ];
}

return ($nr_errors, $nr_warnings, @depend_refs);
}




sub get_gbs_dependencies()
{


if (!exists $GBS_DEPEND{$SRC_TYPE})
{
my @files;




{
my $key = "F$SRC_TYPE";
if (!exists $GBSFILES{$key})
{
my @flag_files = ( FLINCS_get_flags_files( $SRC_TYPE, $COMPONENT),
TARGET_get_bld_flags_file( $GBS_TARGET, $SRC_TYPE));
@flag_files = grep( PATH_file_exists( $_), ENV_perl_path_nospace( @flag_files));
$GBSFILES{$key} = [ @flag_files ];
}
push @files, @{$GBSFILES{$key}};
}




{
my $key = "I$SRC_TYPE";
if (!exists $GBSFILES{$key})
{
my @inc_files = ( FLINCS_get_incs_files( $SRC_TYPE, $COMPONENT),
TARGET_get_bld_incs_file( $GBS_TARGET, $SRC_TYPE));
@inc_files = grep( PATH_file_exists( $_), @inc_files);
$GBSFILES{$key} = [ @inc_files ];
}
push @files, @{$GBSFILES{$key}};
}




foreach my $file ( "$COMPONENT_PATH/scope",
"$GBS_SYSBUILD_PATH/sysbuild")
{
my $gbs_or_usr_file = GBSFILEGLO_select_usr_gbs( $file);
push @files, $gbs_or_usr_file
if (defined $gbs_or_usr_file);
}
push @files, TARGET_get_dependencies( $GBS_TARGET, $SRC_TYPE);	    # target.gbs and command_file(s)

$GBS_DEPEND{$SRC_TYPE} = [ @files ];
}

return @{$GBS_DEPEND{$SRC_TYPE}};
}




sub print_rule($$$$$)
{
my ($out,
$make_targets_ref,
$gen_command,
$dependencies_ref,
$must_add_gbs_dependencies,
) = @_;




foreach my $make_target (@{$make_targets_ref})
{
$make_target = ENV_os_path( $make_target);
$make_target = "\"$make_target\"" if (index( $make_target, ' ') >= 0);
}
my $make_targets = join " \\\n", @{$make_targets_ref};
print( $out "$make_targets :");




my @comments;
foreach my $depend (@{$dependencies_ref})
{
if (substr($depend,0,1) eq '#')
{
push @comments, $depend;
} else
{
if (ENV_is_win32())
{
$depend = ENV_os_path( $depend);
$depend = "\"$depend\"" if (index( $depend, ' ') >= 0);
}
print( $out " \\\n	 $depend");
}
}
if ($must_add_gbs_dependencies)
{
foreach my $depend (get_gbs_dependencies())
{
if (ENV_is_win32())
{
$depend = ENV_os_path( $depend);
$depend = "\"$depend\"" if (index( $depend, ' ') >= 0);
}
print( $out " \\\n	 $depend");
}
}
print( $out "\n");




print( $out "\t$gen_command\n")
if ($gen_command ne '');




map { print( $out "$_\n"); } @comments;
}




sub heading($@)
{
my ($out,
@items
) = @_;

print( $out "#=========================================================\n");
foreach my $item (@items)
{
if ($item eq '')
{
print( $out "#\n");
} else
{
print( $out "#   $item\n");
}
}
print( $out "#=========================================================\n");
}





sub set_includes()
{
my $key = "$COMPONENT:$SRC_TYPE";

if (exists $INCPATHS_CACHE{$key})
{
@INCPATH_REFS = @{ $INCPATHS_CACHE{$key} };
@ABS_INCPATHS = @{ $ABS_INCPATH_CACHE{$key} };

} else
{

@ABS_INCPATHS = ENV_abs_path( $CWD, [ FLINCS_get_incs( $SRC_TYPE, $COMPONENT),   # inc
FLINCS_get_bld_incs( $SRC_TYPE)]);	 # sysinc
$ABS_INCPATH_CACHE{$key} = [ @ABS_INCPATHS ];

@INCPATH_REFS = ();
foreach my $dir (@ABS_INCPATHS)
{
if ($dir =~ m!$GBS_COMP_PATH/([^/]+)/bld/$GBS_TARGET!)
{
push @INCPATH_REFS, [ 'C', $1 ];
} else
{

push @INCPATH_REFS, [ 'P', $dir ];
}
}

$INCPATHS_CACHE{$key} = [ @INCPATH_REFS ];
}
}








sub find_file_rel($)
{
my ($file) = @_;
my ($found, $spec);


my $file_key = "$SRC_TYPE>$file";
my $found_ref = $FOUND_FILES{$file_key};
if (defined $found_ref)
{
($found, $spec) = @{$found_ref};
} else
{



my $found_file = '';
foreach my $inc_path_ref (@INCPATH_REFS)
{
if ($inc_path_ref->[0] eq 'C')
{
my $component = $inc_path_ref->[1];

if (exists $BLDS{$component}->{$file})
{
$found_file = "$GBS_COMP_PATH/$component/bld/$GBS_TARGET/$file";

last;
}
} else	# $inc_path_ref->[0] eq 'P'
{
my $dir = $inc_path_ref->[1];

my $search_file = "$dir/$file";
if (PATH_file_exists( $search_file))
{
$found_file = $search_file;
last;
}
}
}
if ($found_file ne '')
{
$found = 'F';
$spec = $found_file;
} else
{



foreach my $stub_location (@STUBS_PATHS)
{
if (PATH_file_exists( "$stub_location/$file"))
{
my $short_spec = GBSGLO_short_filespecs( "$stub_location/$file");
$found_file = $short_spec;
last;
}
}
if ($found_file ne '')
{
$found = 'S';
$spec = $found_file;
} else
{
$found = 'N';
$spec = '';
}
}
$FOUND_FILES{$file_key} = [ $found, $spec ];
}

return ($found, $spec);
}




sub warning($$@)
{
my ($file,		# may be undef
$line_nr,	# may be undef
@texts
) = @_;

$WARNING_COUNT++;
$file = GBSGLO_short_filespecs( $file)
if (defined $file);
ENV_warn( [ undef, $SRC_SHORT_FILE_SPEC, $file, $line_nr, '' ], [ @texts ] );
}




sub error($$@)
{
my ($file,		# may be undef
$line_nr,	# may be undef
@texts
) = @_;

$ERROR_COUNT++;
$file = GBSGLO_short_filespecs( $file)
if (defined $file);
ENV_error( [ undef, $SRC_SHORT_FILE_SPEC, $file, $line_nr, '' ], [ @texts ] );
}

return 1;


