#===============================================================
#
#   gbstest.pl
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================
#
#  glt: .glt -> .log
#      arg_1   = file_name     (without type)
#      arg_2   = in_file_spec  (relative file spec)
#      arg_3   = out_file_spec (relative file spec)
#      arg_4   = out_file_dir  (out_file_spec directory only)
#      arg_5   = out_file_type (with dot)
#      arg_6-N = command_flags (from command line)
#  Environment:
#      General:
#          GBS_INCS            libraries
#          GBS_GLKB       files to include
#          GBS_MAP             NO / YES
#          GBS_DEBUGGER        NO / YES
#      Application:
#          <none>
#============================================================================

print "*GBSTEST*\n" if ($ENV{GBSDEBUG_FILE});

use strict;
use integer;


use lib $ENV{GBS_SCRIPTS_PATH};
use glo::env;
use glo::args;
use glo::spit;
use glo::slurp;
use gbsenv;




sub get_command_arguments();
sub handle_glt_options( @);
sub glt_error_exit($@);




my $PREFIX = 'GBSTEST';

my $GBS_TARGET = ENV_getenv( 'GBS_TARGET');




my $FILE_NAME;		# (without type)
my $IN_FILE_SPEC;	# (relative file spec)
my $OUT_FILE_SPEC;	# (relative file spec)
my $OUT_FILE_DIR;	# (out_file_spec directory only)
my $OUT_FILE_TYPE;	# (with dot)
my @COMMAND_FLAGS;	# (from command line)

my @GBS_GLKB;
my $GBS_IGNORE_ERRORS;








my $TIMEOUT = 1 * 60;	    # 10 minute. 0 == no-timeout






$| = 1;
my $RC = 0;
GBSENV_init();

get_command_arguments();










{
my @glt_options;
my $call = '';		    # '' or 'call '
my $exec_file;
my @args;




if (@GBS_GLKB)
{





while ( @GBS_GLKB && $GBS_GLKB[0] =~ /^glt_/)
{
push @glt_options, shift @GBS_GLKB;
}




$call = shift @GBS_GLKB;
if (defined $call)
{
if (lc $call eq 'call')
{
$exec_file = shift @GBS_GLKB;
$call .= ' ';
} else
{
$exec_file = $call;
$call = '';
}
}




@args = @GBS_GLKB;
}

handle_glt_options( @glt_options);

if (defined $exec_file)
{
$exec_file = ENV_os_path( $exec_file);
my $command_line = "$call$exec_file";

glt_error_exit( $command_line,  "'$exec_file' does not exist")
if (!-e $exec_file);

$command_line = "$command_line @args"
if (@args);

$command_line = "$command_line @COMMAND_FLAGS"
if (@COMMAND_FLAGS);

SPIT_file( $OUT_FILE_SPEC, "$PREFIX: Executing test '$command_line'\n");
my $start = time();
$RC = ENV_system( $command_line, undef, ">$OUT_FILE_SPEC");
my $end = time();
my $elapsed = $end - $start + 1;

my @log_lines = SLURP_file_raw( $OUT_FILE_SPEC);
if (@log_lines <= 4)
{
ENV_print( 0, \@log_lines);
} else
{
ENV_print( 0, @log_lines[0..1]);
ENV_say( 0, '...');
ENV_print( 0, $log_lines[-1]);
}
my @trails;
push @trails, "$PREFIX: Failed ($RC)\n"
if ($RC > 0);
push @trails,  "$PREFIX: Execution took less than $elapsed seconds\n";
ENV_print( 0, @trails);			    # write to screen
SPIT_file( ">$OUT_FILE_SPEC", @trails);	    # append to logfile
} else
{
ENV_error( "No Executable found in parsed GLT-file (GBS_GLKB EnvVar for this Target)",
"- Contents: '@GBS_GLKB'");
}
}




ENV_exit( $RC);




sub END
{
ENV_print_end_msg(1);
}




sub handle_glt_options(@)
{
my (@glt_options) = @_;



my @keywords = qw( timeout );
my %keywords = map { $_ => 0 } @keywords;

foreach my $line (@glt_options)
{
my ($key, $op, $value) = $line =~ /^glt_(\w*)(=|!=)(.+)$/;
glt_error_exit( $line, "glt syntax is '<keyword><=|!=><value|value_comma_list>'")
if (!defined $key);


my $count = $keywords{$key};
glt_error_exit( $line, "glt keyword must be one of (@keywords)")
if (!defined $count);
glt_error_exit( $line, "glt_$key may only be specified once")
if ($count > 0);
$keywords{$key}++;

if ($key eq 'timeout')
{
glt_error_exit( $line, "glt_maxtime only allows '=' as operator")
if ($op ne '=');
glt_error_exit( $line, "glt_maxtime value must be decimal (time in seconds)")
if ($value !~ /^\d+$/);
$TIMEOUT = $value;

glt_error_exit( $line, 'glt_maxtime not implemented yet!');
ENV_say( 1, "glk_maxtime set to $TIMEOUT (seconds)");
} else
{
ENV_fatal( "Impossible else");
}
}
}




sub glt_error_exit($@)
{
my ($glt_line,
@text_lines
) = @_;

ENV_say( 1, "GLT: $glt_line");
ENV_error_exit( @text_lines);
}




sub get_command_arguments()
{

($FILE_NAME,
$IN_FILE_SPEC,
$OUT_FILE_SPEC,
$OUT_FILE_DIR,
$OUT_FILE_TYPE,
@COMMAND_FLAGS) = ARGS_get( [5],
'perl gbstest.pl file_name in_file_spec out_file_spec out_file_dir out_file_type [ args... ]',
undef);

my $gbs_glt_lines = ENV_getenv( 'GBS_GLKB');

@GBS_GLKB = split( ' ', $gbs_glt_lines);

$GBS_IGNORE_ERRORS = ENV_getenv( 'GBS_IGNORE_ERRORS');
}



