#===============================================================
#
#   sca.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package sca;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
SCA_init
SCA_read_msgs
SCA_add_msg
SCA_get_filespec
SCA_get_enabled_msgs
SCA_msg_is_enabled
SCA_get_format_name
SCA_get_quiet_paths
SCA_get_threshold_refs
);

use strict;
use integer;

use glo::env;
use gbsenv;
use gbsfile;
use gbsfileglo;
use gbsfilecom;
use scamet;
use scadef;




sub SCA_init($$$$$);
sub SCA_read_msgs($);
sub SCA_add_msg($);
sub SCA_get_filespec();
sub SCA_get_enabled_msgs();
sub SCA_msg_is_enabled($);
sub SCA_get_format_name();
sub SCA_get_quiet_paths();
sub SCA_get_threshold_refs();

sub read_sca($);
sub do_errornumber($$$);
sub do_format($$$);
sub do_quiet($$$);
sub do_metrics_threshold($$$);




my $GBS_ROOT_PATH = GBSENV_gbs_root_path();
my $GBS_SYSAUDIT_PATH = "$GBS_ROOT_PATH/sysaudit";

my $CUR_SCA_AUDIT = '';
my $CUR_SCA_TARGET = '';
my $CUR_SCA_SYSLEVEL = -1;		# 0 = System, 1 = SubSys
my $CUR_SCA_REF;

my %SCA_INITS = ();



my %SCAS = ();










my %SPECIAL_MSG_NRS;	    # used only during read_sca





my @FORMAT_VALUES = qw( NONE DEFAULT PCLINT VISUAL_C GNU_C ECLIPSE);
my %FORMAT_VALUES = map { $_ => 1 } @FORMAT_VALUES;







my %SCA_DEF = (

E	=> [ -1,  \&do_errornumber  ],
F	=> [  1,  \&do_format],
Q	=> [  1,  \&do_quiet ],
T   => [  1,  \&do_metrics_threshold ],
);
my @TYPES = sort keys %SCA_DEF;

my @ALL_ENABLED_MSGS;




sub SCA_init($$$$$)
{
my ($audit,
$target,
$all_msg_nrs_ref,	    # All possible msg_nrs
$fatal_msg_nrs_ref,
$specw_msg_nrs_ref,
) = @_;




$CUR_SCA_AUDIT = $audit;
$CUR_SCA_TARGET = $target;
$CUR_SCA_SYSLEVEL = -1;
$CUR_SCA_REF = undef;

$SCAS{$CUR_SCA_AUDIT}->{$CUR_SCA_TARGET}->[0] = [];
$SCAS{$CUR_SCA_AUDIT}->{$CUR_SCA_TARGET}->[1] = [];

$SCA_INITS{$audit}->{$target}->[0] = [ @{$all_msg_nrs_ref} ];
$SCA_INITS{$audit}->{$target}->[1] = [ @{$fatal_msg_nrs_ref}, @{$specw_msg_nrs_ref} ];
}




sub SCA_read_msgs($)
{
my ($subsys,
) = @_;

read_sca( $subsys);
}





sub SCA_add_msg($)
{
my ($msg_nr) = @_;

if (!exists $CUR_SCA_REF->[1]->{$msg_nr})
{
$CUR_SCA_REF->[1]->{$msg_nr} = 1;	# enable
}
}





sub SCA_get_filespec()
{
return wantarray ? (@{$CUR_SCA_REF->[5]}) : $CUR_SCA_REF->[0];	#  @dependencies, $file_spec
}




sub SCA_get_enabled_msgs()
{
return @ALL_ENABLED_MSGS;
}




sub SCA_msg_is_enabled($)
{
my ($msg_nr,
) = @_;
my $enabled;

$enabled = $CUR_SCA_REF->[1]->{$msg_nr};

if (!defined $enabled)
{
ENV_error( "Msg $msg_nr in $CUR_SCA_AUDIT: $CUR_SCA_TARGET is not defined in sca_*.gbs");
$enabled = 0;
}

return $enabled;
}




sub SCA_get_format_name()
{
return $CUR_SCA_REF->[2];	    # $wanted_format
}




sub SCA_get_quiet_paths()
{
return @{$CUR_SCA_REF->[3]};    # @quiet_paths
}




sub SCA_get_threshold_refs()
{
return @{$CUR_SCA_REF->[4]};    # @metric_threshold_refs

}




sub read_sca($)
{
my ($subsys,
) = @_;

my $file_spec = "$GBS_ROOT_PATH/dev/$subsys/audit/$CUR_SCA_AUDIT/sca_$CUR_SCA_TARGET";
my @inc_path = ( "$GBS_ROOT_PATH/dev/$subsys/audit/$CUR_SCA_AUDIT",
"$GBS_SYSAUDIT_PATH/$CUR_SCA_AUDIT",
);


my $cur_file_spec = GBSFILEGLO_select_usr_gbs( $file_spec);
if (defined $cur_file_spec)
{
$CUR_SCA_SYSLEVEL = 1;	    # Found in SubSystem
} else
{
$file_spec = "$GBS_SYSAUDIT_PATH/$CUR_SCA_AUDIT/sca_$CUR_SCA_TARGET";
@inc_path = ( "$GBS_SYSAUDIT_PATH/$CUR_SCA_AUDIT",
);


$cur_file_spec = GBSFILEGLO_select_usr_gbs( $file_spec);
if (defined $cur_file_spec)
{
$CUR_SCA_SYSLEVEL = 0;  # Found in System
} else
{
$CUR_SCA_SYSLEVEL = 0;  # Assume found in System
}
}

$CUR_SCA_REF = $SCAS{$CUR_SCA_AUDIT}->{$CUR_SCA_TARGET}->[$CUR_SCA_SYSLEVEL];
if ($CUR_SCA_SYSLEVEL == 0 && @{$CUR_SCA_REF})
{
ENV_say( 2, "==> $subsys $CUR_SCA_AUDIT $CUR_SCA_TARGET $CUR_SCA_SYSLEVEL From Cache");	# System level only
} else
{
ENV_say( 2, "==> $subsys $CUR_SCA_AUDIT $CUR_SCA_TARGET $CUR_SCA_SYSLEVEL Reading $file_spec...");
my $all_msg_nrs_ref = $SCA_INITS{$CUR_SCA_AUDIT}->{$CUR_SCA_TARGET}->[0];	# All possible msg_nrs
my $special_msg_nrs_ref = $SCA_INITS{$CUR_SCA_AUDIT}->{$CUR_SCA_TARGET}->[1];	# @fatal_msg_nrs, @specw_msg_nrs




$CUR_SCA_REF->[0] = undef;	# $file_spec
$CUR_SCA_REF->[1] = {};		# $msg_nrs
$CUR_SCA_REF->[2] = 'DEFAULT';	# $wanted_format
$CUR_SCA_REF->[3] = [];		# @quiet_list
$CUR_SCA_REF->[4] = [];		# @metric_threshold_refs
$CUR_SCA_REF->[5] = [];		# @dependencies




map { $CUR_SCA_REF->[1]->{$_ * 1} = 1 } @{$all_msg_nrs_ref};




if (defined $cur_file_spec)
{
GBSFILE_open( $cur_file_spec, 0, \@inc_path, undef);	# $must_exist, $allow_target_line_selection
$CUR_SCA_REF->[0] = $cur_file_spec;




%SPECIAL_MSG_NRS = map { $_ * 1 => 1 } (@{$special_msg_nrs_ref});

my $line = GBSFILE_get_line();
while ($line)
{




my ($valid_line, $comment) = split( /[^\\]#/, $line, 2);




my ($op, @values) = split( / /, $valid_line);
GBSFILE_error_exit( "Line syntax = +|-<type> <value(s)>")
if (length( $op) != 2 || @values == 0);
my $plus_minus = substr( $op, 0, 1);
my $type = substr( $op, 1, 1);
my $enable = ($plus_minus eq '+') ? 1 : ($plus_minus eq '-') ? 0 : undef;
GBSFILE_error_exit( "Line syntax = +|-<type> <value(s)>")
if (!defined $enable);




my $sca_ref = $SCA_DEF{$type};
GBSFILE_error_exit( "Invalid type ($type). Must be one of (@TYPES)")
if (!defined $sca_ref);
my ($max_args, $do_func_ref) = @{$sca_ref};
GBSFILE_error_exit( "Only $max_args value(s) allowed")
if ($max_args >= 0 && @values > $max_args);
$do_func_ref->( $enable, $type, \@values);

$line = GBSFILE_get_line();
}
$CUR_SCA_REF->[5] = [ GBSFILE_get_dependencies() ];

ENV_warn( "No Msgs enabled for $CUR_SCA_AUDIT: $CUR_SCA_TARGET")
if (keys %{$CUR_SCA_REF->[1]} == 0);

%SPECIAL_MSG_NRS = ();	    # cleanup
} else
{
ENV_say( 1, "No SCA file. All messages enabled for $CUR_SCA_AUDIT: $CUR_SCA_TARGET");
}
@ALL_ENABLED_MSGS = grep $CUR_SCA_REF->[1]->{$_} == 1, sort { $a <=> $b } keys %{$CUR_SCA_REF->[1]};

}
}




sub do_errornumber($$$)
{
my ($enable,
$type,
$values_ref,
) = @_;

foreach my $msg_nr (@{$values_ref})
{
if ($msg_nr eq 'ALL')
{
if ($enable)
{
map { $CUR_SCA_REF->[1]->{$_} = 1 } keys %{$CUR_SCA_REF->[1]};
} else
{
map { $CUR_SCA_REF->[1]->{$_} = 0 } keys %{$CUR_SCA_REF->[1]};
map { $CUR_SCA_REF->[1]->{$_} = 1 } keys %SPECIAL_MSG_NRS;
}
} else
{
$msg_nr *= 1;
if (!exists $CUR_SCA_REF->[1]->{$msg_nr})
{
GBSFILE_warning( "Unknown Msg_nr $msg_nr in $CUR_SCA_AUDIT: $CUR_SCA_TARGET");
}
if ($enable)
{
$CUR_SCA_REF->[1]->{$msg_nr} = 1;
} else
{
$CUR_SCA_REF->[1]->{$msg_nr} = 0;
if (exists $SPECIAL_MSG_NRS{$msg_nr})
{
GBSFILE_info( "Implicit disable of fatal or special message ($msg_nr)");
}
}
}
}
}




sub do_format($$$)
{
my ($enable,
$type,
$values_ref,
) = @_;

my $value = $values_ref->[0];
GBSFILE_error_exit( "Invalid FORMAT value '$value' (@FORMAT_VALUES)")
if (!exists $FORMAT_VALUES{$value});

$CUR_SCA_REF->[2] = $value;			# $wanted_format
}




sub do_quiet($$$)
{
my ($enable,
$type,
$values_ref,
) = @_;

my $path = GBSFILECOM_validate_path( $values_ref->[0]);

push @{$CUR_SCA_REF->[3]}, $path;		# $quiet_paths
}




sub do_metrics_threshold($$$)
{
my ($enable,
$type,
$values_ref,
) = @_;

my $threshold = $values_ref->[0];
my ($metric_name, $min_value, $max_value, $msg_nr) = uc( $threshold) =~ /^([A-Z0-9]+)=([\d.]+|)(,[\d.]+|)(:\d+|)$/;

if (defined $metric_name)
{
$max_value = substr( $max_value, 1)	    # remove leading ','
if ($max_value ne '');
$msg_nr = substr( $msg_nr, 1)	    	    # remove leading ':'
if ($msg_nr ne '');
GBSFILE_error_exit( "At least one Threshold Limit (min and/or max) must be specified")
if ($min_value eq '' && $max_value eq '');

if (SCADEF_metric_is_enabled( $CUR_SCA_AUDIT, $metric_name))
{

push @{$CUR_SCA_REF->[4]}, [ $metric_name, $min_value, $max_value, $msg_nr ];	# @metric_threshold_refs
} else
{
if (SCAMET_name_is_valid( $metric_name))
{
GBSFILE_error_exit( "Metric mnemonic '$metric_name' not enabled")
} else
{
GBSFILE_error_exit( "Unknown metric: '$metric_name'");
}
}
} else
{
GBSFILE_error_exit( "Threshold value syntax is: <metric_name>=[<min_value>][,<max_value>][:<msg_nr>]");
}
}

return 1;


