#===============================================================
#
#   settings.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright © 2010-2020 - Randy Marques - All rights reserved
#===============================================================

package settings;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
SETTINGS_set_check
);

use strict;
use integer;

BEGIN
{
if ($^O eq 'MSWin32')
{
require glo::winreg;
import  glo::winreg;
}
}

use glo::env;
use glo::slurp;
use glo::spit;
use glo::file;
use glo::shell;
use glo::shortcut;
use gbsenv;




sub SETTINGS_set_check($$);

sub win32_check_registry();
sub check_unwanted($$$);
sub unix_check_profile($);
sub unix_check_envfile($$);
sub unix_insert_lines($$$);
sub unix_write_file($$);
sub copy_setup_files(@);
sub copy_setup_file($$);




my $IS_WIN32 = (ENV_is_win32());
my $IS_LUNIX = ! $IS_WIN32;

my $GBS_BASE_PATH = ENV_getenv_perl_path( 'GBS_BASE_PATH');	#  %APPDATA%/.gbs or ~/.gbs
$GBS_BASE_PATH =~ s/GBS~\d$/.gbs/
if ($IS_WIN32);

my @LUNIX_PROFILE_FILES = ( glob( '~/.bash_profile'), glob( '~/.bash_login'), glob( '~/.profile') );
my @LUNIX_ENV_FILES_REFS = (

[ glob( '~/.bashrc'), glob( '~/.bash_aliases') ],
[ glob( '~/.kshrc'), undef ],
);

my $LUNIX_GBS_FUNCTION = ". $GBS_BASE_PATH/gbsinit.sh";
my $WIN32_GBS_FUNCTION = "$ENV{GBS_BASE_PATH}\\gbsinit.bat";
my $SHELL_GBS_FUNCTION = ($IS_WIN32) ? SHELL_setfunc( 'gbs', $WIN32_GBS_FUNCTION) :
SHELL_setfunc( 'gbs', $LUNIX_GBS_FUNCTION, 1);

my $THIS_GBS_REL;
my $THIS_GBS_SCRIPTS;




sub SETTINGS_set_check($$)
{
my ($new_scripts_root_path,
$new_rel,
) = @_;

ENV_say( 1, "Set/Check Startup Settings...");

$THIS_GBS_REL = $new_rel;					# may have new value
$THIS_GBS_SCRIPTS = "$new_scripts_root_path/$new_rel";	# may have new value


if ($IS_WIN32)
{



win32_check_registry();




copy_setup_files( '*.bat', '*.ico');	# scripts and icons

} else  # LUNIX
{
{



my @found_files;
foreach my $profile_file (@LUNIX_PROFILE_FILES)
{
if (-f $profile_file)
{
push @found_files, $profile_file;
unix_check_profile( $profile_file);
}
}
ENV_error( "No ~/.profile and no ~/.bash_profile found!!")
if (!@found_files);
}
{



my $found_files = 0;
foreach my $ref (@LUNIX_ENV_FILES_REFS)
{
my ($env_file, $aliases_file) = @{$ref};
if (-f $env_file)
{
$found_files++;
if (defined $aliases_file)
{
if (-f $aliases_file)
{
unix_check_envfile( $env_file, 0);
unix_check_envfile( $aliases_file, 1);
} else
{
unix_check_envfile( $env_file, 1);
}
} else
{
unix_check_envfile( $env_file, 1);
}
}
}
ENV_error( "No Env Files (.bashrc, etc) found!!")
if ($found_files == 0);
}




copy_setup_files( '*.sh', '*.png');	# scripts and icons
}

if (ENV_get_desktop_path( 0) ne '')		# user_path
{



ENV_say( 1, "Checking Shortcuts...");
my $update_count = 0;
my $init_file = "$THIS_GBS_SCRIPTS/setup/shortcut";

my @releases;
push @releases, $THIS_GBS_REL;		    # $new_value
push @releases, ''			    # current
if (ENV_is_development());
my $define_spec = ENV_commandfile_spec( "$GBS_BASE_PATH/define");    # Note: shortcuts do no expand EnvVars

foreach my $release (@releases)
{
my $shortcut_name;
my $icon_spec;
my $command_args;
if ($release eq 'beta')
{



$shortcut_name = "GBS Startup [beta]";
$icon_spec = "$GBS_BASE_PATH/gbs_test";
$command_args = "$define_spec beta";
} else
{



$shortcut_name = "GBS Startup [cur]";
$icon_spec = "$GBS_BASE_PATH/gbs";
$command_args = "$define_spec cur";
}
if ($IS_WIN32)
{
$command_args = "/K $command_args";
} else
{
$command_args = "-title \"GBS xterm\" -geometry 100x40+100+100 -bg white -fg darkblue -sb -ls -e \"bash $command_args; bash -login\"";


}




$update_count += SHORTCUT_update( [ 'DESKTOP', 'STARTMENU/GBS' ],
$init_file,
$shortcut_name,
'Application',	    # $type Application or Link
undef,		    # $description,
$icon_spec,
undef,		    # $command_spec
$command_args,
undef,		    # $ working_directory
undef);		    # show: 0 = NORMAL, 1 = Maximised, -1 = Minimised




my $help_file_spec = "$THIS_GBS_SCRIPTS/doc/index.html";
$shortcut_name =~ s/Startup/Help/;
$update_count += SHORTCUT_update( [ 'DESKTOP', 'STARTMENU/GBS' ],
$init_file,
$shortcut_name,		# $shortcut_name,
'Link',			# $type Application or Link
'Startup the GBS help',	# $description,
$icon_spec,		# $icon_spec,
$help_file_spec,		# $command_spec
'',			# $command_args,
undef,			# $ working_directory
undef);			# show: 0 = NORMAL, 1 = Maximised, -1 = Minimised
}





if ($update_count > 0)
{
ENV_warn( "$update_count standard GBS Shortcut(s) on the DeskTop/StartMenu were updated.",
"If you have your own GBS Shortcuts, please update them accordingly.");
}
} else
{
my $dtp = ENV_get_desktop_path( 0);
ENV_say( 1, "Desktop directory ( $dtp) not found.",
"No checking or creation of Shortcuts");
}
}




sub win32_check_registry()
{



ENV_say( 1, "Checking Registry...");
foreach my $name (WINREG_get_userenv_names())
{
if (substr( uc $name, 0, 3) eq 'GBS')
{
my $value = WINREG_get_userenv( $name);
$value = '' if (!defined $value);
check_unwanted( $name, $value, 'Registry')
}
}
}




sub check_unwanted($$$)
{
my ($name,		    # Must start with 'GBS'. Mixed case ok.
$value,
$source) = @_;	    # 'Registry' or $line

my $warn_line;
my $uc_name = uc $name;
if (substr( $uc_name, 0, 4) eq 'GBS_')
{
if (!GBSENV_is_predefined_env( $uc_name))
{
$warn_line = "** This GBS EnvVar may not be defined here! **";
}
} elsif (substr( $uc_name, 0, 7) eq 'GBSEXT_')
{
$warn_line = "** GBSEXT_ EnvVars must be defined in the switch.gbs file! **";
}

if ($warn_line)
{
if ($IS_WIN32)
{
ENV_warn( "Registry: User EnvVar: $name ($value)",
$warn_line);
} else
{
ENV_warn( "This Environment File: User EnvVar $name ($value)",
$warn_line,
$source);
}
}
}




sub unix_check_profile($)
{
my ($profile_file) = @_;	# Unix-profile file

ENV_say( 1, "Checking LOGIN $profile_file...");
my $env_found = 0;
my $env_def;






my @lines = SLURP_file( $profile_file);
for (my $i = 0; $i < @lines; $i++)
{
my $line = $lines[$i];
next if ($line eq '');

if ($line =~ /\sENV=(.*)($|\s|#)/)
{
$env_def = $1;
$env_found = 1;
ENV_say( 1, "- Found: $i $line - OK");
} elsif ($line =~ /^\s*\\?(export|set)\s+([^=]+)="?([^"]*)"?$/)
{
my $command = $1;	    # export or set
my $name = $2;
my $value = $3;
if ($name =~ /^GBS_/ || $name =~ /^GBSEXT_/)
{
check_unwanted( $name, $value, $line);
}
}
}




if ($env_found)
{
my $env_file = ENV_expand_envs( $env_def);
push @LUNIX_ENV_FILES_REFS, [ $env_file, undef ]
if (! grep( $_->[0] eq $env_file, @LUNIX_ENV_FILES_REFS));	# add if not already present
}
}




sub unix_check_envfile($$)
{
my ($env_file,
$alias_must_be_present) = @_;
my $alias_is_present = 0;

ENV_say( 1, "Checking ENV $env_file...");
my $file_changed = 0;
my $alias_index = -1;







my @lines = SLURP_file( $env_file);
for (my $i = 0; $i < @lines; $i++)
{
my $line = $lines[$i];
next if ($line eq '' || $line =~ /^\s*#/);
if ($line =~ /^\s*\\?alias\s+gbs=/)
{
$alias_index = $i;
ENV_say( 1, "- Found: $i $line - OK");
} elsif ($line =~ /^s*\\?(export|set)\s+([^=]+)="?([^"]*)"?$/)
{
my $name = $2;
my $value = $3;
if ($name =~ /^GBS_/ || $name =~ /^GBSEXT_/)
{
ENV_warn( "- GBS EnvVars may not be set in '$env_file'",
"  - $line");
} else
{
ENV_warn( "- It is unnecessary and unwise to set EnvVars in '$env_file'",
"  - $line");
}
}
}




if ($alias_index == -1) 	    # alias gbs not found
{
if ($alias_must_be_present)
{



ENV_say( 1, "Adding '$SHELL_GBS_FUNCTION' to '$env_file'");
my @gbs_lines = (
'',
'#',
'#   GBS aliases',
'#',
$SHELL_GBS_FUNCTION,
'#',
);

@lines = unix_insert_lines( \@lines, -1, \@gbs_lines);	# -1: insert at end
$file_changed = 1;
$alias_is_present = 1;
}
} else
{
$alias_is_present = 1;
}

if ($file_changed)
{

unix_write_file( $env_file, \@lines);
}

return $alias_is_present;
}




sub unix_insert_lines($$$)
{
my ($lines_ref,
$index,	        	# insert before this index (-1 == append)
$inserts_ref,
) = @_;
my @lines;

@lines = @{$lines_ref};
my $i = ($index == -1) ? @lines : $index;
while ($i > 0 && $lines[$i-1] eq '')     { $i--; }
while ($i > 0 && $lines[$i-1] =~ /\s*#/) { $i--; }
while ($i > 0 && $lines[$i-1] eq '')     { $i--; }

my @inserts = @{$inserts_ref};
push @inserts, ''
if (!defined $lines[$i] || $lines[$i] ne '');
splice( @lines, $i, 0, @inserts);

return @lines;
}




sub unix_write_file($$)
{
my ($file_spec,
$lines_ref,
) = @_;

if (-f $file_spec)
{
my $backup_file_spec = ENV_backup_file( $file_spec, '.gbs_saved', 1);
ENV_say( 1, "- Saved $file_spec to $backup_file_spec");
}
ENV_say( 1, "- Writing $file_spec...");
my $writable = (-w $file_spec);
ENV_chmod( 'u+w', $file_spec) if (!$writable);
SPIT_file_nl( $file_spec, $lines_ref);
ENV_chmod( 'u-w', $file_spec) if (!$writable);
}




sub copy_setup_files(@)
{
my (@wild_files) = @_;
my $nr_copied = 0;

my $from_dir = "$THIS_GBS_SCRIPTS/setup";

foreach my $wild_file (@wild_files)
{
foreach my $file_spec (glob( "$from_dir/$wild_file"))
{
$nr_copied += copy_setup_file( $file_spec, $GBS_BASE_PATH);
}
}

return $nr_copied;
}




sub copy_setup_file($$)
{
my ($file_spec,
$to_dir,	# e.g: GBS_BASE_PATH
) = @_;
my $replicated;


my $file = ENV_split_spec_f( $file_spec);
$replicated = FILE_replicate( $file_spec, "$to_dir/$file");
if ($replicated == 1)
{
ENV_say( 1, "Copied $file to $to_dir");
if ($IS_LUNIX && $file =~ /\.sh$/)
{
ENV_say( 2, "- chmod u+x");
ENV_chmod( 'u+x', "$to_dir/$file");
}
}

return $replicated;
}

return 1;

