#===============================================================
#
#   steps.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package steps;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
STEPS_analyze_steps
STEPS_select_steps
STEPS_get_command
STEPS_get_subsystems_order
);

use strict;
use integer;

use glo::env;
use glo::format;
use glo::slurp;
use gbsenv;
use gbsglo;
use gbssfile;
use gbsfileglo;
use fspec;
use ssprop;




sub STEPS_analyze_steps($$);
sub STEPS_select_steps($$);
sub STEPS_get_command($);
sub STEPS_get_subsystems_order();

sub de_alias_step_names(@);
sub validate_step_names(@);
sub expand_step_names(@);
sub select_steps_by_execute(@);
sub display_steps();
sub fill_data();
sub read_steps();
sub step_entry($$);
sub do_step_command($$$$$);
sub do_step_targets($$$$$);
sub do_step_audits($$$$$);
sub step_exit($$);
sub alias_entry($$);
sub do_alias_steps($$$$$);
sub do_alias_target($$$$$);




my $GBS_ROOT_PATH = GBSENV_gbs_root_path();

my @GBS_ALL_TARGETS = ENV_getenv( 'GBS_ALL_TARGETS');		# All platforms
my %GBS_ALL_TARGETS = map { ($_, 1) } @GBS_ALL_TARGETS;

my @GBS_ALL_AUDITS = ENV_getenv( 'GBS_ALL_AUDITS');		# All platforms
my %GBS_ALL_AUDITS = map { ($_, 1) } @GBS_ALL_AUDITS;

my @GBS_TARGETS = ENV_getenv( 'GBS_TARGETS');			# Current platform only
my %GBS_TARGETS = map { ($_, 1) } @GBS_TARGETS;

my @GBS_AUDITS = ENV_getenv( 'GBS_AUDITS');			# All platforms
my %GBS_AUDITS = map { ($_, 1) } @GBS_AUDITS;

my $STEPS_FILE_IS_PRESENT = 0;
my $STEPS_READ = 0;
my $DATA_FILLED = 0;




my $CUR_STEP;
my $CUR_SUBSYS;
my @CUR_TARGETS;
my @CUR_AUDITS;




my @ALL_SUBSYSTEMS;	    # Do not initialze!
my %ALL_SUBSYSTEMS;	    # Do not initialze!




my @STEP_NAMES_ORDER;
my %STEPS_DATA;









my @ALIAS_NAMES_ORDER;




my %ALL_EXEC = (
SELECT  => 1,	# Will be executed if selected
SPECIFY => 1,	# Will only be executed if explicitly specifed
ALLWAYS => 1,	# Will allways be executed, not if job fails
FORCE   => 1,	# Will allways be executed, even if job fails
);

my %STEPS;


























my %STEP_DEFS = (
COMMAND	    => [ 'sc' , undef,     0, undef, undef,  \&do_step_command, undef      ],
EXEC	    => [ 'sss', 'SELECT',  0, undef,     1,  undef,	        \%ALL_EXEC ],
TARGETS	    => [ 'ss' , ['*'],     0, undef,   [1],  \&do_step_targets, undef      ],
AUDITS	    => [ 'ss' , ['*'],     0, undef,   [1],  \&do_step_audits,  undef      ],
);

my %ALIAS_DEFS = (
STEPS	    => [ 'ss' , [],        0, undef,   [1],  \&do_alias_steps,  undef ],
TARGET	    => [ 'ss' , '',	   0, undef,     1,  \&do_alias_target, undef ],
);










my %SECTION_DEFS = (
STEP     => [ 'S', 1, 0,   1, \%STEP_DEFS,  \&step_entry,  \&step_exit ],
ALIAS    => [ 'S', 0, 0,   1, \%ALIAS_DEFS, \&alias_entry, undef ],
);




my %SPECIFIED_STEPS;


my $JOBTYPE;	    # gen, make or audit
my @EXPANDED_STEPS;
my %ALIASED_TARGETS;




sub STEPS_analyze_steps($$)
{
my ($jobtype,    # 'gen', 'make' or 'audit'
$steps_ref,
) = @_;
my @selectable_targets;

$JOBTYPE = $jobtype;
my @steps = @{$steps_ref};
@EXPANDED_STEPS = ();

ENV_say( 1, "Steps: @steps")
if (@steps && $steps[0] ne '?');




fill_data()
if (!$DATA_FILLED);






if (!@steps)
{
@steps = @STEP_NAMES_ORDER;
} elsif ($steps[0] eq '?')
{
ENV_say( 1, "Showing steps...");
display_steps();
ENV_exit( 0);
} else
{



@steps = de_alias_step_names( @steps);
@steps = validate_step_names( @steps);
@steps = expand_step_names( @steps);
}
@EXPANDED_STEPS = select_steps_by_execute( @steps);


if (%ALIASED_TARGETS)
{
@selectable_targets = keys( %ALIASED_TARGETS);
ENV_say( 1, "Aliased Target(s): @selectable_targets");
} else
{
if (%SPECIFIED_STEPS)
{
my %selected_targets;
foreach my $step (keys %SPECIFIED_STEPS)
{

my $targets_ref = $STEPS_DATA{$step}->[6];	    # $targets_ref
map { $selected_targets{$_} = 1 } keys( %{$targets_ref});
}
@selectable_targets = keys( %selected_targets);
ENV_say( 1, "Selectable Target(s): @selectable_targets");
} else
{
@selectable_targets = @GBS_TARGETS;
ENV_say( 1, "Selectable Target(s): All Targets");
}
}

return @selectable_targets;
}







sub STEPS_select_steps($$)
{
my ($target,
$audit_name, # '' for gen & make
) = @_;
my @selected_steps;





foreach my $step_name (@EXPANDED_STEPS)
{

my (undef, undef, undef, undef, undef, undef, $targets_ref, $audits_ref) = @{$STEPS_DATA{$step_name}};

if ($audit_name ne '')
{
if ($audits_ref->{$audit_name})
{
push @selected_steps, $step_name
if (GBSGLO_subsystem_does_audit( $step_name, $audit_name, $target));
} else
{
if (exists $SPECIFIED_STEPS{$step_name})
{
ENV_say( 1, "Step skipped for Audit '$audit_name': '$step_name'");
}
}
} else
{
if ($targets_ref->{$target})
{
push @selected_steps, $step_name;
} else
{
if (exists $SPECIFIED_STEPS{$step_name})
{
ENV_say( 1, "Step skipped for Target '$target': '$step_name'");
}
}
}
}


return @selected_steps;
}




sub de_alias_step_names(@)
{
my @in_steps = @_;
my @out_steps;




my %changed_steps;

foreach my $step_name (@in_steps)
{
if ($step_name eq '.')
{
$step_name = ENV_getenv( 'GBS_SUBSYS');
ENV_error_exit( "Steps: No current SubSystem (.)")
if ($step_name eq '');
push @out_steps, $step_name;
} elsif ($step_name eq '-')
{
push @out_steps, '-';
} elsif (exists $STEPS{ALIAS}->{$step_name})
{
my $alias_ref = $STEPS{ALIAS}->{$step_name};
my @alias_steps = @{$alias_ref->{STEPS}};
my $alias_target = $alias_ref->{TARGET};
push @out_steps, @alias_steps
if (@alias_steps);		# ALIAS possibly changes TARGET only
if ($alias_target)
{
@alias_steps = @STEP_NAMES_ORDER
if (!@alias_steps);
foreach my $step (@alias_steps)
{
my $changed_target = $changed_steps{$step};
if (!defined $changed_target)
{
$changed_steps{$step} = $alias_target;
} else
{
ENV_error_exit( "Alias '$step_name' changed Target more than once for STEP '$step'",
"( $changed_target => $alias_target)")
if ($alias_target ne $changed_target);
}
$STEPS_DATA{$step}->[6] = { $alias_target => 1 };	# $targets_ref
}
$ALIASED_TARGETS{ $alias_target } = 1;

}
} else
{
push @out_steps, $step_name;
}
}

return @out_steps;
}




sub validate_step_names(@)
{
my @in_steps = @_;
my @out_steps;




my $previous_step_index = -1;
if (@in_steps == 1 && $in_steps[0] eq 'ALL' && !exists $STEPS_DATA{'ALL'})
{
ENV_say( 1, "Selection forced to 'ALL'");
map { $SPECIFIED_STEPS{$_} = 1 } @STEP_NAMES_ORDER;
@out_steps = @STEP_NAMES_ORDER;
} else
{
foreach my $step_name (@in_steps)
{
if ($step_name eq '-')
{

} else
{
ENV_error_exit( "Unknown step '$step_name'",
"Use 'gbssys... ?' to show all steps")
if (!exists $STEPS_DATA{$step_name});

my $this_step_index = $STEPS_DATA{$step_name}->[0];	# $step_nr
if (defined $this_step_index)
{
if ($this_step_index < $previous_step_index)
{
my $previous_step_name = $STEP_NAMES_ORDER[$previous_step_index];
ENV_error_exit( "'$previous_step_name' must appear after '$step_name'",
"Proper order is: @STEP_NAMES_ORDER");
} elsif ($this_step_index == $previous_step_index)
{
ENV_error_exit( "'$step_name' specified more than once");
}
$previous_step_index = $this_step_index;
}
$SPECIFIED_STEPS{$step_name} = 1;
}
}
@out_steps = @in_steps;
}
}




sub expand_step_names(@)
{
my @in_steps = @_;
my @out_steps;

if (!@in_steps)
{
@out_steps = @STEP_NAMES_ORDER;
} else
{



my @working_steps = @in_steps;
unshift @working_steps, $STEP_NAMES_ORDER[0] if ($in_steps[0] eq '-');
push @working_steps, $STEP_NAMES_ORDER[-1] if ($in_steps[-1] eq '-');


while (@working_steps)
{
my $step1 = shift @working_steps;

if (@working_steps && $working_steps[0] eq '-')
{
shift @working_steps;	    # skip '-'
my $step2 = shift @working_steps;
ENV_error_exit( "Subsequent hyphens (- -) not allowed")
if ($step2 eq '-');;
my $i1 = $STEPS_DATA{$step1}->[0];			# $step_nr
my $i2 = $STEPS_DATA{$step2}->[0];			# $step_nr
push @out_steps, @STEP_NAMES_ORDER[$i1..$i2];
} else
{
push @out_steps, $step1;
}
}
}

return @out_steps;
}




sub select_steps_by_execute(@)
{
my @in_steps = @_;
my @out_steps;

my %in_steps = map { $_ => 1 } @in_steps;
foreach my $step (@STEP_NAMES_ORDER)
{
my $execute = $STEPS_DATA{$step}->[1];	    # $execute
if (exists $in_steps{$step})
{
if ($execute eq 'SPECIFY' )
{
push @out_steps, $step
if (exists $SPECIFIED_STEPS{$step});
} else  # SELECT, ALLWAYS or FORCE
{
push @out_steps, $step;
}
} else
{
if ($execute eq 'ALLWAYS' || $execute eq 'FORCE')
{
ENV_say( 1, "Step added: '$step' ($execute)");
push @out_steps, $step;
} else
{

}
}
}

return @out_steps;
}




sub STEPS_get_command($)
{
my ($step_name) = @_;
my ($subsys, $command, $force_execute);

fill_data()
if (!$DATA_FILLED);

ENV_error_exit( "Unknown step '$step_name'")
if (!exists $STEPS_DATA{$step_name});

(undef, my $execute, undef, $subsys, undef, $command) = @{$STEPS_DATA{$step_name}};
$force_execute = ($execute eq 'FORCE');

return ($subsys, $command, $force_execute);
}




sub display_steps()
{



{
my @row_refs = ( [ 'Step', 'Exec', 'Type', 'Audits', 'Targets' ] );

foreach my $step_name (@STEP_NAMES_ORDER)
{
my (undef, $execute, $step, $subsys, $ss_type, undef, $targets_ref, $audits_ref) = @{$STEPS_DATA{$step_name}};

my $type;
if ($subsys ne '')
{
$type = $ss_type;
} else
{
$type = 'Command';
}
my $audits_comma_list = join( ',', sort( keys %{$audits_ref}));
my $targets_comma_list = join( ',', sort( keys %{$targets_ref}));
push @row_refs, [ $step_name, $execute, $type, $audits_comma_list, $targets_comma_list ];
}

ENV_say( 0, "- Steps for $GBS_ROOT_PATH:",
FORMAT_table( 0, 2, '  ', undef, @row_refs));
}




{
my @row_refs = ( [ 'Step', 'Command' ] );

foreach my $step_name (@STEP_NAMES_ORDER)
{
my (undef, undef, $step, undef, undef, $command ) = @{$STEPS_DATA{$step_name}};
if ($command)
{
push @row_refs, [ $step_name, $command ];
}
}
if (@row_refs > 1)
{
ENV_say( 0, '',
"- Commands for $GBS_ROOT_PATH:",
FORMAT_table( 0, 2, '  ', undef, @row_refs));
}
}




{
my @row_refs = ([ 'Alias', 'Target', 'Steps' ]);
foreach my $alias (@ALIAS_NAMES_ORDER)
{
my $alias_ref = $STEPS{ALIAS}->{$alias};
my @alias_steps = @{$alias_ref->{STEPS}};
my $alias_target = $alias_ref->{TARGET};
push @row_refs, [ $alias, $alias_target, "@alias_steps" ];
}
if (@row_refs > 1)
{
ENV_say( 0, '',
"- Aliases for $GBS_ROOT_PATH:",
FORMAT_table( 0, 2, '	', undef, @row_refs));
} else
{
ENV_say( 0, '',
"- No Aliases for $GBS_ROOT_PATH");
}
}
}




sub fill_data()
{
read_steps()
if (!$STEPS_READ);

if ($STEPS_FILE_IS_PRESENT)
{







my $nr_steps = @STEP_NAMES_ORDER;
for (my $i = 0; $i < $nr_steps; $i++)
{
my $subsys_or_step_name = $STEP_NAMES_ORDER[$i];
my $step_ref = $STEPS{STEP}->{$subsys_or_step_name};
my $step_nr = $i;
my $execute = $step_ref->{EXEC};
my $stepname= '';
my $subsys = '';
my $ss_type = '';
my $command = '';
my @targets = @{$step_ref->{TARGETS}};
my @audits = @{$step_ref->{AUDITS}};
if (exists $ALL_SUBSYSTEMS{$subsys_or_step_name})
{
$subsys = $subsys_or_step_name;
$stepname = $subsys;
$ss_type = SSPROP_get_type( undef, $subsys);
@targets = GBSGLO_subsystem_targets( $subsys)
if ("@targets" eq '*');
@audits = GBSGLO_subsystem_audits( $subsys)
if ("@audits" eq '*');
} else
{
$stepname = $subsys_or_step_name;
my $command_ref = $step_ref->{COMMAND};


$command = $command_ref->[2];
@targets = @GBS_TARGETS
if ("@targets" eq '*');
@audits = ()
if ("@audits" eq '*');
}
my %targets = map { $_ => 1 } @targets;
my %audits = map { $_ => 1 } @audits;
$STEPS_DATA{$subsys_or_step_name} = [ $step_nr, $execute, $stepname, $subsys, $ss_type, $command,
\%targets, \%audits ];

}
} else
{
my $nr_steps = @ALL_SUBSYSTEMS;
for (my $i = 0; $i < $nr_steps; $i++)
{
my $step_nr = $i;
my $execute = 'SELECT';
my $stepname = $ALL_SUBSYSTEMS[$i];
my $subsys = $stepname;
my $ss_type = SSPROP_get_type( undef, $subsys);
my $command = '';
my %targets = map { $_ => 1 } GBSGLO_subsystem_targets( $subsys);
my %audits = map { $_ => 1 } GBSGLO_subsystem_audits( $subsys);
$STEPS_DATA{$subsys} = [ $step_nr, $execute, $stepname, $subsys, $ss_type, $command,
\%targets, \%audits ];

}
}


$DATA_FILLED = 1;
}




sub read_steps()
{
@ALL_SUBSYSTEMS = GBSGLO_select_dirs( SLURP_dir_dirs( "$GBS_ROOT_PATH/dev", 0));
%ALL_SUBSYSTEMS = map { $_ => 1 } @ALL_SUBSYSTEMS;

my $full_file_spec = FSPEC_steps( 'steps', $GBS_ROOT_PATH);    # .gbs or .usr
if (-e $full_file_spec)
{
$STEPS_FILE_IS_PRESENT = 1;
GBSSFILE_parse( \%STEPS, $full_file_spec, 'step', undef, \%SECTION_DEFS);

foreach my $subsys (@ALL_SUBSYSTEMS)
{
if (!exists $STEPS{STEP}->{$subsys})
{
ENV_error_exit( "SubSystem: '$subsys' not specified in any STEP.",
"Please 'gbsedit steps.gbs|.usr' and add the SubSystem.");
}
}

} else
{
@STEP_NAMES_ORDER = @ALL_SUBSYSTEMS;

ENV_say( 1, "No 'steps.gbs' defined",
"SubSystems will be taken in alphabetical order...");
}

$STEPS_READ = 1;
}




sub step_entry($$)
{
my ($section,	# STEP
$args_ref,	# $stepname
) = @_;

my $stepname =  $args_ref->[0];





GBSFILEGLO_error_exit( "Invalid stepname '$stepname'")
if ($stepname eq '-');




if (exists $ALL_SUBSYSTEMS{$stepname})
{
$CUR_SUBSYS = $stepname;
@CUR_TARGETS = GBSGLO_subsystem_targets( $CUR_SUBSYS);
@CUR_AUDITS = GBSGLO_subsystem_audits( $CUR_SUBSYS);
} else
{
$CUR_SUBSYS = '';
@CUR_TARGETS = @GBS_TARGETS;
@CUR_AUDITS = ();
}
$CUR_STEP = $stepname;
push @STEP_NAMES_ORDER, $stepname;

return ( $stepname );
}




sub do_step_command($$$$$)
{
my ($section,	    # STEP
$subsection,	    # $step_name
$item,		    # COMMAND
$values_ref,	    # [$command_type, $pos_arg_tokens, $command_line, $ok_values, $exec_condition];

$constr_arg,
) = @_;




my (undef, undef, $command_line, undef, undef) = @{$values_ref};
GBSFILEGLO_error_exit( "Cannot specify COMMAND for SubSystem STEP '$subsection'", "($command_line)")
if ($CUR_SUBSYS ne '');


return $values_ref;
}




sub do_step_targets($$$$$)
{
my ($section,   # STEP
$subsection,# $step_name
$item,	    # TARGETS
$values_ref,
$constr_arg,
) = @_;


my @targets = @{$values_ref};

my $targets = "@targets";
if ($targets eq '*')
{
@targets = @CUR_TARGETS;
} elsif ($targets eq '! *' || $targets eq '!')
{
@targets = ();
} elsif ("@targets" eq '-')
{

} else
{



my $target_op = '';
$target_op = shift @targets 	# '!'
if ($targets[0] eq '!');

my %allowed_targets = map { $_ => 1 } @CUR_TARGETS;
my %targets;
foreach my $target (@targets)
{
GBSFILEGLO_error_exit( "No such Target: $target")
if (!exists $GBS_ALL_TARGETS{$target});
GBSFILEGLO_warning( "Step '$CUR_STEP' does not execute for Target '$target'")
if (!exists $allowed_targets{$target});
GBSFILEGLO_error_exit( "Duplicate Target: $target")
if (exists $targets{$target});
$targets{$target} = 1;
}
@targets = keys %targets;
if ($target_op eq '!')
{
map { delete $allowed_targets{$_} } @targets;
@targets = keys %allowed_targets;
}
}

return [ sort( @targets) ];
}




sub do_step_audits($$$$$)
{
my ($section,   # STEP
$subsection,# $step_name
$item,	    # AUDITS
$values_ref,
$constr_arg,
) = @_;


my @audits = @{$values_ref};

my $audits = "@audits";
if ($audits eq '*')
{
@audits = @CUR_AUDITS;
} elsif ($audits eq '! *' || $audits eq '!')
{
@audits = ();
} elsif ("@audits" eq '-')
{

} else
{



my $audit_op = '';
$audit_op = shift @audits 	# '!'
if ($audits[0] eq '!');

my %allowed_audits = map { $_ => 1 } @CUR_AUDITS;
my %audits;
foreach my $audit (@audits)
{
GBSFILEGLO_error_exit( "No such Audit: $audit")
if (!exists $GBS_ALL_AUDITS{$audit});
GBSFILEGLO_warning( "Step '$CUR_STEP' does not execute for Audit '$audit'")
if (!exists $allowed_audits{$audit});
GBSFILEGLO_error_exit( "Duplicate Audit: $audit")
if (exists $audits{$audit});
$audits{$audit} = 1;
}
@audits = keys %audits;
if ($audit_op eq '!')
{
map { delete $allowed_audits{$_} } @audits;
@audits = keys %allowed_audits;
}
}

return [ sort( @audits) ];
}




sub step_exit($$)
{
my ($section,	# STEP
$args_ref,	# $stepname
) = @_;

my ($stepname) = @{$args_ref};




if ($CUR_SUBSYS eq '' &&
!defined $STEPS{STEP}->{$stepname}->{COMMAND})
{
GBSFILEGLO_error_exit( "COMMAND required after non-SubSystem STEP (STEP=$stepname)");
}

}




sub alias_entry($$)
{
my ($section,	# ALIAS
$args_ref,	# $alias
) = @_;

my $alias = $args_ref->[0];

push @ALIAS_NAMES_ORDER, $alias;

return ($alias);
}




sub do_alias_steps($$$$$)
{
my ($section,   # ALIAS
$subsection,# $alias_name
$item,	    # STEPS
$values_ref,
$constr_arg,
) = @_;



my @steps = @{$values_ref};
my @step_names_order = @STEP_NAMES_ORDER;
foreach my $step (@steps)
{
next if ($step eq '-');
GBSFILEGLO_error_exit( "No such step '$step'")
if (!exists $STEPS{STEP}->{$step});

while (@step_names_order && $step ne $step_names_order[0])
{
shift @step_names_order;
}
GBSFILEGLO_error_exit( "Wrong step-sequence for step '$step' (@STEP_NAMES_ORDER)")
if (!@step_names_order);
}

return [ @steps ];
}




sub do_alias_target($$$$$)
{
my ($section,   # ALIAS
$subsection,# $alias_name
$item,	    # TARGET
$value_ref,
$constr_arg,
) = @_;



my $value = $value_ref->[0];
GBSFILEGLO_error_exit( "No such Target: $value")
if (!exists $GBS_ALL_TARGETS{$value});

return $value;
}








sub STEPS_get_subsystems_order()
{
my @subsys_order;




$GBS_ROOT_PATH = ENV_getenv_perl_path( 'GBS_ROOT_PATH');
@ALL_SUBSYSTEMS = GBSGLO_select_dirs( SLURP_dir_dirs( "$GBS_ROOT_PATH/dev", 0));
%ALL_SUBSYSTEMS = map { $_ => 1 } @ALL_SUBSYSTEMS;

my $full_file_spec = FSPEC_steps( 'steps', $GBS_ROOT_PATH);
if (-e $full_file_spec)
{
my %all_subsystems = %ALL_SUBSYSTEMS;	    # Copy because we are going to delete
my @lines = grep( $_ !~ /^\s/, SLURP_file_full( $full_file_spec, qr/^#/, undef));
foreach my $line (@lines)
{

my ($step_name) = $line =~ /^STEP\s+(.*)/;
if (defined $step_name && exists $all_subsystems{$step_name})
{
push @subsys_order, $step_name;
delete $all_subsystems{$step_name};
}
}
my @remaining_subsys = keys %all_subsystems;
if (@remaining_subsys)
{
push @subsys_order, @remaining_subsys;
ENV_warn( "SubSystem(s) '@remaining_subsys' do not appear in 'steps.gbs'",
"Temporarily added at end",
"Please use 'gbsedit steps.gbs' to add the missing SubSystem(s)");
}
} else
{
@subsys_order = @ALL_SUBSYSTEMS;
}

return @subsys_order;
}

return 1;


