#===============================================================
#
#   system.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package system;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
SYSTEM_get
SYSTEM_get_os
SYSTEM_get_os_refs
SYSTEM_get_os_ref_names
SYSTEM_get_os_ref_value
SYSTEM_get_osnames
SYSTEM_create
SYSTEM_put
SYSTEM_put_os
SYSTEM_put_os_refs
SYSTEM_add_os_ref
SYSTEM_remove_os_ref
SYSTEM_write
SYSTEM_close
SYSTEM_fix
);

use strict;
use integer;

use glo::env;
use glo::scm;
use gbsfile;




sub SYSTEM_get($$);
sub SYSTEM_get_os($$$);
sub SYSTEM_get_os_refs($$$);
sub SYSTEM_get_os_ref_names($$$);
sub SYSTEM_get_os_ref_value($$$$);
sub SYSTEM_get_osnames($$);
sub SYSTEM_create($);
sub SYSTEM_put($$$);
sub SYSTEM_put_os($$$$);
sub SYSTEM_put_os_refs($$$$);
sub SYSTEM_add_os_ref($$$$);
sub SYSTEM_remove_os_ref($$$$);
sub SYSTEM_write();
sub SYSTEM_close();
sub SYSTEM_fix($;@);

sub read_file($);
sub get_item($);
sub get_os_item($$);
sub put_item($$);
sub put_os_item($$$);
sub write_file();
sub general_init();
sub check_value($$$$);








my $CUR_OS = $^O;
my @SUPPORTED_SCMS = SCM_get_names( '*');   # All platforms
my $DEFAULT_SCMS = $SUPPORTED_SCMS[0];

my %DEFS = (




id		    => [ 0, 0, 'system',       0, [ 'system' ] ],
system_name     => [ 0, 0, 'NoName',       0, undef ],
build_start	    => [ 0, 0, '0',            0, qr/^\d{9}$/ ],		# VVVYYMMDD
build_end	    => [ 0, 0, '',             1, qr/^\d{9}$/ ],		# VVVYYMMDD
scms	    => [ 1, 0, $DEFAULT_SCMS,  0, \@SUPPORTED_SCMS ],
scms_data	    => [ 0, 0, '',             1, undef ],
scms_repository => [ 1, 0, '',             1, undef ],
root_version    => [ 0, 0, '0.00',         0, qr/^\d+\.\d\d$/ ],		# r.vv
targets	    => [ 1, 1, [],             0, undef ],			# [ $target, $plugin ]
audits	    => [ 1, 1, [],             1, undef ],			# [ $audit, $plugn ]
tools	    => [ 1, 1, [],             1, undef ],			# [ $tool, $plugin ]
);
my @ITEMS = qw( id system_name root_version build_start build_end
scms scms_data scms_repository targets audits tools);




my $CURRENT_FILE = '';
my $FILE_CHANGED = 0;

my %DATA;



my %OTHER_OS;


my @MISSING_ITEMS;
my $RUNNING_FIX = 0;




sub SYSTEM_get($$)
{
my ($root_path,
$item,
) = @_;

my $def_ref = $DEFS{$item};
ENV_fatal( "No such SYSTEM item ($item)")
if (!defined $def_ref);
ENV_fatal( "Must use SYSTEM_get_os")
if ($def_ref->[0] == 1);	# $is_os_item

read_file( $root_path);

return get_item( $item);
}





sub SYSTEM_get_os($$$)
{
my ($root_path,
$osname,	    # '.' == Current OS
$item,
) = @_;
my $value;		    # $value or $values_ref

my $def_ref = $DEFS{$item};
ENV_fatal( "No such SYSTEM item ($item)")
if (!defined $def_ref);
ENV_fatal( "Must use SYSTEM_get")
if ($def_ref->[0] == 0);	# $is_os_item

read_file( $root_path);

$osname = $CUR_OS
if ($osname eq '.');

$value = get_os_item( $osname, $item);		# scalar: $value or $values_ref

return $value;
}





sub SYSTEM_get_os_refs($$$)
{
my ($root_path,
$osname,		# '.' == Current OS, '*' == All OS
$item,
) = @_;
my @values_refs;		# ( [ $name, $value ], ...)

my $def_ref = $DEFS{$item};
ENV_fatal( "No such SYSTEM item ($item)")
if (!defined $def_ref);
ENV_fatal( "Must use SYSTEM_get")
if ($def_ref->[0] == 0);	# $is_os_item

read_file( $root_path);

$osname = $CUR_OS
if ($osname eq '.');

my @osnames = ($osname eq '*') ? keys %{$DATA{$item}} : ($osname);
foreach my $osname (@osnames)
{
push @values_refs, @{get_os_item( $osname, $item)};
}

return @values_refs;
}





sub SYSTEM_get_os_ref_names($$$)
{
my ($root_path,
$osname,		# '.' == Current OS, '*' == All OS
$item,
) = @_;
my @names;

my $def_ref = $DEFS{$item};
ENV_fatal( "No such SYSTEM item ($item)")
if (!defined $def_ref);
ENV_fatal( "Must use SYSTEM_get")
if ($def_ref->[0] == 0);	# $is_os_item

read_file( $root_path);

$osname = $CUR_OS
if ($osname eq '.');

my @osnames = ($osname eq '*') ? keys %{$DATA{$item}} : ($osname);
foreach my $osname (@osnames)
{
push @names, map { $_->[0] } @{get_os_item( $osname, $item)};	# [ $name, $value ]
}

return @names;
}




sub SYSTEM_get_os_ref_value($$$$)
{
my ($root_path,
$osname,		# '.' == Current OS
$item,
$name,
) = @_;
my $value;

my $def_ref = $DEFS{$item};
ENV_fatal( "No such SYSTEM item ($item)")
if (!defined $def_ref);
ENV_fatal( "Must use SYSTEM_get")
if ($def_ref->[0] == 0);	# $is_os_item

read_file( $root_path);

$osname = $CUR_OS
if ($osname eq '.');

foreach my $ref (@{get_os_item( $osname, $item)})	# [ $name, $value ]
{
if ($ref->[0] eq $name)
{
$value = $ref->[1];
last;
}
}
ENV_fatal( "No such SYSTEM item name: $osname $item $name")
if (!defined $value);

return $value;
}






sub SYSTEM_get_osnames($$)
{
my ($root_path,
$item,
) = @_;
my @osnames;

my $def_ref = $DEFS{$item};
ENV_fatal( "No such SYSTEM item ($item)")
if (!defined $def_ref);
ENV_fatal( "No osname for item $item")
if ($def_ref->[0] == 0);	# $is_os_item

read_file( $root_path);

@osnames = keys( %{$DATA{$item}});
@osnames = reverse @osnames
if ($osnames[0] ne $CUR_OS);

return @osnames;
}




sub SYSTEM_create($)
{
my ($root_path,
) = @_;

general_init();
$CURRENT_FILE = "$root_path/system.gbs";
$FILE_CHANGED = 0;
}




sub SYSTEM_put($$$)
{
my ($root_path,
$item,
$value,	    # $value or $values_ref
) = @_;

my $def_ref = $DEFS{$item};
ENV_fatal( "No such SYSTEM item ($item)")
if (!defined $def_ref);
ENV_fatal( "Must use SYSTEM_put_os")
if ($def_ref->[0] == 1);	# $is_os_item

read_file( $root_path);
put_item( $item, $value);
}




sub SYSTEM_put_os($$$$)
{
my ($root_path,
$osname,    # '.' == Current OS
$item,
$value,	    # $value or $values_ref
) = @_;

my $def_ref = $DEFS{$item};
ENV_fatal( "No such SYSTEM item ($item)")
if (!defined $def_ref);
ENV_fatal( "Must use SYSTEM_put")
if ($def_ref->[0] == 0);	# $is_os_item

read_file( $root_path);

$osname = $CUR_OS
if ($osname eq '.');

put_os_item( $osname, $item, $value);
}




sub SYSTEM_put_os_refs($$$$)
{
my ($root_path,
$osname,	    # '.' == Current OS
$item,
$values_refs_ref,   # [ [ $name, $value ], ... ]
) = @_;

my $def_ref = $DEFS{$item};
ENV_fatal( "No such SYSTEM item ($item)")
if (!defined $def_ref);
ENV_fatal( "Must use SYSTEM_put")
if ($def_ref->[0] == 0);	# $is_os_item

read_file( $root_path);

$osname = $CUR_OS
if ($osname eq '.');

my @values_refs = sort( { $a->[0] cmp $b->[0] } @{$values_refs_ref});

put_os_item( $osname, $item, \@values_refs);
}




sub SYSTEM_add_os_ref($$$$)
{
my ($root_path,
$osname,	    # '.' == Current OS
$item,
$values_ref,	    # [ $name, $value ]
) = @_;

my $def_ref = $DEFS{$item};
ENV_fatal( "No such SYSTEM item ($item)")
if (!defined $def_ref);
ENV_fatal( "Must use SYSTEM_put")
if ($def_ref->[0] == 0);	# $is_os_item

read_file( $root_path);

$osname = $CUR_OS
if ($osname eq '.');

my @values_refs = (@{get_os_item( $osname, $item)}, $values_ref);		    # [ $name, $value ]
@values_refs = sort( { $a->[0] cmp $b->[0] } @values_refs);

put_os_item( $osname, $item, \@values_refs);
}




sub SYSTEM_remove_os_ref($$$$)
{
my ($root_path,
$osname,	    # '.' == Current OS
$item,
$name,		    # [ $name, $value ]
) = @_;

print "remove $name\n";
my $def_ref = $DEFS{$item};
ENV_fatal( "No such SYSTEM item ($item)")
if (!defined $def_ref);
ENV_fatal( "Must use SYSTEM_put")
if ($def_ref->[0] == 0);	# $is_os_item

read_file( $root_path);

$osname = $CUR_OS
if ($osname eq '.');

my @values_refs = grep( $_->[0] ne $name, @{get_os_item( $osname, $item)});	    # [ $name, $value ]

put_os_item( $osname, $item, \@values_refs);
}




sub SYSTEM_write()
{
if ($FILE_CHANGED)
{
write_file();
}
}




sub SYSTEM_close()
{
if ($CURRENT_FILE ne '')
{
$CURRENT_FILE = '';
$FILE_CHANGED = 0;
%DATA = ();
%OTHER_OS = ();
}
}




sub SYSTEM_fix($;@)
{
my ($root_path,
@item_pairs,	    # repair-pairs
) = @_;
my $is_updated = 0;

$RUNNING_FIX = 1;
read_file( $root_path);
while (@item_pairs)
{
my $name = shift @item_pairs;
my $value = shift @item_pairs;
put_item( $name, $value);
$is_updated = 1;
}

if (@MISSING_ITEMS)
{
if (get_item( 'build_start') == 0)
{
my $gbs_build = ENV_getenv( 'GBS_BUILD');   # VVVYYMMDD
put_item( build_start => $gbs_build);
}
@MISSING_ITEMS = ();
$is_updated = 1;
}
write_file();
$RUNNING_FIX = 0;

return $is_updated;
}




sub read_file($)
{
my ($root_path) = @_;

my $file_spec = "$root_path/system.gbs";

if ($file_spec ne $CURRENT_FILE)
{
general_init();

GBSFILE_open( $file_spec, 1, undef, undef);
my $line = GBSFILE_get_line();
while ($line)
{

my ($name, $value) = split( /=/,$line, 2);
$value = '' if (!defined $value);
my ($item, $osname) = split( /-/, $name, 2);
$osname = '' if (!defined $osname);
my $def_ref = $DEFS{$item};
if (defined $def_ref)
{
my ($is_os_item, $is_ref_list, $default, $empty_ok, $check) = @{$def_ref};
if ($is_ref_list)
{
my $values_ref;
if ($value ne '')
{
$values_ref = [ split( ',', $value) ];
} else
{
if (!$empty_ok)
{
GBSFILE_warning( "Item '$name' must have a value");
$values_ref = [ ];
}
}
if ($is_os_item)
{
push @{$DATA{$item}->{$osname}}, $values_ref;
$OTHER_OS{$osname} = 1
if ($osname ne $CUR_OS);
} else
{
push @{$DATA{$item}}, $values_ref;
}
} else
{
if ($value ne '')
{
check_value( $item, $default, $check, $value)
if (defined $check);
} else
{
if (!$empty_ok)
{
GBSFILE_warning( "Item '$name' must have a value");
$value = $default;
}
}
if ($is_os_item)
{
$DATA{$item}->{$osname} = $value;
$OTHER_OS{$osname} = 1
if ($osname ne $CUR_OS);
} else
{
$DATA{$item} = $value;
}
}
} else
{
GBSFILE_warning( "Invalid item '$name'");
}
$line = GBSFILE_get_line();
}




foreach my $item (@ITEMS)
{
my (undef, undef, $default, $empty_ok) = @{$DEFS{$item}};
if (!$empty_ok && !defined $DATA{$item})
{
GBSFILE_warning( "Item '$item' missing");
$DATA{$item} = $default;
push @MISSING_ITEMS, $item;
}
}

ENV_warn( "Errors found. Use gbsmaint 7 5 (Fix system.gbs) to repair")
if (@MISSING_ITEMS && !$RUNNING_FIX);

$CURRENT_FILE = $file_spec;
$FILE_CHANGED = 0;
}
}




sub get_item($)
{
my ($item,
) = @_;
my $value;

$value = $DATA{$item};

return $value;
}




sub get_os_item($$)
{
my ($osname,
$item,
) = @_;
my $value;



$value = $DATA{$item}->{$osname};
if (!defined $value)
{
my $default = $DEFS{$item}->[2];   # $default;
my @available_keys = sort( grep( $DATA{$item}->{$_}, keys( %OTHER_OS)));
ENV_warn( "$item: No data for this OS ($osname)",
"Available for OS: (@available_keys)",
"Assume: $default");
$value = $default;
$DATA{$item}->{$osname} = $value;
}

return $value;
}





sub put_item($$)
{
my ($item,
$value,	    # or $values_ref
) = @_;

$DATA{$item} = $value;

$FILE_CHANGED = 1;
}






sub put_os_item($$$)
{
my ($osname,    # Note that osname must already be present as scms!
$item,
$value,	    # or $values_ref
) = @_;

ENV_error( "Invalid OSname '$osname' in name '$item'")
if (!exists $DATA{scms}->{$osname});

if (ref $value)
{
$value = [ sort( { $a->[0] cmp $b->[0] } @{$value}) ];
}

$DATA{$item}->{$osname} = $value;

$FILE_CHANGED = 1;
}





sub write_file()
{
my @lines;
push @lines, '#';
push @lines, '#   SYSTEM.GBS';
push @lines, '#';

foreach my $item (@ITEMS)
{
my ($is_os_item, $is_ref_list, $default) = @{$DEFS{$item}};
if ($is_os_item)
{
while (my ($os, $value) = each %{$DATA{$item}})
{
my $value = $DATA{$item}->{$os};
$value = $default
if (!defined $value);
if ($is_ref_list)
{
push @lines, map { "$item-$os=" . join( ',', @{$_}) } @{$value};
} else
{
push @lines, "$item-$os=$value";
}
}
} else
{
my $value = $DATA{$item};
if ($is_ref_list)
{
push @lines, map { "$item=" . join( ',', @{$_}) } @{$value};
} else
{
push @lines, "$item=$value";
}
}
}

push @lines, '###EOF###';
map { $_ .= "\n" } @lines;


ENV_backup_file( $CURRENT_FILE, '.sav', 0)
if (-e $CURRENT_FILE);
SCM_store_file( $CURRENT_FILE, \@lines);

$FILE_CHANGED = 0;
}




sub general_init()
{
%DATA = ();
%OTHER_OS = ();
foreach my $item (@ITEMS)
{
my ($is_os_item, $is_ref_list, undef, $empty_ok) = @{$DEFS{$item}};
my $default = ($is_ref_list) ? [] : ($empty_ok) ? '' : undef;
if ($is_os_item)
{
$DATA{$item}->{$CUR_OS} = $default;
} else
{
$DATA{$item} = $default;
}
}
$DATA{id} = 'system';
}




sub check_value($$$$)
{
my ($item,
$default,
$check,
$value,
) = @_;

my $old_value = $value;
if (ref $check eq 'ARRAY')
{
if (!grep( $value eq $_, @{$check}))
{
if (grep (lc $value eq lc $_, @{$check}))
{
$value = lc $value;
} else
{
$value = $default;
}
GBSFILE_warning( "Invalid value: $old_value",
"Possible value(s): (@{$check})",
"Assumed: $value");
}
} elsif (ref $check eq 'Regexp')
{
if ($value !~ $check)
{
$value = $default;
GBSFILE_warning( "Invalid format value: $old_value",
"Regexp: $check",
"Assumed: $value");
}
} else
{
ENV_fatal( 'Impossible else. $check=' . ref $check);
}
}

return 1;
