#===============================================
#
#   validate.pm
#
#===============================================================
#   This file is a part of the GBS (Generic Build Support)
#   package created by Randy Marques, Netherlands
#   For information: http://www.genericbuildsupport.com
#   Copyright  2010-2020 - Randy Marques - All rights reserved
#===============================================================

package validate;
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
VALIDATE_gbs
VALIDATE_root
VALIDATE_subsys
VALIDATE_subsys_full_gbs
VALIDATE_subsys_audit
VALIDATE_component
VALIDATE_target
VALIDATE_target_no_subsys
VALIDATE_targets
VALIDATE_audit
VALIDATE_audit_no_subsys
VALIDATE_audits
VALIDATE_integrator
VALIDATE_administrator
);

use strict;
use integer;

use glo::env;
use glo::list;
use glo::ask;
use gbsenv;
use gbsglo;




sub VALIDATE_gbs();
sub VALIDATE_root();
sub VALIDATE_subsys();
sub VALIDATE_subsys_full_gbs();
sub VALIDATE_subsys_audit();
sub VALIDATE_component();
sub VALIDATE_target();
sub VALIDATE_target_no_subsys();
sub VALIDATE_targets(@);
sub VALIDATE_audit($);
sub VALIDATE_audit_no_subsys($);
sub VALIDATE_audits(@);
sub VALIDATE_integrator();
sub VALIDATE_administrator();

sub validate_user($$);




my $GBS_EXEC_MODE = ENV_getenv('GBS_EXEC_MODE');
my $GBS_ROOT_PATH = GBSENV_gbs_root_path();
my $GBS_SUBSYS = '';
my $GBS_SUBSYS_PATH = '';

my @GBS_ALL_TARGETS = ENV_getenv( 'GBS_ALL_TARGETS');
my @GBS_TARGETS = ENV_getenv( 'GBS_TARGETS');
my $GBS_TARGET = ENV_getenv( 'GBS_TARGET');

my @GBS_ALL_AUDITS = ENV_getenv( 'GBS_ALL_AUDITS');
my @GBS_AUDITS = ENV_getenv( 'GBS_AUDITS');


my $ROOT_VALIDATED = 0;




sub VALIDATE_gbs()
{
ENV_error_exit( "Unable to continue: GBS not active")
if ($GBS_EXEC_MODE eq '');
}






sub VALIDATE_root()
{
if (!$ROOT_VALIDATED)
{
VALIDATE_gbs();
ENV_error_exit( "Unable to continue: No current Root")
if ($GBS_ROOT_PATH eq "");
$ROOT_VALIDATED = 1;
}

return $GBS_ROOT_PATH;
}






sub VALIDATE_subsys()
{
VALIDATE_root();
$GBS_SUBSYS = ENV_getenv( 'GBS_SUBSYS');
ENV_error_exit( "Unable to continue: No current SubSystem")
if ($GBS_SUBSYS eq "");
$GBS_SUBSYS_PATH = "$GBS_ROOT_PATH/dev/$GBS_SUBSYS";

return ($GBS_ROOT_PATH, $GBS_SUBSYS);
}






sub VALIDATE_subsys_full_gbs()
{
VALIDATE_subsys();

ENV_error_exit( "This command is only possible for full GBS Subsystems")
if (! GBSGLO_subsystem_is_full_gbs( $GBS_SUBSYS));

return ($GBS_ROOT_PATH, $GBS_SUBSYS);
}






sub VALIDATE_subsys_audit()
{
VALIDATE_subsys_full_gbs();

ENV_error_exit( "No Audit(s) defined for SubSystem '$GBS_SUBSYS'")
if (!scalar glob "$GBS_SUBSYS_PATH/audit/*");

return ($GBS_ROOT_PATH, $GBS_SUBSYS);
}






sub VALIDATE_component()
{
my $component;

VALIDATE_subsys_full_gbs();

$component = ENV_getenv( 'GBS_COMPONENT');
ENV_error_exit( "Unable to continue: No current Component")
if ($component eq "");

return ($GBS_ROOT_PATH, $GBS_SUBSYS, $component);
}




sub VALIDATE_target()
{
VALIDATE_target_no_subsys();

if ($GBS_SUBSYS ne '')
{
ENV_error_exit( "SubSystem ($GBS_SUBSYS) does not generate for Target '$GBS_TARGET'")
if (! GBSGLO_subsystem_does_build( $GBS_SUBSYS, $GBS_TARGET));
}
}




sub VALIDATE_target_no_subsys()
{
ENV_error_exit( "Unable to continue: No current Target")
if ($GBS_TARGET eq "");
ENV_error_exit( "No such Target '$GBS_TARGET' (@GBS_TARGETS)")
if (! -d "$GBS_ROOT_PATH/sysbuild/$GBS_TARGET");
}






sub VALIDATE_targets(@)
{
my @in_targets = @_;		# possibly wild Targets, '.' == current, '-' = None
my @out_targets;

if (!@GBS_ALL_TARGETS)
{
ENV_error_exit( "No Targets defined")
if (!@GBS_ALL_TARGETS);
ENV_error_exit( "No Targets defined for this Platform ($^O)")
if (!@GBS_TARGETS);
}

foreach my $target (@in_targets)
{
if ($target eq '.')
{
ENV_error_exit( "No current Target")
if ($GBS_TARGET eq '');
push @out_targets, $GBS_TARGET;
} elsif ($target eq '-')
{
push @out_targets, $target;
} else
{
my @exp_targets = ENV_wildcard( $target, [ @GBS_TARGETS ]);
if (!@exp_targets)
{
@exp_targets = ENV_wildcard( $target, [ @GBS_ALL_TARGETS ]);
if (@exp_targets)
{
ENV_error_exit( "No such Target '$target' for this Platform ($^O)");
} else
{
ENV_error_exit( "No such Target '$target' (@GBS_TARGETS)");
}
}
push @out_targets, @exp_targets;
}
}

if (@out_targets)
{
LIST_unique( \@out_targets);
} else
{
ENV_error_exit( "No current Target")
if ($GBS_TARGET eq '');
@out_targets = ($GBS_TARGET);
}

return @out_targets;
}






sub VALIDATE_audit($)
{
my ($audit) = @_;

VALIDATE_audit_no_subsys( $audit);
ENV_error_exit( "SubSystem ($GBS_SUBSYS) does not audit for Audit '$audit' and Target '$GBS_TARGET'")
if (! GBSGLO_subsystem_does_audit( $GBS_SUBSYS, $audit, $GBS_TARGET));

return $audit;
}






sub VALIDATE_audit_no_subsys($)
{
my ($audit) = @_;

my @ALL_AUDITS = ENV_getenv( 'GBS_ALL_AUDITS');
ENV_error_exit( "No such Audit '$audit' (@ALL_AUDITS)")
if (! -d "$GBS_ROOT_PATH/sysaudit/$audit");

return $audit;
}






sub VALIDATE_audits(@)
{
my @audits = @_;
my @new_audits;

my @ALL_AUDITS = ENV_getenv( 'GBS_ALL_AUDITS');
ENV_error_exit( "No Audits defined")
if (!@ALL_AUDITS);

foreach my $audit (@audits)
{
if ($audit eq '.')
{
my $gbs_audit = ENV_getenv( 'GBS_AUDIT');
ENV_error_exit( 'No current Audit')
if ($gbs_audit eq '');
push @new_audits, $gbs_audit;
} elsif ($audit eq '-')
{
push @new_audits, $audit;
} else
{
my @exp_audits = ENV_wildcard( $audit, \@ALL_AUDITS );
ENV_error_exit( "No such Audit '$audit' (@ALL_AUDITS)")
if (!@exp_audits);
push @new_audits, @exp_audits;
}
}


if (@new_audits)
{
LIST_unique( \@new_audits);
}

return @new_audits;
}




sub VALIDATE_integrator()
{
return validate_user( GBSENV_is_integrator(), 'Integrator');
}




sub VALIDATE_administrator()
{
return validate_user( GBSENV_is_administrator(), 'Administrator');
}




sub validate_user($$)
{
my ($is_ok,
$user_type, 	# 'Administrator', 'Integrator'
) = @_;


if (!$is_ok)
{
ENV_beep();
ENV_say( 1, "**");
ENV_say( 1, "** Warning: You are not an $user_type for this System");
ENV_say( 1, "**");
$is_ok = (ASK_YN( "Continue anyway?", 'N') eq 'Y');

ENV_warn_exit() if (!$is_ok && !defined wantarray)
}

return $is_ok;
}

return 1;

